@@ -35,10 +35,17 @@
           ;; more-or-less confined to serve-event, except for a test which now
           ;; detects whether COMPUTE-POLLFDS is defined and therefore testable.
           :OS-PROVIDES-POLL
+          ;; Silly feature we add in because additive features are more sensible
+          ;; than subtractive ones. So you opt out, not in, to having *LOAD-TRUENAME*
+          ;; eagerly bound.
+          :ANSI-COMPLIANT-LOAD-TRUENAME
           ;; The final batch of symbols is strictly for C. The LISP_FEATURE_
           ;; prefix on the corresponding #define is unfortunate.
           :GCC-TLS :USE-SYS-MMAP
-          :OS-PROVIDES-BLKSIZE-T)")) ; only for 'src/runtime/wrap.h'
+          ;; only for 'src/runtime/wrap.h'
+          :OS-PROVIDES-BLKSIZE-T
+          ;; only for src/runtime/run-program.c
+          :OS-PROVIDES-CLOSE-RANGE-WRAPPER)"))
        (public-features
         (cons
          sb-impl::!sbcl-architecture
@@ -71,7 +78,7 @@
            ;; Developer mode features. A release build will never have them,
            ;; hence it makes no difference whether they're public or not.
            :METASPACE
-           :SB-FLUID :SB-DEVEL :SB-DEVEL-LOCK-PACKAGES)")))
+           :SB-DEVEL :SB-DEVEL-LOCK-PACKAGES)")))
        (removable-features
         (append non-target-features public-features)))
   (defconstant sb-impl:+internal-features+
@@ -211,16 +218,10 @@ Please check that all strings which were not recognizable to the compiler
 (as the first argument to WARN, etc.) are wrapped in SB-FORMAT:TOKENS"))
       wps)))
 
-;;; Either set some more package docstrings, or remove any and all docstrings
-;;; that snuck in (as can happen with any file compiled in warm load)
-;;; depending on presence of the :sb-doc internal feature.
-(if (member :sb-doc sb-impl:+internal-features+)
-  (setf (documentation (find-package "COMMON-LISP") t)
-        "public: home of symbols defined by the ANSI language specification"
-        (documentation (find-package "COMMON-LISP-USER") t)
-        "public: the default package for user code and data"
-        (documentation (find-package "KEYWORD") t)
-        "public: home of keywords")
+;;; If the SB-DOC internal feature is not present, remove any and all
+;;; docstrings that snuck in (as can happen with any file compiled in
+;;; warm load).
+(unless (member :sb-doc sb-impl:+internal-features+)
   (let ((count 0))
     (macrolet ((clear-it (place)
                  `(when ,place
@@ -309,7 +310,7 @@ Please check that all strings which were not recognizable to the compiler
 
   ;; Unintern no-longer-needed stuff before the possible PURIFY in
   ;; SAVE-LISP-AND-DIE.
-  #-(or sb-fluid sb-devel) (!unintern-init-only-stuff)
+  #-sb-devel (!unintern-init-only-stuff)
 
   ;; Mark interned immobile symbols so that COMPILE-FILE knows
   ;; which symbols will always be physically in immobile space.
@@ -426,6 +427,7 @@ Please check that all strings which were not recognizable to the compiler
         #.(find-package "SB-PCL")
         #.(find-package "SB-MOP")
         #.(find-package "SB-PRETTY")
+        #.(find-package "SB-BROTHERTREE")
         #.(find-package "SB-KERNEL"))
        ;; Assume all and only external symbols must be retained
        (eq accessibility :external))
