@@ -46,8 +46,7 @@
         (let ((v (%make-lisp-obj (sap-int (sap+ pointer other-pointer-lowtag)))))
           (setf (%array-fill-pointer v) length
             ;; then store the widetag
-                (sap-ref-8 pointer #+big-endian (1- sb-vm:n-word-bytes)
-                                   #+little-endian 0)
+                (sap-ref-8 pointer (+ #+big-endian (1- n-word-bytes)))
                 widetag)
           v)
         (error 'simple-storage-condition
@@ -118,7 +117,7 @@
   ;; provided that no zero bits would be right-shifted out.
   (aver (zerop (logand newval lowtag-mask)))
   (setf (deref varyobj-pages index)
-        (logior (ash newval (- 8 (1+ sb-vm:word-shift)))
+        (logior (ash newval (- 8 (1+ word-shift)))
                 (logand (deref varyobj-pages index) #xFF)))
   newval)
 
@@ -225,7 +224,7 @@
            (- (nth-value 1 (ceiling free-ptr immobile-card-bytes)))))
       (setf (sap-ref-word (int-sap free-ptr) 0) simple-array-fixnum-widetag
             (%array-fill-pointer
-             (%make-lisp-obj (logior free-ptr sb-vm:other-pointer-lowtag)))
+             (%make-lisp-obj (logior free-ptr other-pointer-lowtag)))
             ;; Convert bytes to words, subtract the header and vector length.
             (- (ash n-trailing-bytes (- word-shift)) 2)))))
 
@@ -432,9 +431,8 @@
                   (logior (ash (1- symbol-size) n-widetag-bits) symbol-widetag)))))
     ;; symbol-hash was initialized to 0
     (%set-symbol-global-value symbol (make-unbound-marker))
-    (setf (symbol-%info symbol) nil)
-    (%primitive sb-vm::set-slot symbol name
-                'make-symbol sb-vm:symbol-name-slot sb-vm:other-pointer-lowtag)
+    (%primitive set-slot symbol nil 'make-symbol symbol-info-slot other-pointer-lowtag)
+    (%primitive set-slot symbol name 'make-symbol symbol-name-slot other-pointer-lowtag)
     (%set-symbol-package symbol nil)
     symbol))
 
@@ -452,21 +450,25 @@
 (defun immobile-space-obj-p (obj)
   (immobile-space-addr-p (get-lisp-obj-address obj)))
 
+(define-load-time-global *codeblob-tree* nil)
+
 ;;; Enforce limit on boxed words based on maximum total number of words
 ;;; that can be indicated in the header for 32-bit words.
 ;;; 22 bits = 4MiB, quite generous for one code object.
+;;;   (I think the reasoning behind 22 is that after adding 8 for the widetag,
+;;;    it's 30, and the other 2 bits are the fixnum tag)
 ;;; 25 bits is the maximum unboxed size expressed in bytes,
 ;;; if n-word-bytes = 8 and almost the entire code object is unboxed
 ;;; which is, practically speaking, not really possible.
-(declaim (ftype (sfunction (t (unsigned-byte 16) (unsigned-byte 22) (unsigned-byte 25))
-                           code-component)
+(declaim (ftype (sfunction (t (unsigned-byte 22) (unsigned-byte 25))
+                           (values code-component (integer 0)))
                 allocate-code-object))
 ;;; Allocate a code component with BOXED words in the header
 ;;; followed by UNBOXED bytes of raw data.
 ;;; BOXED must be the exact count of boxed words desired. No adjustments
 ;;; are made for alignment considerations or the fixed slots.
-(defun allocate-code-object (space n-named-calls boxed unboxed)
-  (declare (ignorable space n-named-calls))
+(defun allocate-code-object (space boxed unboxed)
+  (declare (ignorable space))
   (let* ((total-words
            (the (unsigned-byte 22) ; Enforce limit on total words as well
                 (align-up (+ boxed (ceiling unboxed n-word-bytes)) 2)))
@@ -478,55 +480,31 @@
                  ;; Indicate that there are initially 2 boxed words, otherwise
                  ;; immobile space GC thinks this object is freeable.
                  (alloc-immobile-code (ash total-words word-shift)
-                                        (logior (ash total-words code-header-size-shift)
-                                                code-header-widetag)
-                                        (ash 2 n-fixnum-tag-bits)
-                                        other-pointer-lowtag
-                                        (eq space :immobile)))
-              ;;; x86-64 has a vop which is nothing more than wrapping
-               ;; pseudo-atomic around a call to alloc_code_object() in the C runtime.
-               ;; The vop is defined in such a way that it can't be inserted into
-               ;; this fuction, but instead needs an out-of-line call to a helper function
-               ;; (because it clobbers all registers and doesn't indicate that)
-               #+(and x86-64 (not win32))
-               (alloc-dynamic-space-code total-words)
-               #-(and x86-64 (not win32))
+                                      (logior (ash total-words code-header-size-shift)
+                                              code-header-widetag)
+                                      (* boxed n-word-bytes)
+                                      other-pointer-lowtag
+                                      (eq space :immobile)))
+               ;; x86-64 has a vop which wraps pseudo-atomic around the foreign call,
+               ;; as is the custom for allocation trampolines.
+               #+x86-64 (%primitive alloc-code total-words boxed)
+               #-x86-64
                (without-gcing
                  (%make-lisp-obj
                   (alien-funcall (extern-alien "alloc_code_object"
-                                               (function unsigned (unsigned 32)))
-                                 total-words))))
-           #+cheneygc
-           (%primitive var-alloc total-words 'alloc-code
-                       ;; subtract 1 because var-alloc always adds 1 word
-                       ;; for the header, which is not right for code objects.
-                       -1 code-header-widetag other-pointer-lowtag nil)))
-
+                                               (function unsigned (unsigned 32) (unsigned 32)))
+                                 total-words boxed))))))
     (with-pinned-objects (code)
       (let ((sap (sap+ (int-sap (get-lisp-obj-address code))
                        (- other-pointer-lowtag))))
-        ;; The immobile space allocator pre-zeroes, and also it needs a nonzero
-        ;; value in the boxed word count because otherwise it looks like
-        ;; an immobile space page filler. So don't do any more zeroing there.
-        ;; (Could dead immobile objects be converted to use FILLER-WIDETAG instead?)
-        (unless (immobile-space-obj-p code)
-          ;; Before writing the boxed word count, zeroize up to and including 1 word
-          ;; after the boxed header so that all pointers can be safely read by GC,
-          ;; and so that the jump table count word is 0.
-          (loop for byte-index from (ash boxed word-shift) downto (ash 2 word-shift)
-                by n-word-bytes
-                do (setf (sap-ref-word-jit sap byte-index) 0)))
-        ;; The 1st slot beyond the header stores the boxed header size in bytes
-        ;; as an untagged number, which has the same representation as a tagged
-        ;; value denoting a word count if WORD-SHIFT = N-FIXNUM-TAG-BITS.
-        ;; This boxed-size MUST be 0 prior to writing any pointers into the object
-        ;; because the boxed words will not necessarily have been pre-zeroed;
-        ;; scavenging them prior to zeroing them out would see wild pointers.
-        (setf (sap-ref-word-jit sap (ash code-boxed-size-slot word-shift))
-              ;; For 32-bit words, we'll have to add another primitive-object slot.
-              ;; But so far nothing makes use of the n-named-calls value.
-              (logior #+64-bit (ash n-named-calls 32)
-                      (ash boxed word-shift)))))
+        ;; Record it in the balanced tree.
+        (let ((tree *codeblob-tree*) (addr (sap-int sap)))
+          (loop (let ((newtree (sb-brothertree:insert addr tree)))
+                  ;; check that it hasn't been promoted from gen0 -> gen1 already
+                  ;; (very unlikely, but certainly possible).
+                  (unless (eq (generation-of code) 0) (return))
+                  (let ((oldval (cas *codeblob-tree* tree newtree)))
+                    (if (eq oldval tree) (return) (setq tree oldval))))))))
 
     ;; FIXME: there may be random values in the unboxed payload and it's not obvious
     ;; that all callers of ALLOCATE-CODE-OBJECT always write all raw bytes.
@@ -536,10 +514,4 @@
     ;; But what about other things that create code objects?
     ;; It could be a subtle source of nondeterministic core images.
 
-    ;; FIXME: Sort out 64-bit and cheneygc.
-    #+(and 64-bit cheneygc)
-    (setf (code-header-ref code 0)
-          (%make-lisp-obj
-           (logior (ash total-words 32)
-                   sb-vm:code-header-widetag)))
-    code))
+    (values code total-words)))
