@@ -136,6 +136,9 @@
 
   (/show0 "entering !COLD-INIT")
   #+sb-show (setq */show* t)
+  (setq sb-vm::*codeblob-tree* nil)
+  (setq sb-kernel::*defstruct-hooks* '(sb-kernel::!bootstrap-defstruct-hook)
+        sb-kernel::*struct-accesss-fragments-delayed* nil)
   (let ((stream (!make-cold-stderr-stream)))
     (setq *error-output* stream
           *standard-output* stream
@@ -405,7 +408,8 @@ process to continue normally."
   (when (eq *invoke-debugger-hook* 'sb-debug::debugger-disabled-hook)
     (sb-debug::disable-debugger))
   (call-hooks "initialization" *init-hooks*)
-  #+sb-thread (finalizer-thread-start))
+  #+sb-thread (finalizer-thread-start)
+  (sb-vm::setup-cpu-specific-routines))
 
 ;;;; some support for any hapless wretches who end up debugging cold
 ;;;; init code
