@@ -20,18 +20,18 @@
   #+gencgc sb-vm:dynamic-space-start
   #-gencgc (extern-alien "current_dynamic_space" unsigned-long))
 
-#+(or x86 x86-64)
+#+gencgc
 (progn
   (declaim (inline dynamic-space-free-pointer))
   (defun dynamic-space-free-pointer ()
-    (extern-alien "dynamic_space_free_pointer" system-area-pointer)))
+    (sap+ (int-sap (current-dynamic-space-start))
+          (* (extern-alien "next_free_page" signed) sb-vm:gencgc-page-bytes))))
 
 (declaim (inline dynamic-usage))
-#+gencgc
-(defun dynamic-usage ()
-  (extern-alien "bytes_allocated" os-vm-size-t))
-#-gencgc
 (defun dynamic-usage ()
+  #+gencgc
+  (extern-alien "bytes_allocated" os-vm-size-t)
+  #-gencgc
   (truly-the word
              (- (sap-int (sb-c::dynamic-space-free-pointer))
                 (current-dynamic-space-start))))
@@ -63,6 +63,13 @@
 (define-load-time-global *n-bytes-freed-or-purified* 0)
 (defun gc-reinit ()
   (setq *gc-inhibit* nil)
+  ;; This GC looks utterly unnecessary, but when I tried removing it,
+  ;; I got heap exhaustions sooner than when I left it in.
+  ;; So why is that? Well, it seems like there's some state that is getting
+  ;; saved into the core which needs resetting, such as these SETQs,
+  ;; but how is that affecting the C runtime's decision on when to
+  ;; perform the first auto-triggered GC? How is not doing an explicit
+  ;; GC delaying the decision? It seems wrong.
   (gc)
   (setf *n-bytes-freed-or-purified* 0
         *gc-run-time* 0))
@@ -78,26 +85,58 @@ and submit it as a patch."
   (+ (dynamic-usage)
      *n-bytes-freed-or-purified*))
 
-;;; * If symbols are 7 words incl header, as they are on 32-bit w/threads,
-;;;   then the part of NIL that manifests as symbol slots consumes 6 words
-;;;   (less the header) because NIL's symbol widetag precedes it by 1 word.
-;;;   So, there are 6 post-header words and 2 pre-header: one containing
-;;;   the widetag (not that it is ever read), and one 0 word.
-;;; * If symbols are 6 words incl header, as they are on 64-bit and
-;;;   32-bit w/o threads, then the symbol-like part of NIL is 5 words,
-;;;   which aligns up to 6, plus the two pre-header words.
-;;; So either way, NIL's alignment padding makes it come out to 8 words in total.
-(defconstant sb-vm::sizeof-nil-in-words (+ 2 (sb-int:align-up (1- sb-vm:symbol-size) 2)))
-
 (defun primitive-object-size (object)
   "Return number of bytes of heap or stack directly consumed by OBJECT"
   (cond ((not (sb-vm:is-lisp-pointer (get-lisp-obj-address object))) 0)
         ((eq object nil) (ash sb-vm::sizeof-nil-in-words sb-vm:word-shift))
         ((simple-fun-p object) (code-object-size (fun-code-header object)))
+        #-(or x86 x86-64 arm64 riscv) ((lra-p object) 1)
         (t
          (with-alien ((sizer (function unsigned unsigned) :extern "primitive_object_size"))
            (with-pinned-objects (object)
              (alien-funcall sizer (get-lisp-obj-address object)))))))
+
+(defun sb-vm:hexdump (thing &optional (n-words nil wordsp)
+                            ;; pass NIL explicitly if T crashes on you
+                            (decode t))
+  (multiple-value-bind (obj addr count)
+      (if (typep thing 'word) ; ambiguous in the edge case, but assume it's
+          ;; an address (though you might be trying to dump a bignum's data)
+          (values nil thing (if wordsp n-words 1))
+          (values
+           thing
+           (logandc2 (get-lisp-obj-address thing) sb-vm:lowtag-mask)
+           (if wordsp
+               n-words
+               (if (and (typep thing 'code-component) (plusp (code-n-entries thing)))
+                   ;; Display up through the first fun header
+                   (+ (code-header-words thing)
+                      (ash (%code-fun-offset thing 0) (- sb-vm:word-shift))
+                      sb-vm:simple-fun-insts-offset)
+                   ;; at most 16 words
+                   (min 16 (ash (primitive-object-size thing) (- sb-vm:word-shift)))))))
+    (with-pinned-objects (obj)
+      (dotimes (i count)
+        (let ((word (sap-ref-word (int-sap addr) (ash i sb-vm:word-shift))))
+          (multiple-value-bind (lispobj ok fmt)
+              (cond ((and (typep thing 'code-component)
+                          (< 1 i (code-header-words thing)))
+                     (values (code-header-ref thing i) t))
+                    #+compact-symbol
+                    ((and (typep thing '(and symbol (not null)))
+                          (= i sb-vm:symbol-name-slot))
+                     (values (list (sb-impl::symbol-package-id thing)
+                                   (symbol-name thing))
+                             t
+                             "{~{~A,~S~}}"))
+                    (decode
+                     (make-lisp-obj word nil)))
+            (let ((*print-lines* 1)
+                  (*print-pretty* t))
+              (format t "~x: ~v,'0x~:[~; = ~@?~]~%"
+                      (+ addr (ash i sb-vm:word-shift))
+                      (* 2 sb-vm:n-word-bytes)
+                      word ok (or fmt "~S") lispobj))))))))
 
 ;;;; GC hooks
 
@@ -109,8 +148,7 @@ run in any thread.")
 
 ;;;; internal GC
 
-(define-alien-routine collect-garbage int
-  (#+gencgc last-gen #-gencgc ignore int))
+(define-alien-routine collect-garbage int (last-gen int))
 
 #+(or sb-thread sb-safepoint)
 (progn
@@ -121,35 +159,10 @@ run in any thread.")
   (defun gc-stop-the-world ())
   (defun gc-start-the-world ()))
 
-#+gencgc
-(progn
-  (define-alien-variable ("gc_logfile" %gc-logfile) (* char))
-  (defun (setf gc-logfile) (pathname)
-    (let ((new (when pathname
-                 (make-alien-string
-                  (native-namestring (translate-logical-pathname pathname)
-                                     :as-file t))))
-          (old %gc-logfile))
-      (setf %gc-logfile new)
-      (when old
-        (free-alien old))
-      pathname))
-  (defun gc-logfile ()
-    "Return the pathname used to log garbage collections. Can be SETF.
-Default is NIL, meaning collections are not logged. If non-null, the
-designated file is opened before and after each collection, and generation
-statistics are appended to it."
-    (let ((val (cast %gc-logfile c-string)))
-      (when val
-        (native-pathname val)))))
-
 (declaim (inline dynamic-space-size))
 (defun dynamic-space-size ()
   "Size of the dynamic space in bytes."
   (extern-alien "dynamic_space_size" os-vm-size-t))
-#+gencgc
-(define-symbol-macro sb-vm:dynamic-space-end
-  (+ (dynamic-space-size) sb-vm:dynamic-space-start))
 
 ;;;; SUB-GC
 
@@ -408,16 +421,37 @@ Note: currently changes to this value are lost when saving core."
 
 ;;;; GENCGC specifics
 ;;;;
-;;;; For documentation convenience, these have stubs on non-GENCGC platforms
-;;;; as well.
-#+gencgc
+(define-symbol-macro sb-vm:dynamic-space-end
+    ;; FIXME: what uses this? Nothing? If so, remove it.
+    (+ (dynamic-space-size) sb-vm:dynamic-space-start))
+
+(define-alien-variable ("gc_logfile" %gc-logfile) (* char))
+
+(defun (setf gc-logfile) (pathname)
+  (let ((new (when pathname
+               (make-alien-string
+                (native-namestring (translate-logical-pathname pathname)
+                                   :as-file t))))
+        (old %gc-logfile))
+    (setf %gc-logfile new)
+    (when old
+      (free-alien old))
+    pathname))
+(defun gc-logfile ()
+    "Return the pathname used to log garbage collections. Can be SETF.
+Default is NIL, meaning collections are not logged. If non-null, the
+designated file is opened before and after each collection, and generation
+statistics are appended to it."
+  (let ((val (cast %gc-logfile c-string)))
+    (when val
+      (native-pathname val))))
+
 (deftype generation-index ()
   `(integer 0 ,sb-vm:+pseudo-static-generation+))
 
 ;;; FIXME: GENERATION (and PAGE, as seen in room.lisp) should probably be
 ;;; defined in Lisp, and written to header files by genesis, instead of this
 ;;; OAOOMiness -- this duplicates the struct definition in gencgc.c.
-#+gencgc
 (define-alien-type generation
     (struct generation
             (bytes-allocated os-vm-size-t)
@@ -428,34 +462,84 @@ Note: currently changes to this value are lost when saving core."
             (cum-sum-bytes-allocated os-vm-size-t)
             (minimum-age-before-gc double)))
 
-#+gencgc
 (define-alien-variable generations
     (array generation #.(1+ sb-vm:+pseudo-static-generation+)))
 
-(macrolet ((def (slot doc &optional setfp)
+;;; Why is PAGE-INDEX-T in SB-KERNEL but PAGE and the page table are in SB-VM?
+(define-alien-type (struct sb-vm::page)
+    (struct sb-vm::page
+            ;; To cut down the size of the page table, the scan_start_offset
+            ;; - a/k/a "start" - is measured in 4-byte integers regardless
+            ;; of word size. This is fine for 32-bit address space,
+            ;; but if 64-bit then we have to scale the value. Additionally
+            ;; there is a fallback for when even the scaled value is too big.
+            (sb-vm::start #+64-bit (unsigned 32) #-64-bit signed)
+            ;; On platforms with small enough GC pages, this field
+            ;; will be a short. On platforms with larger ones, it'll
+            ;; be an int. It should probably never be an int.
+            (sb-vm::words-used (unsigned
+                                #.(if (typep sb-vm::gencgc-page-words '(unsigned-byte 16))
+                                      16
+                                      32)))
+            (sb-vm::flags (unsigned 8)) ; in C this is {type, need_zerofill, pinned}
+            (sb-vm::gen (signed 8))))
+(define-alien-variable ("page_table" sb-vm:page-table) (* (struct sb-vm::page)))
+(declaim (inline sb-vm:find-page-index))
+(define-alien-routine ("ext_find_page_index" sb-vm:find-page-index) page-index-t (address unsigned))
+
+(defun generation-of (object)
+  (with-pinned-objects (object)
+    (let* ((addr (get-lisp-obj-address object))
+           (page (sb-vm:find-page-index addr)))
+      (cond ((>= page 0) (slot (deref sb-vm:page-table page) 'sb-vm::gen))
+            #+immobile-space
+            ((immobile-space-addr-p addr)
+             ;; SIMPLE-FUNs don't contain a generation byte
+             (when (simple-fun-p object)
+               (setq addr (get-lisp-obj-address (fun-code-header object))))
+             (let ((sap (int-sap (logandc2 addr sb-vm:lowtag-mask))))
+               (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
+                       #xF)))))))
+
+(export 'page-protected-p)
+(macrolet ((addr->mark (addr)
+             `(sap-ref-8 (extern-alien "gc_card_mark" system-area-pointer)
+                         (logand (ash ,addr (- (integer-length (1- sb-vm:gencgc-card-bytes))))
+                                 (extern-alien "gc_card_table_mask" int))))
+           (markedp (val)
+             #+soft-card-marks `(eql ,val 0)
+             ;; With physical protection there are 2 single-bit fields packed in the mark byte.
+             ;; The 0 bit is the card mark bit (inverse of protected-p)
+             #-soft-card-marks `(oddp ,val)))
+  (defun page-protected-p (object) ; OBJECT must be pinned by caller
+    (not (markedp (addr->mark (get-lisp-obj-address object)))))
+  (defun object-card-marks (obj)
+    (aver (eq (heap-allocated-p obj) :dynamic))
+    ;; Return a bit-vector with a 1 for each marked card the object is on
+    ;; and a 0 for each unmarked card.
+    (with-pinned-objects (obj)
+      (let* ((base (logandc2 (get-lisp-obj-address obj) sb-vm:lowtag-mask))
+             (last (+ base (1- (primitive-object-size obj))))
+             (aligned-base (logandc2 base (1- sb-vm:gencgc-card-bytes)))
+             (aligned-last (logandc2 last (1- sb-vm:gencgc-card-bytes)))
+             (result (make-array (1+ (truncate (- aligned-last aligned-base)
+                                               sb-vm:gencgc-card-bytes))
+                                 :element-type 'bit)))
+        (loop for addr from aligned-base to aligned-last by sb-vm:gencgc-card-bytes
+              for i from 0
+              do (setf (aref result i) (if (markedp (addr->mark addr)) 1 0)))
+        result))))
+
+(macrolet ((def (slot doc &optional adjustable)
              `(progn
                 (defun ,(symbolicate "GENERATION-" slot) (generation)
                   ,doc
-                  #+gencgc
                   (declare (generation-index generation))
-                  #-gencgc
-                  (declare (ignore generation))
-                  #-gencgc
-                  (error "~S is a GENCGC only function and unavailable in this build"
-                         ',slot)
-                  #+gencgc
                   (slot (deref generations generation) ',slot))
-                ,@(when setfp
-                        `((defun (setf ,(symbolicate "GENERATION-" slot)) (value generation)
-                            #+gencgc
-                            (declare (generation-index generation))
-                            #-gencgc
-                            (declare (ignore value generation))
-                            #-gencgc
-                            (error "(SETF ~S) is a GENCGC only function and unavailable in this build"
-                                   ',slot)
-                            #+gencgc
-                            (setf (slot (deref generations generation) ',slot) value)))))))
+                ,@(when adjustable
+                    `((defun (setf ,(symbolicate "GENERATION-" slot)) (value generation)
+                        (declare (generation-index generation))
+                        (setf (slot (deref generations generation) ',slot) value)))))))
   (def bytes-consed-between-gcs
       "Number of bytes that can be allocated to GENERATION before that
 generation is considered for garbage collection. This value is meaningless for
@@ -489,23 +573,17 @@ Experimental: interface subject to change.")
       "Number of times garbage collection has been done on GENERATION without
 promotion. Available on GENCGC platforms only.
 
-Experimental: interface subject to change."))
+Experimental: interface subject to change.")
   (defun generation-average-age (generation)
     "Average age of memory allocated to GENERATION: average number of times
 objects allocated to the generation have seen younger objects promoted to it.
 Available on GENCGC platforms only.
 
 Experimental: interface subject to change."
-    #+gencgc
     (declare (generation-index generation))
-    #-gencgc (declare (ignore generation))
-    #-gencgc
-    (error "~S is a GENCGC only function and unavailable in this build."
-           'generation-average-age)
-    #+gencgc
     (alien-funcall (extern-alien "generation_average_age"
                                  (function double generation-index-t))
-                   generation))
+                   generation)))
 
 (macrolet ((cases ()
              `(cond ((< (current-dynamic-space-start) addr
