@@ -45,8 +45,6 @@
                           complex-array-widetag))
         (setf (svref infos code) info)))
 
-    (setf (svref infos filler-widetag) (make-room-info 'filler))
-
     (dotimes (i (length *specialized-array-element-type-properties*))
       (let ((saetp (aref *specialized-array-element-type-properties* i)))
         (setf (svref infos (saetp-typecode saetp)) saetp)))
@@ -113,7 +111,9 @@
 (defun %space-bounds (space)
   (macrolet ((bounds (a b) `(values (%make-lisp-obj ,a) (%make-lisp-obj ,b))))
     (ecase space
-      (:static
+      (:static ;; These bounds are appropriate for computing the space usage
+               ;; but NOT for computing iteration bounds, because there are
+               ;; "unformatted" words preceding the lowest addressable object.
        (bounds static-space-start
                (sap-int *static-space-free-pointer*)))
       (:read-only
@@ -187,21 +187,31 @@
         (end (descriptor-sap end)))
     (loop
      (if (sap>= start end) (return))
-     (binding* ((widetag (widetag@baseptr start))
-                (obj (lispobj@baseptr start widetag))
-                ((typecode size)
+     (let ((word (sap-ref-word start 0)))
+       (cond
+         ((= (logand word widetag-mask) filler-widetag) ; pseudo-object
+          (let ((size (ash (filler-nwords word) word-shift)))
+            (setq start (sap+ start size))))
+         ((= word most-positive-word)
+          ;; has to be a pseudo-cons resulting from removing an insignificant
+          ;; sign word of a bignum. Don't call FUN
+          (setq start (sap+ start (* 2 n-word-bytes))))
+         (t
+          (binding*
+              ((widetag (widetag@baseptr start))
+               (obj (lispobj@baseptr start widetag))
+               ((typecode size)
                  ;; PRIMITIVE-OBJECT-SIZE works on conses, but they're exceptions already
                  ;; because of absence of a widetag, so may as well not call the sizer.
                  (if (listp obj)
                      (values list-pointer-lowtag (* 2 n-word-bytes))
                      (values widetag (primitive-object-size obj)))))
-       ;; SIZE is surely a fixnum. Non-fixnum would imply at least
-       ;; a 512MB object if 32-bit words, and is inconceivable if 64-bit.
-       ;; But check to be sure.
-       (aver (not (logtest (the fixnum size) lowtag-mask)))
-       (unless (= typecode filler-widetag)
-         (funcall fun obj typecode size))
-       (setq start (sap+ start size))))
+              ;; SIZE is surely a fixnum. Non-fixnum would imply at least
+              ;; a 512MB object if 32-bit words, and is inconceivable if 64-bit.
+              ;; But check to be sure.
+              (aver (not (logtest (the fixnum size) lowtag-mask)))
+              (funcall fun obj typecode size)
+              (setq start (sap+ start size)))))))
     (when strict-bound
      ;; If START is not eq to END, then we have blown past our endpoint.
       #+sb-devel
@@ -215,27 +225,8 @@
 
 ;;; Access to the GENCGC page table for better precision in
 ;;; MAP-ALLOCATED-OBJECTS
-#+gencgc
+#+immobile-space
 (progn
-  (define-alien-type (struct page)
-      (struct page
-              ;; To cut down the size of the page table, the scan_start_offset
-              ;; - a/k/a "start" - is measured in 4-byte integers regardless
-              ;; of word size. This is fine for 32-bit address space,
-              ;; but if 64-bit then we have to scale the value. Additionally
-              ;; there is a fallback for when even the scaled value is too big.
-              (start #+64-bit (unsigned 32) #-64-bit signed)
-              ;; On platforms with small enough GC pages, this field
-              ;; will be a short. On platforms with larger ones, it'll
-              ;; be an int. It should probably never be an int.
-              (words-used (unsigned
-                           #.(if (typep gencgc-page-words '(unsigned-byte 16))
-                                 16
-                                 32)))
-              (flags (unsigned 8)) ; in C this is {type, need_zerofill, pinned}
-              (gen (signed 8))))
-  #+immobile-space
-  (progn
     (define-alien-type (struct immobile-page)
         ;; ... and yet another place for Lisp to become out-of-sync with C.
         (struct immobile-page
@@ -247,11 +238,7 @@
                 (page-link (unsigned 16))
                 (prior-free-index (unsigned 16))))
     (define-alien-variable "fixedobj_pages" (* (struct immobile-page))))
-  (declaim (inline find-page-index))
-  (define-alien-routine ("ext_find_page_index" find-page-index)
-    long (index unsigned))
-  (define-alien-variable "next_free_page" sb-kernel::page-index-t)
-  (define-alien-variable "page_table" (* (struct page))))
+(define-alien-variable "next_free_page" sb-kernel::page-index-t)
 
 #+immobile-space
 (progn
@@ -325,15 +312,10 @@ We could try a few things to mitigate this:
             (:static
              ;; Static space starts with NIL, which requires special
              ;; handling, as the header and alignment are slightly off.
-             (multiple-value-bind (start end) (%space-bounds space)
-               (declare (ignore start))
-               (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
-               (let ((start (+ (logandc2 nil-value lowtag-mask)
-                               (ash (- sizeof-nil-in-words 2) word-shift))))
-                 (map-objects-in-range fun
-                                       (ash start (- n-fixnum-tag-bits))
-                                       end))))
-
+             (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
+             (let ((start (%make-lisp-obj static-space-objects-start))
+                   (end (%make-lisp-obj (sap-int *static-space-free-pointer*))))
+               (map-objects-in-range fun start end)))
             ((:read-only #-gencgc :dynamic)
              ;; Read-only space (and dynamic space on cheneygc) is a block
              ;; of contiguous allocations.
@@ -423,10 +405,7 @@ We could try a few things to mitigate this:
                                                 end-page-bytes-used)
                                              most-positive-word)))))
       (when (sap> end start)
-        ;; The bits in the 6-bit 'type' field have fixed positions,
-        ;; but the position of the field itself depends on endianness.
-        (let ((flags (ldb (byte 6 (+ #+big-endian 2))
-                          (slot (deref page-table start-page) 'flags))))
+        (let ((flags (slot (deref page-table start-page) 'flags)))
           ;; The GEN slot is declared as (SIGNED 8) which does not satisfy the
           ;; type restriction on the first argument to LOGBITP.
           ;; Masking it to 3 bits fixes that, and allows using the other 5 bits
@@ -434,6 +413,8 @@ We could try a few things to mitigate this:
           (when (and (logbitp (logand (slot (deref page-table start-page) 'gen) 7)
                               generation-mask)
                      (= (logand flags page-type-mask) page-type-constraint))
+            ;; FIXME: should exclude (0 . 0) conses on PAGE_TYPE_{BOXED,UNBOXED}
+            ;; resulting from zeroing the tail of a bignum or vector etc.
             (map-objects-in-range fun
                                   (%make-lisp-obj (sap-int start))
                                   (%make-lisp-obj (sap-int end))
@@ -722,6 +703,8 @@ We could try a few things to mitigate this:
            (type stream stream) (type spaces space)
            (type (or index null) type larger smaller count))
   (multiple-value-bind (start end) (%space-bounds space)
+    (when (eq space :static)
+      (setq start (%make-lisp-obj static-space-objects-start)))
     (let* ((space-start (ash start n-fixnum-tag-bits))
            (space-end (ash end n-fixnum-tag-bits))
            (space-size (- space-end space-start))
@@ -914,9 +897,13 @@ We could try a few things to mitigate this:
          ;; These two are in fact generally the most frequently occurring type.
          ,.(make-case 'cons `(car ,obj) `(cdr ,obj))
          ,.(make-case* 'instance
-            `(progn
+            ;; %INSTANCE-LAYOUT is defknown'ed to return a LAYOUT,
+            ;; but heap walking might encounter an instance with no layout,
+            ;; hence the need to access the slot opaquely.
+            `(unless (eql 0 #+compact-instance-header (%primitive %instance-layout ,obj)
+                            #-compact-instance-header (%instance-ref ,obj 0))
                (,functoid (%instance-layout ,obj) ,@more)
-               (do-instance-tagged-slot (.i. ,obj nil)
+               (do-instance-tagged-slot (.i. ,obj)
                  (,functoid (%instance-ref ,obj .i.) ,@more))))
          (function
           (typecase ,obj
@@ -1037,7 +1024,6 @@ We could try a few things to mitigate this:
           (case (widetag-of this)
             (#.value-cell-widetag
              (test (value-cell-ref this)))
-            (#.filler-widetag)
             (t
              (bug "Unknown object type #x~x addr=~x"
                   (widetag-of this)
@@ -1168,8 +1154,8 @@ We could try a few things to mitigate this:
                         (setq page-num next-page seen-filler nil))))))))
     (let ((i 0))
       (loop while (< i total-pages)
-            do (let ((type (ldb (byte 2 0) (slot (deref page-table i) 'flags))))
-                 (if (= type 3)
+            do (let ((type (slot (deref page-table i) 'flags)))
+                 (if (= (logand type 7) 7)
                      (setq i (dump-page i))
                      (incf i)))))
     (let* ((n-pages (count 1 pages))
@@ -1196,21 +1182,6 @@ We could try a few things to mitigate this:
         (%make-lisp-obj varyobj-space-start)
         (%make-lisp-obj (sap-int *varyobj-space-free-pointer*))))))
 
-#+gencgc
-(defun generation-of (object)
-  (with-pinned-objects (object)
-    (let* ((addr (get-lisp-obj-address object))
-           (page (find-page-index addr)))
-      (cond ((>= page 0) (slot (deref page-table page) 'gen))
-            #+immobile-space
-            ((immobile-space-addr-p addr)
-             ;; SIMPLE-FUNs don't contain a generation byte
-             (when (simple-fun-p object)
-               (setq addr (get-lisp-obj-address (fun-code-header object))))
-             (let ((sap (int-sap (logandc2 addr lowtag-mask))))
-               (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
-                       #xF)))))))
-
 ;;; Show objects in a much simpler way than print-allocated-objects.
 ;;; Probably don't use this for generation 0 of dynamic space. Other spaces are ok.
 ;;; (And this is removed from the image; it's meant for developers)
@@ -1253,8 +1224,7 @@ We could try a few things to mitigate this:
          ;; gen0 conses on MIXED pages, but even that is not enough- pinned conses
          ;; will promote but keep their MIXED page type. So don't bother with this.
          #+use-cons-region
-         (let* ((flags (slot (deref page-table (find-page-index obj-addr)) 'flags))
-                (type (ldb (byte 6 (+ #+big-endian 2)) flags))
+         (let* ((type (slot (deref page-table (find-page-index obj-addr)) 'flags))
                 (ok (if (consp obj)
                         (or (= type #b101) ; PAGE_TYPE_CONS
                             (and (eq (car obj) 0) (eq (cdr obj) 0)))
@@ -1281,47 +1251,6 @@ We could try a few things to mitigate this:
 ;;; this is a valid test that genesis separated code and data.
 (!ensure-genesis-code/data-separation)
 
-(defun hexdump (thing &optional (n-words nil wordsp)
-                                ;; pass NIL explicitly if T crashes on you
-                                (decode t))
-  (multiple-value-bind (obj addr count)
-      (if (typep thing 'word) ; ambiguous in the edge case, but assume it's
-          ;; an address (though you might be trying to dump a bignum's data)
-          (values nil thing (if wordsp n-words 1))
-          (values
-           thing
-           (logandc2 (get-lisp-obj-address thing) lowtag-mask)
-           (if wordsp
-               n-words
-               (if (and (typep thing 'code-component) (plusp (code-n-entries thing)))
-                   ;; Display up through the first fun header
-                   (+ (code-header-words thing)
-                      (ash (%code-fun-offset thing 0) (- word-shift))
-                      simple-fun-insts-offset)
-                   ;; at most 16 words
-                   (min 16 (ash (primitive-object-size thing) (- word-shift)))))))
-    (with-pinned-objects (obj)
-      (dotimes (i count)
-        (let ((word (sap-ref-word (int-sap addr) (ash i word-shift))))
-          (multiple-value-bind (lispobj ok fmt)
-              (cond ((and (typep thing 'code-component)
-                          (< 1 i (code-header-words thing)))
-                     (values (code-header-ref thing i) t))
-                    #+compact-symbol
-                    ((and (typep thing '(and symbol (not null)))
-                          (= i symbol-name-slot))
-                     (values (list (sb-impl::symbol-package-id thing)
-                                   (symbol-name thing))
-                             t
-                             "{~{~A,~S~}}"))
-                    (decode
-                     (make-lisp-obj word nil)))
-            (let ((*print-lines* 1)
-                  (*print-pretty* t))
-              (format t "~x: ~v,'0x~:[~; = ~@?~]~%"
-                      (+ addr (ash i word-shift))
-                      (* 2 n-word-bytes)
-                      word ok (or fmt "~S") lispobj))))))))
 #+sb-thread
 (defun show-tls-map ()
   (let ((list
@@ -1429,7 +1358,7 @@ We could try a few things to mitigate this:
     (when (dotimes (i (length v))
             (declare (optimize (sb-c::aref-trapping 0)))
             (let ((val (svref v i)))
-              (when (= (get-lisp-obj-address val) no-tls-value-marker-widetag)
+              (when (= (get-lisp-obj-address val) unwritten-vector-element-marker)
                 (return t))))
       (push (make-weak-pointer v) result)
       (let* ((origin (vector-extra-data v))
