@@ -81,7 +81,7 @@
        sb-kernel::*gc-epoch*))
 
 (defun start-lisp (toplevel callable-exports)
-  (named-lambda start-lisp ()
+  (named-lambda %start-lisp ()
     (cond (callable-exports
            (reinit t)
            (dolist (export callable-exports)
@@ -166,8 +166,8 @@ The following &KEY arguments are defined:
      This is only meaningful if the runtime was built with the :SB-CORE-COMPRESSION
      feature enabled. If NIL (the default), saves to uncompressed core files. If
      :SB-CORE-COMPRESSION was enabled at build-time, the argument may also be
-     an integer from -1 to 9, corresponding to zlib compression levels, or T
-     (which is equivalent to the default compression level, -1).
+     an integer from -7 to 22, corresponding to zstd compression levels, or T
+     (which is equivalent to the default compression level, 9).
 
   :APPLICATION-TYPE
      Present only on Windows and is meaningful only with :EXECUTABLE T.
@@ -214,10 +214,10 @@ sufficiently motivated to do lengthy fixes."
   (let ((toplevel (%coerce-callable-to-fun toplevel))
         *streams-closed-by-slad*)
     #+sb-core-compression
-    (check-type compression (or boolean (integer -1 9)))
+    (check-type compression (or boolean (integer -7 22)))
     #-sb-core-compression
     (when compression
-      (error "Unable to save compressed core: this runtime was not built with zlib support"))
+      (error "Unable to save compressed core: this runtime was not built with zstd support"))
     (when *dribble-stream*
       (restart-case (error "Dribbling to ~s is enabled." (pathname *dribble-stream*))
         (continue ()
@@ -227,7 +227,7 @@ sufficiently motivated to do lengthy fixes."
           :report "Abort saving the core."
           (return-from save-lisp-and-die))))
     (when (eql t compression)
-      (setf compression -1))
+      (setf compression 9))
     (flet ((foreign-bool (value)
              (if value 1 0)))
       (let ((name (native-namestring (physicalize-pathname core-file-name)
@@ -315,7 +315,7 @@ sufficiently motivated to do lengthy fixes."
   (call-hooks "save" *save-hooks*)
   #+win32 (itimer-emulation-deinit)
   #+sb-thread
-  (let (err)
+  (let (error)
     (with-system-mutex (sb-thread::*make-thread-lock*)
       (finalizer-thread-stop)
       #+pauseless-threadstart (sb-thread::join-pthread-joinables #'identity)
@@ -328,10 +328,10 @@ sufficiently motivated to do lengthy fixes."
           (let* ((interactive (sb-thread::interactive-threads))
                  (other (union (set-difference threads interactive)
                                (union starting joinable))))
-            (make-condition 'save-with-multiple-threads-error
-                            :interactive-threads interactive
-                            :other-threads other)))))
-    (when err (error err))
+            (setf error (make-condition 'save-with-multiple-threads-error
+                                        :interactive-threads interactive
+                                        :other-threads other))))))
+    (when error (error error))
     #+allocator-metrics (setq sb-thread::*allocator-metrics* nil)
     (setq sb-thread::*sprof-data* nil))
   (tune-image-for-dump)
@@ -355,8 +355,11 @@ sufficiently motivated to do lengthy fixes."
   #+immobile-code (sb-vm::statically-link-core)
   (invalidate-fd-streams)
   (finalizers-deinit)
+  (sb-vm::restore-cpu-specific-routines)
   ;; Do this last, to have some hope of printing if we need to.
   (stream-deinit)
+  (setf sb-c::*compile-elapsed-time* 0
+        sb-c::*compile-file-elapsed-time* 0)
   (setf * nil ** nil *** nil
         - nil + nil ++ nil +++ nil
         /// nil // nil / nil))
@@ -463,7 +466,8 @@ sufficiently motivated to do lengthy fixes."
                          ((not (memq accessor
                                      '(%closure-fun
                                        %fun-layout %instance-layout
-                                       symbol-package symbol-name fdefn-name
+                                       symbol-package symbol-name symbol-%info
+                                       fdefn-name
                                        %numerator %denominator
                                        %realpart %imagpart
                                        %make-lisp-obj ; fdefn referent
@@ -502,9 +506,7 @@ sb-c::
            (let ((di (sb-vm::%%code-debug-info obj)))
              ;; Discard memoized debugger's debug info
              (when (typep di 'sb-c::compiled-debug-info)
-               (let ((thing (sb-c::compiled-debug-info-tlf-num+offset di)))
-                 (when (consp thing)
-                   (setf (sb-c::compiled-debug-info-tlf-num+offset di) (car thing))))))
+               (setf (sb-c::compiled-debug-info-memo-cell di) nil)))
            (dotimes (i (sb-kernel:code-n-entries obj))
              (let* ((fun (sb-kernel:%code-entry-point obj i))
                     (arglist (%simple-fun-arglist fun))
