@@ -416,8 +416,9 @@
 
 ;;; The fun-table-count is a uint16_t immediately preceding the trailer length
 ;;; containing two subfields:
-;;;  12 bits for the number of simple-funs in the code component
-;;;   4 bits for the number of pad bytes added to align the fun-offset-table
+;;;  11 bits for the number of simple-funs in the code component
+;;;   5 bits for the number of pad bytes added to align the fun-offset-table
+;;; See also code_n_funs() in code.h
 (declaim (inline code-fun-table-count))
 (defun code-fun-table-count (code-obj)
   (if (eql (code-trailer-len code-obj) 0)
@@ -431,7 +432,7 @@
 ;;; Keep in sync with C function code_n_funs()
 (defun code-n-entries (code-obj)
   (declare (type code-component code-obj))
-  (ash (code-fun-table-count code-obj) -4))
+  (ash (code-fun-table-count code-obj) -5))
 
 ;;; Index to start of named-call fdefns
 ;;; FIXME: Naming symmetry between this and code-n-named-calls might be nice.
@@ -450,7 +451,7 @@
 ;;; to its FUN-INDEXth function.
 (declaim (inline %code-fun-offset))
 (defun %code-fun-offset (code-obj fun-index)
-  (declare ((unsigned-byte 12) fun-index))
+  (declare ((unsigned-byte 11) fun-index))
   (code-trailer-ref code-obj (* -4 (+ fun-index 2))))
 
 ;;; Subtract from %CODE-CODE-SIZE the number of trailing data bytes which aren't
@@ -468,8 +469,8 @@
 (defun %code-text-size (code-obj)
   (- (%code-code-size code-obj)
      (code-trailer-len code-obj)
-     ;; Subtract between 0 and 15 bytes of padding
-     (logand (code-fun-table-count code-obj) #xf)))
+     ;; Subtract between 0 and 31 bytes of padding
+     (logand (code-fun-table-count code-obj) #x1f)))
 
 (defun %code-entry-point (code-obj fun-index)
   (declare (type (unsigned-byte 16) fun-index))
@@ -631,33 +632,3 @@
   (do-closure-values (value closure)
     (when (funcall test value)
       (return value))))
-
-(in-package "SB-C")
-
-;;; Decode the packed TLF-NUM+OFFSET slot, which might have ancillary
-;;; data for the debugger pushed in.  So if it's a cons, take the CDR.
-;;; This is target-only code, so doesn't belong in 'debug-info.lisp'
-(defun unpack-tlf-num+offset (tlf-num+offset
-                              &aux (integer (if (consp tlf-num+offset)
-                                                (car tlf-num+offset)
-                                                tlf-num+offset))
-                                   (bytepos 0))
-  (flet ((unpack-1 ()
-           (let ((shift 0) (acc 0))
-             (declare (notinline sb-kernel:%ldb)) ; lp#1573398
-             (loop
-              (let ((byte (ldb (byte 8 bytepos) integer)))
-                (incf bytepos 8)
-                (setf acc (logior acc (ash (logand byte #x7f) shift)))
-                (if (logtest byte #x80)
-                    (incf shift 7)
-                    (return acc)))))))
-    (let ((v1 (unpack-1))
-          (v2 (unpack-1)))
-      (values (if (eql v1 0) nil (1- v1))
-              (if (eql v2 0) nil (1- v2))))))
-
-(defun compiled-debug-info-tlf-number (cdi)
-    (nth-value 0 (unpack-tlf-num+offset (compiled-debug-info-tlf-num+offset cdi))))
-  (defun compiled-debug-info-char-offset (cdi)
-    (nth-value 1 (unpack-tlf-num+offset (compiled-debug-info-tlf-num+offset cdi))))
