@@ -90,8 +90,6 @@
 ;;; that's what I'm going with.
 ;;;
 (defun %instance-sxhash (instance)
-  ;; to avoid consing in fmix
-  (declare (inline #+64-bit murmur3-fmix64 #-64-bit murmur3-fmix32))
   ;; LAYOUT must not acquire an extra slot for the stable hash,
   ;; because the bitmap length is derived from the instance length.
   ;; It would probably be simple to eliminate this as a special case
@@ -113,8 +111,7 @@
                    #+sb-thread (%primitive sb-vm::set-instance-hashed instance))
                  (get-lisp-obj-address instance))))
     ;; perturb the address
-    (logand (#+64-bit murmur3-fmix64 #-64-bit murmur3-fmix32 addr)
-            most-positive-fixnum)))
+    (logand (murmur3-fmix-word addr) most-positive-fixnum)))
 
 (declaim (inline instance-sxhash))
 (defun instance-sxhash (instance)
@@ -280,6 +277,11 @@
 ;;; To avoid "note: Return type not fixed values ..."
 (declaim (ftype (sfunction (t) hash-code) pathname-sxhash))
 
+(defun sap-hash (x)
+  ;; toss in a LOGNOT so that (the word a) and (int-sap a) hash differently
+  (let ((w (logand (lognot (sap-int x)) most-positive-word)))
+    (logand (murmur3-fmix-word w) most-positive-fixnum)))
+
 (defun sxhash (x)
   ;; profiling SXHASH is hard, but we might as well try to make it go
   ;; fast, in case it is the bottleneck somewhere.  -- CSR, 2003-03-14
@@ -297,8 +299,7 @@
   ;;    so we should pick off SYMBOL sooner than INSTANCE as well.
   ;;  * INSTANCE (except for PATHNAME) doesn't recurse anyway - in fact
   ;;    it is particularly dumb (by design), so performing that test later
-  ;;    doesn't incur much of a penalty. And our users probably know that
-  ;;    SXHASH on instance doesn't really do anything.
+  ;;    doesn't incur much of a penalty.
   ;; Anyway, afaiu, the code below was previously ordered by gut feeling
   ;; rather than than actual measurement, so having any rationale for ordering
   ;; is better than having no rationale. And as a further comment observes,
@@ -348,6 +349,7 @@
                     (fsc-instance-hash x)
                     ;; funcallable structure, not funcallable-standard-object
                     9550684))
+               (system-area-pointer (sap-hash x))
                (t 42))))
     (sxhash-recurse x +max-hash-depthoid+)))
 
