@@ -210,7 +210,7 @@
 (defconstant trampoline-entry-offset n-word-bytes)
 (defun make-simplifying-trampoline (fun)
   (let ((code (truly-the (values code-component &optional)
-                         (allocate-code-object :dynamic 0 3 24)))) ; KLUDGE
+                         (allocate-code-object :dynamic 3 24)))) ; KLUDGE
     (setf (%code-debug-info code) fun)
     (let ((sap (sap+ (code-instructions code) trampoline-entry-offset))
           (ea (+ (logandc2 (get-lisp-obj-address code) lowtag-mask)
@@ -481,7 +481,7 @@
                       (cons (code-header-ref code wordindex) locs)))))))))
   code)
 
-(sb-c::when-vop-existsp (:translate sb-c::word+)
+(sb-c::when-vop-existsp (:translate sb-c::unsigned+)
   (defconstant cf-bit 0)
   (defconstant sf-bit 7)
   (defconstant of-bit 11)
@@ -490,3 +490,6 @@
     (let ((flags (context-flags context)))
       (values (logbitp of-bit flags)
               (logbitp cf-bit flags)))))
+
+(def-cpu-feature :avx2
+    (plusp (sb-alien:extern-alien "avx2_supported" int)))
