@@ -169,13 +169,15 @@
 (compile 'rename-file-a-la-unix)
 
 (export '(*target-sbcl-version* *generated-sources-root*
+          *build-dependent-generated-sources-root*
           stem-source-path find-bootstrap-file read-from-file))
 (defvar *sources-root* "")
 (defvar *generated-sources-root* "")
+(defvar *build-dependent-generated-sources-root* "")
 (defvar *src-cold-shared-pathname* *load-pathname*)
 
 ;;; See remark in COMPILE-STEM about strings vs. The Common Lisp Way
-(defun find-bootstrap-file (namestring)
+(defun find-bootstrap-file (namestring &optional build-dependent)
   (cond ((char= (char namestring 0) #\^)
          ;; If it starts with a "^" then it means "src/cold/..."
          (let ((this *src-cold-shared-pathname*)
@@ -190,7 +192,9 @@
          ;; in the tree as checked in, or generated by a prior build step.
          (concatenate 'string
                       (if (eql (mismatch "output/" namestring) 7) ; a generated source
-                          *generated-sources-root*
+                          (if build-dependent
+                              *build-dependent-generated-sources-root*
+                              *generated-sources-root*)
                           *sources-root*)
                       namestring))
         (t
@@ -201,9 +205,10 @@
 
 ;;; Return an expression read from the file named NAMESTRING.
 ;;; For user-supplied inputs, protect against more than one expression
-;;; appearing in the file. For in-tree inputs we needn't bother.
-(defun read-from-file (namestring &optional (enforce-single-expr t))
-  (with-open-file (s (find-bootstrap-file namestring))
+;;; appearing in the file.  (Our ^build-order.lisp-expr file has more than
+;;; one expression in it, so we need to be able to not enforce.)
+(defun read-from-file (namestring &key (enforce-single-expr t) build-dependent)
+  (with-open-file (s (find-bootstrap-file namestring build-dependent))
     (let* ((result (read s))
            (eof-result (cons nil nil)))
       (unless enforce-single-expr
@@ -281,6 +286,9 @@
         ;; Win32 conditionally adds :sb-futex in grovel-features.sh
         (when (target-featurep '(:and :sb-thread (:or :linux :freebsd)))
           (pushnew :sb-futex sb-xc:*features*))
+        (when (target-featurep :immobile-space)
+          (pushnew :compact-instance-header sb-xc:*features*)
+          (pushnew :immobile-code sb-xc:*features*))
         (when (target-featurep :64-bit)
           (push :compact-symbol sb-xc:*features*))
         (when (target-featurep '(:and :sb-thread (:not :win32)))
@@ -341,7 +349,7 @@
          ("(not (or elf mach-o win32))"
           "No execute object file format feature defined")
          ("(and cons-profiling (not sb-thread))" ":CONS-PROFILING requires :SB-THREAD")
-         ("(and sb-linkable-runtime (not (or arm arm64 x86 x86-64)))"
+         ("(and sb-linkable-runtime (not (or arm arm64 x86 x86-64 ppc ppc64)))"
           ":SB-LINKABLE-RUNTIME not supported on selected architecture")
          ("(and sb-linkable-runtime (not (or darwin freebsd linux win32)))"
           ":SB-LINKABLE-RUNTIME not supported on selected operating system")
@@ -481,8 +489,7 @@
                               (pathname-type (compile-file-pathname stem)))))
         (:target-compile
          (values *target-obj-prefix*
-                 (cond ((find :extra-artifact flags) "")
-                       ((find :assem flags) *target-assem-obj-suffix*)
+                 (cond ((find :assem flags) *target-assem-obj-suffix*)
                        (t *target-obj-suffix*)))))
     (concatenate 'string obj-prefix (stem-remap-target stem) obj-suffix)))
 (compile 'stem-object-path)
@@ -503,7 +510,7 @@
           ;; to produce warnings as a bug workaround.
           (let ((cl:*features* (cons feature cl:*features*))
                 (*readtable* *xc-readtable*))
-            (read-from-file "^build-order.lisp-expr" nil))))
+            (read-from-file "^build-order.lisp-expr" :enforce-single-expr nil))))
     (setf *stems-and-flags* (cons build-phase list)))
   ;; Now check for duplicate stems and bogus flags.
   (let ((stems (make-hash-table :test 'equal)))
