@@ -441,10 +441,10 @@
               ;; Only the first 6 bits count for shifts.
               ;; This sets all bits to 1 if AMOUNT is larger than 63,
               ;; cutting the amount to 63.
-              (inst csinv temp temp zr-tn :lt)
+              (inst csinv temp temp zr-tn :lo)
               (inst asr result number temp))
              (:unsigned
-              (inst csel result number zr-tn :lt)
+              (inst csel result number zr-tn :lo)
               (inst lsr result result temp)))))
 
     (inst b END)
@@ -454,7 +454,7 @@
            (inst lsl result number temp))
           (t
            (inst cmp temp n-word-bits)
-           (inst csel result number zr-tn :lt)
+           (inst csel result number zr-tn :lo)
            (inst lsl result result temp)))
     END))
 
@@ -485,10 +485,10 @@
               ;; Only the first 6 bits count for shifts.
               ;; This sets all bits to 1 if AMOUNT is larger than 63,
               ;; cutting the amount to 63.
-              (inst csinv temp temp zr-tn :lt)
+              (inst csinv temp temp zr-tn :lo)
               (inst asr temp-result number temp))
              (unsigned-reg
-              (inst csel temp-result number zr-tn :lt)
+              (inst csel temp-result number zr-tn :lo)
               (inst lsr temp-result temp-result temp)))))
 
     (inst b END)
@@ -498,7 +498,7 @@
            (inst lsl temp-result number temp))
           (t
            (inst cmp temp n-word-bits)
-           (inst csel temp-result number zr-tn :lt)
+           (inst csel temp-result number zr-tn :lo)
            (inst lsl temp-result temp-result temp)))
     END
     (inst lsl result temp-result n-fixnum-tag-bits)))
@@ -541,10 +541,10 @@
                                                 (specifier-type `(mod ,n-word-bits)))))
                            (inst cmp amount n-word-bits)
                            (cond ((location= amount result)
-                                  (inst csel tmp-tn number zr-tn :lt)
+                                  (inst csel tmp-tn number zr-tn :lo)
                                   (inst lsl result tmp-tn amount))
                                  (t
-                                  (inst csel result number zr-tn :lt)
+                                  (inst csel result number zr-tn :lo)
                                   (inst lsl result result amount))))
                           (t
                            (inst lsl result number amount)))))
@@ -980,6 +980,60 @@
 (define-conditional-vop > :gt :hi)
 (define-conditional-vop eql :eq :eq)
 
+(define-vop (<-unsigned-signed)
+  (:translate <)
+  (:args (unsigned :scs (unsigned-reg))
+         (signed :scs (signed-reg)))
+  (:arg-types unsigned-num signed-num)
+  (:conditional :lo)
+  (:policy :fast-safe)
+  (:generator 7
+    (inst cmp signed 0)
+    (inst ccmp unsigned signed :ge #b0010)))
+
+(define-vop (>-unsigned-signed <-unsigned-signed)
+  (:translate >)
+  (:conditional :hi))
+
+(define-vop (<-signed-unsigned)
+  (:translate <)
+  (:args (signed :scs (signed-reg))
+         (unsigned :scs (unsigned-reg)))
+  (:arg-types signed-num unsigned-num)
+  (:conditional :lo)
+  (:policy :fast-safe)
+  (:generator 7
+    (inst cmp signed 0)
+    (inst ccmp signed unsigned :ge #b0000)))
+
+(define-vop (>-signed-unsigned <-signed-unsigned)
+  (:translate >)
+  (:conditional :hi))
+
+(define-vop (eql-unsigned-signed)
+  (:translate eql)
+  (:args (unsigned :scs (unsigned-reg))
+         (signed :scs (signed-reg)))
+  (:arg-types unsigned-num signed-num)
+  (:conditional :eq)
+  (:policy :fast-safe)
+  (:generator 7
+    (inst cmp signed 0)
+    (inst ccmp unsigned signed :ge #b0000)))
+
+(define-vop (eql-signed-unsigned eql-unsigned-signed)
+  (:args (signed :scs (signed-reg))
+         (unsigned :scs (unsigned-reg)))
+  (:arg-types signed-num unsigned-num))
+
+(define-vop (eq-unsigned-signed eql-unsigned-signed)
+  (:variant-cost 6)
+  (:translate eq))
+
+(define-vop (eq-signed-unsigned eql-signed-unsigned)
+  (:variant-cost 6)
+  (:translate eq))
+
 (define-vop (generic-eql/fixnum fast-if-eql/fixnum)
   (:args (x :scs (any-reg descriptor-reg))
          (y :scs (any-reg)))
@@ -1125,6 +1179,17 @@
 (define-full-setter bignum-set * bignum-digits-offset other-pointer-lowtag
   (unsigned-reg) unsigned-num sb-bignum:%bignum-set)
 
+(define-partial-reffer sb-c::%half-bignum-ref bignum
+  :word nil bignum-digits-offset other-pointer-lowtag (unsigned-reg signed-reg)
+  positive-fixnum
+  sb-bignum:%half-bignum-ref)
+
+(define-partial-setter sb-c::%half-bignum-set bignum
+  :word bignum-digits-offset other-pointer-lowtag (unsigned-reg signed-reg)
+  positive-fixnum
+  (setf sb-bignum:%half-bignum-ref))
+
+
 (define-vop (digit-0-or-plus)
   (:translate sb-bignum:%digit-0-or-plusp)
   (:policy :fast-safe)
@@ -1133,8 +1198,8 @@
   (:conditional)
   (:info target not-p)
   (:generator 2
-     (inst cmp digit 0)
-     (inst b (if not-p :lt :ge) target)))
+    (inst cmp digit 0)
+    (inst b (if not-p :lt :ge) target)))
 
 (define-vop (add-w/carry)
   (:translate sb-bignum:%add-with-carry)
@@ -1168,7 +1233,7 @@
     (inst cmp c 1) ;; Set carry if (fixnum 0 or 1) c=0, else clear.
     (inst sbcs result a b)
     (unless (eq (tn-kind borrow) :unused)
-     (inst cset borrow :cs))))
+      (inst cset borrow :cs))))
 
 (define-vop (bignum-mult-and-add-3-arg)
   (:translate sb-bignum:%multiply-and-add)
@@ -1279,6 +1344,18 @@
         (unless (= i 64)
           (inst adc rem rem rem))))))
 
+(define-vop (half-bignum-floor bignum-floor)
+  (:translate sb-bignum:%half-bigfloor)
+  (:args (div-high :scs (unsigned-reg) :to :save)
+         (div-low :scs (unsigned-reg) :target x)
+         (divisor :scs (unsigned-reg) :to (:result 1)))
+  (:temporary (:sc unsigned-reg :from (:argument 1)) x)
+  (:generator 30
+    (move x div-low)
+    (inst bfm x div-high 32 31)
+    (inst udiv quo x divisor)
+    (inst msub rem quo divisor x)))
+
 (define-vop (signify-digit)
   (:translate sb-bignum:%fixnum-digit-with-correct-sign)
   (:policy :fast-safe)
@@ -1341,7 +1418,7 @@
       (inst b :ne error))))
 
 (define-vop ()
-  (:translate sb-c::signed-word*)
+  (:translate sb-c::signed*)
   (:args (x :scs (signed-reg))
          (y :scs (signed-reg immediate)))
   (:arg-types signed-num signed-num (:constant t))
@@ -1369,7 +1446,7 @@
       (inst b :ne error))))
 
 (define-vop ()
-  (:translate sb-c::word*)
+  (:translate sb-c::unsigned*)
   (:args (x :scs (unsigned-reg))
          (y :scs (unsigned-reg immediate)))
   (:arg-types unsigned-num unsigned-num (:constant t))
@@ -1406,8 +1483,8 @@
          (y :scs (signed-reg immediate)))
   (:arg-types signed-num signed-num)
   (:temporary (:sc signed-reg) high low)
-  (:temporary (:sc signed-reg :from (:argument 2)) header pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc signed-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (r :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:vop-var vop)
@@ -1422,15 +1499,17 @@
           (inst lsl low x (1- (integer-length value)))
           (inst mul low x y))
       (inst smulh high x y))
-    (inst mov header (compute-object-header (+ 2 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 2))
     (inst cmp high (asr low 63))
     (inst b :ne allocate)
     (inst adds r low low)
     (inst b :vc done)
-    (inst mov header (compute-object-header (+ 1 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 1))
+    #+bignum-assertions
+    (inst mov high 0)
     allocate
     (with-fixed-allocation
-        (r pa-flag nil (+ 2 bignum-digits-offset) :lip lip)
+        (r lr nil (+ 2 bignum-digits-offset))
       (storew-pair header 0 low bignum-digits-offset tmp-tn)
       (storew high tmp-tn 2))
     DONE))
@@ -1441,8 +1520,8 @@
          (y :scs (unsigned-reg immediate)))
   (:arg-types unsigned-num unsigned-num)
   (:temporary (:sc unsigned-reg) high low)
-  (:temporary (:sc unsigned-reg :from (:argument 2)) header pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc unsigned-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (r :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:vop-var vop)
@@ -1461,17 +1540,17 @@
                  (setf y high))
                (inst mul low x y)
                (inst umulh high x y))))
-    (inst mov header (compute-object-header (+ 3 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 3))
     (inst tbnz high 63 allocate)
-    (inst mov header (compute-object-header (+ 2 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 2))
     (inst cbnz high allocate)
     (inst tbnz low 63 allocate)
     (inst adds r low low)
     (inst b :vc done)
-    (inst mov header (compute-object-header (+ 1 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 1))
     allocate
     (with-fixed-allocation
-        (r pa-flag nil (+ 2 bignum-digits-offset) :lip lip)
+        (r lr nil (+ 2 bignum-digits-offset))
       (storew-pair header 0 low bignum-digits-offset tmp-tn)
       (storew high tmp-tn 2))
     DONE))
@@ -1482,8 +1561,8 @@
          (y :scs (signed-reg immediate)))
   (:arg-types signed-num signed-num)
   (:temporary (:sc signed-reg) high low)
-  (:temporary (:sc signed-reg :from (:argument 2)) header pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc signed-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (r :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:vop-var vop)
@@ -1495,14 +1574,14 @@
               (inst adds low x (add-sub-immediate y high))))
         (inst adds low x y))
     (inst csetm high :cs)
-    (inst mov header (compute-object-header (+ 2 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 2))
     (inst b :vs allocate)
     (inst adds r low low)
     (inst b :vc done)
-    (inst mov header (compute-object-header (+ 1 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 1))
     allocate
     (with-fixed-allocation
-        (r pa-flag nil (+ 2 bignum-digits-offset) :lip lip)
+        (r lr nil (+ 2 bignum-digits-offset))
       (storew-pair header 0 low bignum-digits-offset tmp-tn)
       (storew high tmp-tn 2))
     DONE))
@@ -1513,8 +1592,8 @@
          (y :scs (signed-reg immediate)))
   (:arg-types signed-num signed-num)
   (:temporary (:sc signed-reg) high low)
-  (:temporary (:sc signed-reg :from (:argument 2)) header pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc signed-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (r :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:vop-var vop)
@@ -1526,14 +1605,14 @@
               (inst subs low x (add-sub-immediate y high))))
         (inst subs low x y))
     (inst csetm high :cs)
-    (inst mov header (compute-object-header (+ 2 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 2))
     (inst b :vs allocate)
     (inst adds r low low)
     (inst b :vc done)
-    (inst mov header (compute-object-header (+ 1 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 1))
     allocate
     (with-fixed-allocation
-        (r pa-flag nil (+ 2 bignum-digits-offset) :lip lip)
+        (r lr nil (+ 2 bignum-digits-offset))
       (storew-pair header 0 low bignum-digits-offset tmp-tn)
       (storew high tmp-tn 2))
     DONE))
@@ -1544,8 +1623,8 @@
          (y :scs (unsigned-reg immediate)))
   (:arg-types unsigned-num unsigned-num)
   (:temporary (:sc unsigned-reg) high low)
-  (:temporary (:sc unsigned-reg :from (:argument 2)) header pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc unsigned-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (r :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:vop-var vop)
@@ -1553,16 +1632,16 @@
     (inst adds low x (if (sc-is y immediate)
                          (add-sub-immediate (tn-value y) high)
                          y))
-    (inst mov header (compute-object-header (+ 2 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 2))
     (inst cset high :cs)
     (inst b :cs allocate)
     (inst b :mi allocate)
     (inst adds r low low)
     (inst b :vc done)
-    (inst mov header (compute-object-header (+ 1 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 1))
     allocate
     (with-fixed-allocation
-        (r pa-flag nil (+ 2 bignum-digits-offset) :lip lip)
+        (r lr nil (+ 2 bignum-digits-offset))
       (storew-pair header 0 low bignum-digits-offset tmp-tn)
       (storew high tmp-tn 2))
     DONE))
@@ -1573,8 +1652,8 @@
          (y :scs (unsigned-reg immediate)))
   (:arg-types unsigned-num unsigned-num)
   (:temporary (:sc unsigned-reg) high low)
-  (:temporary (:sc unsigned-reg :from (:argument 2)) header pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc unsigned-reg :from (:argument 2)) header)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (r :scs (descriptor-reg)))
   (:policy :fast-safe)
   (:vop-var vop)
@@ -1582,7 +1661,7 @@
     (inst subs low x (if (sc-is y immediate)
                          (add-sub-immediate (tn-value y) high)
                          y))
-    (inst mov header (compute-object-header (+ 2 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 2))
     (inst csetm high :cc)
     (inst b :cc negative)
     (inst b :mi allocate)
@@ -1592,16 +1671,16 @@
     positive
     (inst adds r low low)
     (inst b :vc done)
-    (inst mov header (compute-object-header (+ 1 bignum-digits-offset) bignum-widetag))
+    (inst mov header (bignum-header-for-length 1))
     allocate
     (with-fixed-allocation
-        (r pa-flag nil (+ 2 bignum-digits-offset) :lip lip)
+        (r lr nil (+ 2 bignum-digits-offset))
       (storew-pair header 0 low bignum-digits-offset tmp-tn)
       (storew high tmp-tn 2))
     DONE))
 
 (define-vop ()
-  (:translate sb-c::signed-word+)
+  (:translate sb-c::signed+)
   (:args (x :scs (signed-reg))
          (y :scs (signed-reg immediate)))
   (:arg-types signed-num signed-num (:constant t))
@@ -1623,7 +1702,7 @@
       (inst b :vs error))))
 
 (define-vop ()
-  (:translate sb-c::signed-word-)
+  (:translate sb-c::signed-)
   (:args (x :scs (signed-reg))
          (y :scs (signed-reg immediate)))
   (:arg-types signed-num signed-num (:constant t))
@@ -1645,7 +1724,7 @@
       (inst b :vs error))))
 
 (define-vop ()
-  (:translate sb-c::word+)
+  (:translate sb-c::unsigned+)
   (:args (x :scs (unsigned-reg))
          (y :scs (unsigned-reg immediate)))
   (:arg-types unsigned-num unsigned-num (:constant t))
@@ -1664,7 +1743,7 @@
       (inst b :cs error))))
 
 (define-vop ()
-  (:translate sb-c::word-)
+  (:translate sb-c::unsigned-)
   (:args (x :scs (unsigned-reg))
          (y :scs (unsigned-reg immediate)))
   (:arg-types unsigned-num unsigned-num (:constant t))
@@ -1681,3 +1760,79 @@
         (setf y (add-sub-immediate (tn-value y))))
       (inst subs r x y)
       (inst b :cc error))))
+
+(define-vop ()
+  (:translate sb-c::unsigned+signed)
+  (:args (x :scs (unsigned-reg) :to :result)
+         (y :scs (signed-reg immediate)))
+  (:arg-types unsigned-num signed-num (:constant t))
+  (:info type)
+  (:results (r :scs (unsigned-reg)))
+  (:temporary (:scs (unsigned-reg) :from (:argument 2)) temp)
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (if (sc-is y immediate)
+          (assemble ()
+            (let ((value (tn-value y)))
+              (setf y (add-sub-immediate (abs value)))
+              (cond ((plusp value)
+                     (inst adds r x y)
+                     (inst b :cs error))
+                    (t
+                     (inst subs r x y)
+                     (inst b :cc error)))))
+          (assemble ()
+            (when (sc-is y immediate)
+              (setf y (add-sub-immediate (tn-value y))))
+            (inst tbnz y 63 neg)
+            (inst adds r x y)
+            (inst b :cs error)
+            (inst b done)
+            neg
+            (inst neg temp y)
+            (inst subs r x temp)
+            (inst b :cc error)
+            done)))))
+
+(define-vop ()
+  (:translate sb-c::unsigned-signed)
+  (:args (x :scs (unsigned-reg) :to :result)
+         (y :scs (signed-reg immediate)))
+  (:arg-types unsigned-num signed-num (:constant t))
+  (:info type)
+  (:results (r :scs (unsigned-reg)))
+  (:temporary (:scs (unsigned-reg) :from (:argument 2)) temp)
+  (:result-types unsigned-num)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:generator 2
+    (let* ((*location-context* (unless (eq type 'fixnum)
+                                 type))
+           (error (generate-error-code vop 'sb-kernel::add-sub-overflow-error r)))
+      (if (sc-is y immediate)
+          (assemble ()
+            (let ((value (tn-value y)))
+              (setf y (add-sub-immediate (abs value)))
+              (cond ((minusp value)
+                     (inst adds r x y)
+                     (inst b :cs error))
+                    (t
+                     (inst subs r x y)
+                     (inst b :cc error)))))
+          (assemble ()
+            (when (sc-is y immediate)
+              (setf y (add-sub-immediate (tn-value y))))
+            (inst tbz y 63 pos)
+            (inst neg temp y)
+            (inst adds r x temp)
+            (inst b :cs error)
+            (inst b done)
+            pos
+            (inst subs r x y)
+            (inst b :cc error)
+            done)))))
