@@ -237,13 +237,81 @@
                 "S")
             value)))
 
+(defun decode-vector-size (q size)
+  (case q
+    (0
+     (case size
+       (#b00 "8B")
+       (#b01 "4H")
+       (#b10 "2S")))
+    (1
+     (case size
+       (#b00 "16B")
+       (#b01 "8H")
+       (#b10 "4S")
+       (#b11 "2D")))))
+
 (defun print-simd-reg (value stream dstate)
   (declare (ignore dstate))
-  (destructuring-bind (size offset) value
+  (multiple-value-bind (q size offset)
+      (if (= (length value) 3)
+          (destructuring-bind (q size offset) value
+            (values q size offset))
+          (destructuring-bind (q offset) value
+            (values q 0 offset)))
     (format stream "V~d.~a" offset
-            (if (zerop size)
-                "8B"
-                "16B"))))
+            (decode-vector-size q size))))
+
+(defun print-simd-immh-reg (value stream dstate)
+  (declare (ignore dstate))
+  (if (= (length value) 2)
+      (destructuring-bind (immh offset) value
+        (format stream "V~d.~a" offset
+                (cond ((logbitp 0 immh)
+                       "8H")
+                      ((logbitp 1 immh)
+                       "4S")
+                      ((logbitp 2 immh)
+                       "2D"))))
+      (destructuring-bind (q immh offset) value
+        (format stream "V~d.~a" offset
+                (cond ((logbitp 0 immh)
+                       (if (zerop q)
+                           "8B"
+                           "16B"))
+                      ((logbitp 1 immh)
+                       (if (zerop q)
+                           "4H"
+                           "8H"))
+                      ((logbitp 2 immh)
+                       (if (zerop q)
+                           "2S"
+                           "4S")))))))
+
+(defun print-simd-reg-cmode (value stream dstate)
+  (declare (ignore dstate))
+  (destructuring-bind (q cmode offset) value
+    (format stream "V~d.~a" offset
+            (cond ((eq cmode #b1110)
+                   (if (zerop q)
+                       "8B"
+                       "16B"))
+                  ((zerop (logand cmode #b1001))
+                   (if (zerop q)
+                       "2S"
+                       "4S"))))))
+
+(defun print-simd-modified-imm (value stream dstate)
+  (declare (ignore dstate))
+  (destructuring-bind (abc cmode defgh) value
+    (let ((shift
+            (cond ((eq cmode #b1110)
+                   0)
+                  ((zerop (logand cmode #b1001))
+                   (ash cmode 2)))))
+      (princ (dpb abc (byte 3 5) defgh) stream)
+      (when (plusp shift)
+        (format t ", LSL #~d" shift)))))
 
 (defun print-vbhs (value stream dstate)
   (declare (ignore dstate))
@@ -307,6 +375,9 @@
   (declare (ignore dstate))
   (princ (svref +condition-name-vec+ value) stream))
 
+(defun print-negated-cond (value stream dstate)
+  (print-cond (logxor 1 value) stream dstate))
+
 (defun use-label (value dstate)
   (let* ((value (if (consp value)
                     (logior (ldb (byte 2 0) (car value))
@@ -397,7 +468,8 @@
               (slot2 (find (1+ offset) thread-slots :key #'slot-offset)))
          (when slot1
            (note (lambda (stream)
-                   (if (memq (slot-name slot1) '(sb-vm::boxed-tlab
+                   (if (memq (slot-name slot1) '(sb-vm::mixed-tlab
+                                                 sb-vm::cons-tlab
                                                  sb-vm::unboxed-tlab))
                        (format stream "~(~a~).{free-pointer, end-addr}" (slot-name slot1))
                        (format stream "~(~A, ~A~)" (slot-name slot1) (slot-name slot2))))
@@ -430,6 +502,8 @@
                           (prog1 (sap-ref-8 sap offset)
                             (incf offset)))))
          (first-arg (ldb (byte 8 13) inst))
+         (first-offset (ldb (byte 5 0) first-arg))
+         (first-sc (ldb (byte 2 5) first-arg))
          (length (sb-kernel::error-length error-number))
          (index offset))
     (declare (type sb-sys:system-area-pointer sap)
@@ -443,8 +517,11 @@
           (t
            (collect ((sc+offsets)
                      (lengths))
-             (unless (= first-arg sb-vm::zr-offset)
-               (sc+offsets (make-sc+offset sb-vm:descriptor-reg-sc-number first-arg))
+             (unless (= first-offset sb-vm::zr-offset)
+               (sc+offsets (make-sc+offset (case first-sc
+                                             (1 sb-vm:unsigned-reg-sc-number)
+                                             (2 sb-vm:signed-reg-sc-number)
+                                             (t sb-vm:descriptor-reg-sc-number)) first-offset))
                (lengths 0))
              (loop repeat length do
                    (let ((old-index index))
