@@ -19,7 +19,7 @@
     ;; Note:
     ;; This is not FLUSHABLE because it's defined to signal errors.
     (movable)
-  ;; :DERIVE-TYPE RESULT-TYPE-SPEC-NTH-ARG 2 ? Nope... (COERCE 1 'COMPLEX)
+  ;; :DERIVE-TYPE RESULT-TYPE-SPEC-NTH-ARG 1 ? Nope... (COERCE 1 'COMPLEX)
   ;; returns REAL/INTEGER, not COMPLEX.
   )
 ;; These each check their input sequence for type-correctness,
@@ -89,7 +89,7 @@
 
 #+(or x86 x86-64 arm arm64)
 (defknown fixnum-mod-p (t fixnum) boolean
-  (movable foldable flushable always-translatable))
+  (movable flushable always-translatable))
 
 
 ;;;; classes
@@ -111,8 +111,8 @@
 
 ;;; This is not FLUSHABLE, since it's required to signal an error if
 ;;; unbound.
-(defknown (symbol-value symeval) (symbol) t ()
-  :derive-type #'symeval-derive-type)
+(defknown symbol-value (symbol) t ()
+  :derive-type #'symbol-value-derive-type)
 (defknown about-to-modify-symbol-value (symbol t &optional t t) null
   ())
 ;;; From CLHS, "If the symbol is globally defined as a macro or a
@@ -517,15 +517,15 @@
   (flushable))
 
 (defknown copy-seq (proper-sequence) consed-sequence (flushable)
-  :derive-type (sequence-result-nth-arg 1 :preserve-dimensions t))
+  :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t))
 
 (defknown length (proper-sequence) index (foldable flushable dx-safe))
 
 (defknown reverse (proper-sequence) consed-sequence (flushable)
-  :derive-type (sequence-result-nth-arg 1 :preserve-dimensions t))
+  :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t))
 
 (defknown nreverse ((modifying sequence)) sequence (important-result)
-  :derive-type (sequence-result-nth-arg 1 :preserve-dimensions t
+  :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t
                                           :preserve-vector-type t))
 
 (defknown make-sequence (type-specifier index
@@ -533,10 +533,10 @@
                                         (:initial-element t))
   consed-sequence
   (movable)
-  :derive-type (creation-result-type-specifier-nth-arg 1))
+  :derive-type (creation-result-type-specifier-nth-arg 0))
 
 (defknown concatenate (type-specifier &rest proper-sequence) consed-sequence ()
-  :derive-type (creation-result-type-specifier-nth-arg 1))
+  :derive-type (creation-result-type-specifier-nth-arg 0))
 
 (defknown %concatenate-to-string (&rest sequence) simple-string
   (flushable no-verify-arg-count))
@@ -633,7 +633,7 @@
      (:key (function-designator ((nth-arg 1 :sequence t)))))
   consed-sequence
   (flushable call)
-  :derive-type (sequence-result-nth-arg 2))
+  :derive-type (sequence-result-nth-arg 1))
 
 (defknown substitute
   (t t proper-sequence &rest t &key (:from-end t)
@@ -645,7 +645,7 @@
      (:key (function-designator ((nth-arg 2 :sequence t)))))
   consed-sequence
   (flushable call)
-  :derive-type (sequence-result-nth-arg 3))
+  :derive-type (sequence-result-nth-arg 2))
 
 (defknown (remove-if remove-if-not)
   ((function-designator ((nth-arg 1 :sequence t :key :key))) proper-sequence
@@ -656,7 +656,7 @@
    (:key (function-designator ((nth-arg 1 :sequence t)))))
   consed-sequence
   (flushable call)
-  :derive-type (sequence-result-nth-arg 2))
+  :derive-type (sequence-result-nth-arg 1))
 
 (defknown (substitute-if substitute-if-not)
   (t (function-designator ((nth-arg 2 :sequence t :key :key))) proper-sequence
@@ -667,7 +667,7 @@
      (:key (function-designator ((nth-arg 2 :sequence t)))))
   consed-sequence
   (flushable call)
-  :derive-type (sequence-result-nth-arg 3))
+  :derive-type (sequence-result-nth-arg 2))
 
 (defknown delete
   (t (modifying sequence) &rest t &key (:from-end t)
@@ -678,7 +678,7 @@
      (:key (function-designator ((nth-arg 1 :sequence t)))))
   sequence
   (call important-result)
-  :derive-type (sequence-result-nth-arg 2))
+  :derive-type (sequence-result-nth-arg 1))
 
 (defknown nsubstitute
   (t t (modifying sequence) &rest t &key (:from-end t)
@@ -689,7 +689,7 @@
      (:key (function-designator ((nth-arg 2 :sequence t)))))
   sequence
   (call)
-  :derive-type (sequence-result-nth-arg 3))
+  :derive-type (sequence-result-nth-arg 2))
 
 (defknown (delete-if delete-if-not)
   ((function-designator ((nth-arg 1 :sequence t :key :key))) (modifying sequence)
@@ -698,7 +698,7 @@
    (:key (function-designator ((nth-arg 1 :sequence t)))))
   sequence
   (call important-result)
-  :derive-type (sequence-result-nth-arg 2))
+  :derive-type (sequence-result-nth-arg 1))
 
 (defknown (nsubstitute-if nsubstitute-if-not)
   (t (function-designator ((nth-arg 2 :sequence t :key :key))) (modifying sequence)
@@ -707,7 +707,7 @@
      (:key (function-designator ((nth-arg 2 :sequence t)))))
   sequence
   (call)
-  :derive-type (sequence-result-nth-arg 3))
+  :derive-type (sequence-result-nth-arg 2))
 
 (defknown remove-duplicates
   (proper-sequence &rest t &key
@@ -721,7 +721,7 @@
             (:key (function-designator ((nth-arg 0 :sequence t)))))
   consed-sequence
   (flushable call)
-  :derive-type (sequence-result-nth-arg 1))
+  :derive-type (sequence-result-nth-arg 0))
 
 (defknown delete-duplicates
   ((modifying sequence)
@@ -735,7 +735,7 @@
    (:key (function-designator ((nth-arg 0 :sequence t)))))
   sequence
   (call important-result)
-  :derive-type (sequence-result-nth-arg 1))
+  :derive-type (sequence-result-nth-arg 0))
 
 (defknown find
   (t proper-sequence &rest t &key
@@ -857,7 +857,7 @@
                                                      (nth-arg 2 :sequence t))))))
   sequence
   (call important-result)
-  :derive-type (creation-result-type-specifier-nth-arg 1))
+  :derive-type (creation-result-type-specifier-nth-arg 0))
 
 (defknown read-sequence ((modifying sequence) stream
                          &key
@@ -1260,7 +1260,8 @@
   (string-designator &key
    (:start (inhibit-flushing index 0))
    (:end (inhibit-flushing sequence-end nil)))
-  simple-string (flushable))
+  simple-string (flushable)
+  :derive-type (sequence-result-nth-arg 0 :preserve-dimensions t))
 
 (defknown (nstring-upcase nstring-downcase nstring-capitalize)
   ((modifying string) &key (:start index) (:end sequence-end))
@@ -1715,17 +1716,6 @@
    (:entry-points list)
    (:emit-cfasl t))
   (values (or pathname null) boolean boolean))
-(defknown sb-c::compile-files
-    (cons &key (:output-file pathname-designator)
-               (:verbose t)
-               (:print t)
-               (:external-format external-format-designator)
-               (:progress t)
-               (:trace-file t)
-               (:block-compile t)
-               (:entry-points list)
-               (:emit-cfasl t))
-  (values (or pathname null) boolean boolean))
 
 (defknown (compile-file-pathname)
   (pathname-designator &key (:output-file (or pathname-designator
@@ -1802,8 +1792,8 @@
 
 ;;;; miscellaneous extensions
 
-(defknown (symbol-global-value sym-global-val) (symbol) t ()
-  :derive-type #'symeval-derive-type)
+(defknown symbol-global-value (symbol) t ()
+  :derive-type #'symbol-value-derive-type)
 (defknown set-symbol-global-value (symbol t) t ()
   :derive-type #'result-type-last-arg)
 
@@ -1899,9 +1889,9 @@
   (always-translatable))
 ;;; The lowest-level vector SET operators should not return a value.
 ;;; Functions built upon them may return the input value.
-(defknown data-vector-set (array index t) (values) (always-translatable))
+(defknown data-vector-set (array index t) (values) (dx-safe always-translatable))
 (defknown data-vector-set-with-offset (array fixnum fixnum t) (values)
-  (always-translatable))
+  (dx-safe always-translatable))
 (defknown hairy-data-vector-ref (array index) t (foldable))
 (defknown hairy-data-vector-set (array index t) t ())
 (defknown hairy-data-vector-ref/check-bounds (array index) t (foldable))
@@ -2033,7 +2023,6 @@
 (defknown %alien-funcall ((or string system-area-pointer) alien-type &rest t) *)
 
 ;; Used by WITH-PINNED-OBJECTS
-#+(or x86 x86-64)
 (defknown sb-vm::touch-object (t) (values)
   (always-translatable))
 
@@ -2195,17 +2184,19 @@
 (defknown fixnum*
   (fixnum fixnum t)
   fixnum
-  (movable foldable unsafely-flushable commutative
-           always-translatable))
+  (movable foldable always-translatable))
 
-(defknown (signed-word* signed-word+ signed-word-)
+(defknown (signed* signed+ signed-)
   (sb-vm:signed-word sb-vm:signed-word t)
   sb-vm:signed-word
-  (movable foldable unsafely-flushable commutative
-           always-translatable))
+  (movable foldable commutative always-translatable))
 
-(defknown (word* word+ word-)
+(defknown (unsigned* unsigned+ unsigned-)
   (word word t)
   word
-  (movable foldable unsafely-flushable commutative
-           always-translatable))
+  (movable foldable commutative always-translatable))
+
+(defknown (unsigned+signed unsigned-signed)
+  (word sb-vm:signed-word t)
+  word
+  (movable foldable commutative always-translatable))
