@@ -72,9 +72,11 @@
   "The most positive integer that is of type SB-EXT:WORD.")
 
 (defconstant maximum-bignum-length
-  ;; Compute number of bits in the maximum length's representation
-  ;; leaving one bit for a GC mark bit.
-  (ldb (byte (- n-word-bits n-widetag-bits 1) 0) -1))
+  ;; 32-bit: leave one bit for a GC mark bit
+  #-64-bit (ldb (byte (- n-word-bits n-widetag-bits 1) 0) -1)
+  ;; 64-bit: restrict to a reasonably large theoretical size of 32GiB per bignum.
+  ;; I haven't exercised our math routines to anywhere near that limit.
+  #+64-bit #xFFFFFFFF)
 
 ;;; The ANSI-specified minimum is 8.
 ;;; Since the array rank is stored as rank-1 in the array header,
@@ -160,6 +162,11 @@
 ;;; which can be expressed in 8 bits.
 (defconstant short-header-max-words #x7fff)
 
+#+gencgc
+(defconstant max-conses-per-page
+  (floor (* gencgc-page-bytes n-byte-bits)
+         (1+ (* n-word-bytes 2 n-byte-bits)))) ; 1 extra bit per cons
+
 ;;; Amount to righ-shift an instance header to get the length.
 ;;; Similar consideration as above with regard to use of generation# byte.
 (defconstant instance-length-shift 10)
@@ -207,33 +214,6 @@
 
 #+sb-xc-host (deftype sb-xc:fixnum () `(signed-byte ,n-fixnum-bits))
 
-#-darwin-jit
-(progn
-  (declaim (inline (setf sap-ref-word-jit)
-                   (setf signed-sap-ref-32-jit)
-                   signed-sap-ref-32-jit
-                   (setf sap-ref-32-jit)
-                   sap-ref-32-jit
-                   (setf sap-ref-8-jit)))
-  (defun (setf sap-ref-word-jit) (value sap offset)
-    (setf (sap-ref-word sap offset) value))
-
-  (defun (setf signed-sap-ref-32-jit) (value sap offset)
-    (setf (signed-sap-ref-32 sap offset) value))
-
-  (defun signed-sap-ref-32-jit (sap offset)
-    (signed-sap-ref-32 sap offset))
-
-  (defun (setf sap-ref-32-jit) (value sap offset)
-    (setf (sap-ref-32 sap offset) value))
-
-  (defun sap-ref-32-jit (sap offset)
-    (sap-ref-32 sap offset))
-
-  #-sb-xc-host
-  (defun (setf sap-ref-8-jit) (value sap offset)
-    (setf (sap-ref-8 sap offset) value)))
-
 ;;; Supporting code for LAYOUT allocated in metadata space a/k/a "metaspace".
 ;;; These objects are manually allocated and freed.
 ;;; Tracts are the unit of allocation requested from the OS.
@@ -285,3 +265,6 @@
   `(setf (slab-sizeclass ,slab) (ash ,sizeclass n-fixnum-tag-bits)
          (slab-capacity ,slab) ,capacity
          (slab-chunk-size ,slab) ,chunksize))
+
+#+gencgc
+(defconstant gencgc-page-words (/ gencgc-page-bytes n-word-bytes))
