@@ -53,23 +53,22 @@
 
 ;;; Point FUN's 'self' slot to FUN.
 ;;; FUN must be pinned when calling this.
-(declaim (inline assign-simple-fun-self))
-(defun assign-simple-fun-self (fun)
-  (let ((self ;; x86 backends store the address of the entrypoint in 'self'
-          #+(or x86 x86-64)
-          (%make-lisp-obj
-           (truly-the word (+ (get-lisp-obj-address fun)
-                              (ash sb-vm:simple-fun-insts-offset sb-vm:word-shift)
-                              (- sb-vm:fun-pointer-lowtag))))
-          ;; non-x86 backends store the function itself (what else?) in 'self'
-          #-(or x86 x86-64) fun))
-    #-darwin-jit
-    (setf (sb-vm::%simple-fun-self fun) self)
-    #+darwin-jit
-    (sb-vm::jit-patch (+ (get-lisp-obj-address fun)
-                  (- sb-vm:fun-pointer-lowtag)
-                  (* sb-vm:simple-fun-self-slot sb-vm:n-word-bytes))
-               (get-lisp-obj-address self))))
+#-darwin-jit ; done entirely by C for #+darwin-jit
+(defmacro assign-simple-fun-self (fun)
+  `(let* ((fun ,fun)
+          (self
+           ;; a few architectures store the untagged address of the entrypoint in 'self'
+           #+(or x86 x86-64 arm64)
+           (%make-lisp-obj
+            (truly-the word (+ (get-lisp-obj-address fun)
+                               (ash sb-vm:simple-fun-insts-offset sb-vm:word-shift)
+                               (- sb-vm:fun-pointer-lowtag))))
+           ;; all others store the function itself (what else?) in 'self'
+           #-(or x86 x86-64 arm64) fun))
+     (setf (sb-vm::%simple-fun-self fun) self)))
+
+(define-load-time-global sb-fasl::*asm-routine-index-to-name* #())
+(declaim (simple-vector sb-fasl::*asm-routine-index-to-name*))
 
 (flet ((fixup (code-obj offset name kind flavor preserved-lists statically-link-p)
          (declare (ignorable statically-link-p))
@@ -129,7 +128,9 @@
              (setf (sb-vm::%code-fixups code-obj)
                    (sb-c:pack-code-fixup-locs abs-fixups rel-fixups gc-barrier-fixups))))
 
-         ;; Assign all SIMPLE-FUN-SELF slots
+         ;; Assign all SIMPLE-FUN-SELF slots unless #+darwin-jit in which case the simple-funs
+         ;; are assigned by jit_memcpy_codeblob()
+         #-darwin-jit
          (dotimes (i (code-n-entries code-obj))
            (let ((fun (%code-entry-point code-obj i)))
              (assign-simple-fun-self fun)
@@ -139,23 +140,31 @@
                                (- 4 sb-vm:fun-pointer-lowtag))
                    (truly-the (unsigned-byte 32)
                      (get-lisp-obj-address (wrapper-friend #.(find-layout 'function)))))))
-         ;; And finally, make the memory range executable
-         #-(or x86 x86-64) (sb-vm:sanctify-for-execution code-obj)
+         ;; And finally, make the memory range executable.
+         ;; x86 doesn't need it, and darwin-jit doesn't do it because the
+         ;; temporary object is not executable.
+         #-(or x86 x86-64 darwin-jit) (sb-vm:sanctify-for-execution code-obj)
          ;; Return fixups amenable to static linking
          (aref preserved-lists 0)))
 
-  (defun apply-fasl-fixups (fop-stack code-obj n-fixups &aux (top (svref fop-stack 0)))
+  (defun apply-fasl-fixups (code-obj fixups index count &aux (end (1- (+ index count))))
     (dx-let ((preserved (make-array 5 :initial-element nil)))
-      (macrolet ((pop-fop-stack () `(prog1 (svref fop-stack top) (decf top))))
-        (binding* ((alloc-points (pop-fop-stack) :exit-if-null))
-          (setf (gethash code-obj *allocation-patch-points*) alloc-points))
-        (dotimes (i n-fixups (setf (svref fop-stack 0) top))
-          (multiple-value-bind (offset kind flavor)
-              (sb-fasl::!unpack-fixup-info (pop-fop-stack))
-            (fixup code-obj offset (pop-fop-stack) kind flavor preserved nil))))
+      (awhen (svref fixups index)
+        (setf (gethash code-obj *allocation-patch-points*) it))
+      (loop
+        (when (>= index end) (return))
+        (binding* (((offset kind flavor data)
+                    (sb-fasl::!unpack-fixup-info (svref fixups (incf index))))
+                   (name
+                    (cond ((member flavor '(:code-object :gc-barrier)) nil)
+                          ((and (plusp data)
+                                (member flavor '(:assembly-routine :assembly-routine*)))
+                           (aref sb-fasl::*asm-routine-index-to-name* data))
+                          (t (svref fixups (incf index))))))
+          (fixup code-obj offset name kind flavor preserved nil)))
       (finish-fixups code-obj preserved)))
 
-  (defun apply-core-fixups (fixup-notes code-obj)
+  (defun apply-core-fixups (code-obj fixup-notes)
     (declare (list fixup-notes))
     (dx-let ((preserved (make-array 5 :initial-element nil)))
       (dolist (note fixup-notes)
@@ -168,7 +177,12 @@
                  preserved t)))
       (finish-fixups code-obj preserved))))
 
-;;; Return a behaviorally identical copy of CODE.
+;;; Return a behaviorally identical copy of CODE which is used for TRACE
+;;; in "funobj encapsulation" mode where we just switch an entry point
+;;; so that it jumps to a tracing routine and then back again.
+;;; The code that gets copied is just the tracing wrapper.
+;;; See the example at COMPILE-FUNOBJ-ENCAPSULATION in ntrace
+#+(or x86 x86-64)
 (defun copy-code-object (code)
   ;; Must have one simple-fun
   (aver (= (code-n-entries code) 1))
@@ -178,16 +192,13 @@
   #+x86-64
   (aver (not (nth-value
               1 (sb-c:unpack-code-fixup-locs (sb-vm::%code-fixups code)))))
+  (aver (zerop (code-n-named-calls code)))
   (let* ((nbytes (code-object-size code))
          (boxed (code-header-words code)) ; word count
          (unboxed (- nbytes (ash boxed sb-vm:word-shift))) ; byte count
-         (copy (allocate-code-object
-                :dynamic (code-n-named-calls code) boxed unboxed)))
+         (copy (allocate-code-object :dynamic boxed unboxed)))
     (with-pinned-objects (code copy)
-      #-darwin-jit
       (%byte-blt (code-instructions code) 0 (code-instructions copy) 0 unboxed)
-      #+darwin-jit
-      (sb-vm::jit-memcpy (code-instructions copy) (code-instructions code) unboxed)
       ;; copy boxed constants so that the fixup step (if needed) sees the 'fixups'
       ;; slot from the new object.
       (loop for i from 2 below boxed
@@ -200,33 +211,6 @@
       (assign-simple-fun-self (%code-entry-point copy 0)))
     copy))
 
-;;; Note the existence of FUNCTION.
-(defun note-fun (info function object)
-  (declare (type function function)
-           (type core-object object))
-  (let ((patch-table (core-object-patch-table object)))
-    (dolist (patch (gethash info patch-table))
-      (setf (code-header-ref (car patch) (the index (cdr patch))) function))
-    (remhash info patch-table))
-  (setf (gethash info (core-object-entry-table object)) function)
-  (values))
-
-;;; Stick a reference to the function FUN in CODE-OBJECT at index I. If the
-;;; function hasn't been compiled yet, make a note in the patch table.
-(defun reference-core-fun (code-obj i fun object)
-  (declare (type core-object object) (type functional fun)
-           (type index i))
-  (let* ((info (leaf-info fun))
-         (found (gethash info (core-object-entry-table object))))
-    ;; A  core component should not have cross-component references.
-    ;; If it could, then the entries would have placed into one component.
-    (aver found)
-    (if found
-        (setf (code-header-ref code-obj i) found)
-        (push (cons code-obj i)
-              (gethash info (core-object-patch-table object)))))
-  (values))
-
 ;;; Dump a component to core. We pass in the assembler fixups, code
 ;;; vector and node info.
 
@@ -252,15 +236,59 @@
 ;;; are 0, so the jump table count is 0.
 ;;; Similar considerations pertain to x86[-64] fixups within the machine code.
 
-(defun assign-code-serialno (code-obj)
+(defun code-header/trailer-adjust (code-obj expected-nwords n-fdefns)
+  (declare (ignorable expected-nwords n-fdefns))
+  ;; Serial# shares a word with the jump-table word count,
+  ;; so we can't assign serial# until after all raw bytes are copied in.
+  ;; Do we need unique IDs on the various strange kind of code blobs? These would
+  ;; include code from MAKE-SIMPLIFYING-TRAMPOLINE, ENCAPSULATE-FUNOBJ, MAKE-BPT-LRA.
   (let* ((serialno (ldb (byte (byte-size sb-vm::code-serialno-byte) 0)
                         (atomic-incf *code-serialno*)))
          (insts (code-instructions code-obj))
          (jumptable-word (sap-ref-word insts 0)))
     (aver (zerop (ash jumptable-word -14)))
-    (setf (sb-vm::sap-ref-word-jit insts 0) ; insert serialno
+    (setf (sap-ref-word insts 0) ; insert serialno
           (logior (ash serialno (byte-position sb-vm::code-serialno-byte))
-                  jumptable-word))))
+                  jumptable-word)))
+  #+64-bit
+  (let ((base (sap+ (int-sap (get-lisp-obj-address code-obj)) (- sb-vm:other-pointer-lowtag)))
+        (physical-nwords ; upper 4 bytes of the header word
+         (ash (get-header-data code-obj) -24)))
+    (setf (sap-ref-32 base (+ (ash sb-vm:code-boxed-size-slot sb-vm:word-shift)
+                              #+little-endian 4))
+          n-fdefns)
+    (when (/= physical-nwords expected-nwords)
+      ;; Oversized allocation must be exactly 2 words more than requested
+      (aver (= (- physical-nwords 2) expected-nwords))
+      ;; Point just beyond the trailer word (physically and where it should be)
+      (let* ((new-trailer (sap+ base (ash physical-nwords sb-vm:word-shift)))
+             (old-trailer (sap+ base (ash (- physical-nwords 2) sb-vm:word-shift)))
+             (trailer-length (sap-ref-16 old-trailer -2)) ; in bytes
+             (trailer-nelements (floor trailer-length 4)))
+        ;; this is a memmove() and memset()
+        (dotimes (i trailer-nelements)
+          ;; Transfer 4 bytes per element (uint32_t), highest address first
+          ;; since we're moving upward in memory.
+          (let ((offset (* (1+ i) -4)))
+            (setf (sap-ref-32 new-trailer offset) (sap-ref-32 old-trailer offset) )))
+        ;; Zeroize at most 4 elements in the "old" trailer. These will be the
+        ;; items at the lowest addresses (the highest indices in negative order).
+        ;; If there are fewer than 4 elements, then only zeroize that many.
+        (loop for offset from (* trailer-nelements -4) by 4
+              repeat (min trailer-nelements 4)
+              do (setf (sap-ref-32 old-trailer offset) 0))
+        ;; Increase the trailer length by 2 lispwords
+        (incf (sap-ref-16 new-trailer -2) (* 2 sb-vm:n-word-bytes)))))
+  ;; Enforce that the final unboxed data word is published to memory
+  ;; before the debug-info is set.
+  (sb-thread:barrier (:write)))
+
+#+darwin-jit
+(defun assign-code-constants (code data)
+  (let* ((sb-vm::*pinned-objects* ; Pin DATA plus every element of it.
+          (list* code data (nconc (coerce data 'list) sb-vm::*pinned-objects*))))
+    (sb-vm::jit-copy-code-constants (get-lisp-obj-address code)
+                                    (get-lisp-obj-address data))))
 
 (defun make-core-component (component segment length fixup-notes alloc-points object)
   (declare (type component component)
@@ -268,65 +296,44 @@
            (type index length)
            (list fixup-notes)
            (type core-object object))
-  (let* ((debug-info (debug-info-for-component component))
+  (binding*
+        ((debug-info (debug-info-for-component component))
          (2comp (component-info component))
          (constants (ir2-component-constants 2comp))
-         (nboxed (align-up (length constants) sb-c::code-boxed-words-align))
+         (n-boxed-words (length constants))
+         (boxed-data
+          ;; <header, boxed_size, debug_info, fixups> are absent from the simple-vector
+          (or #+darwin-jit
+              (make-array (- n-boxed-words sb-vm:code-constants-offset) :initial-element 0)))
+         (const-patch-start-index
+          (+ sb-vm:code-constants-offset (* (length (ir2-component-entries 2comp))
+                                            sb-vm:code-slots-per-simple-fun)))
          (n-named-calls
           ;; Pre-scan for fdefinitions to ensure their existence.
           ;; Doing so guarantees that storing them into the boxed header now
           ;; can't create any old->young pointer, which is important since gencgc
           ;; does not deal with untagged pointers when looking for old->young.
           (do ((count 0)
-               (index (+ sb-vm:code-constants-offset
-                         (* (length (ir2-component-entries 2comp))
-                            sb-vm:code-slots-per-simple-fun))
-                      (1+ index)))
-              ((>= index (length constants)) count)
-            (let* ((const (aref constants index))
-                   (kind (if (listp const) (car const) const)))
-              (case kind
-                ((member :named-call :fdefinition)
-                 (setf (second const) (find-or-create-fdefn (second const)))
-                 (when (eq kind :named-call) (incf count)))))))
-         (code-obj (allocate-code-object (component-mem-space component)
-                                         n-named-calls nboxed length))
-         (named-call-fixups
-      ;; The following operations need the code pinned:
-      ;; 1. copying into code-instructions (a SAP)
-      ;; 2. apply-core-fixups and sanctify-for-execution
-      ;; A very specific store order is necessary to allow using uninitialized memory
-      ;; pages for code. Storing of the debug-info slot does not need the code pinned,
-      ;; but that store must occur between steps 1 and 2.
-          (with-pinned-objects (code-obj)
-           (let ((bytes (the (simple-array assembly-unit 1)
-                             (segment-contents-as-vector segment))))
-             ;; Note that this does not have to take care to ensure atomicity
-             ;; of the store to the final word of unboxed data. Even if BYTE-BLT were
-             ;; interrupted in between the store of any individual byte, this code
-             ;; is GC-safe because we no longer need to know where simple-funs are embedded
-             ;; within the object to trace pointers. We *do* need to know where the funs
-             ;; are when transporting the object, but it's currently pinned.
-             #-darwin-jit
-             (%byte-blt bytes 0 (code-instructions code-obj) 0 (length bytes))
-             #+darwin-jit
-             (with-pinned-objects (bytes)
-               (sb-vm::jit-memcpy (code-instructions code-obj) (vector-sap bytes) (length bytes)))
-
-             ;; Serial# shares a word with the jump-table word count,
-             ;; so we can't assign serial# until after all raw bytes are copied in.
-             (assign-code-serialno code-obj))
-           ;; Enforce that the final unboxed data word is published to memory
-           ;; before the debug-info is set.
-           (sb-thread:barrier (:write))
-           ;; Until debug-info is assigned, it is illegal to create a simple-fun pointer
-           ;; into this object, because the C code assumes that the fun table is in an
-           ;; invalid/incomplete state (i.e. can't be read) until the code has debug-info.
-           ;; That is, C code can't deal with an interior code pointer until the fun-table
-           ;; is valid. This store must occur prior to calling %CODE-ENTRY-POINT, and
-           ;; applying fixups calls %CODE-ENTRY-POINT, so we have to do this before that.
-            (setf (%code-debug-info code-obj) debug-info)
-            (apply-core-fixups fixup-notes code-obj))))
+               (index const-patch-start-index (1+ index)))
+              ((>= index n-boxed-words) count)
+            (let ((const (aref constants index)))
+              (when (and (listp const) (case (car const)
+                                         (:named-call (incf count))
+                                         (:fdefinition t)))
+                (setf (second const) (find-or-create-fdefn (second const)))))))
+         ((code-obj total-nwords)
+          (allocate-code-object (component-mem-space component)
+                                (align-up n-boxed-words code-boxed-words-align)
+                                length))
+         (bytes
+          (the (simple-array assembly-unit 1) (segment-contents-as-vector segment)))
+         (n-simple-funs (length (ir2-component-entries 2comp)))
+         (named-call-fixups nil))
+    (declare (ignorable boxed-data))
+    (sb-fasl::with-writable-code-instructions
+        (code-obj total-nwords debug-info n-named-calls n-simple-funs)
+      :copy (%byte-blt bytes 0 (code-instructions code-obj) 0 (length bytes))
+      :fixup (setq named-call-fixups (apply-core-fixups code-obj fixup-notes)))
 
     (when alloc-points
       #+(and x86-64 sb-thread)
@@ -334,49 +341,52 @@
           (setf (gethash code-obj *allocation-patch-points*) alloc-points)
           (funcall 'sb-aprof::patch-code code-obj alloc-points)))
 
-      ;; Don't need code pinned now
-      ;; (It will implicitly be pinned on the conservatively scavenged backends)
-    (let* ((entries (ir2-component-entries 2comp))
-           (fun-index (length entries)))
-      (dolist (entry-info entries)
-        (let ((fun (%code-entry-point code-obj (decf fun-index)))
-              (w (+ sb-vm:code-constants-offset
-                    (* sb-vm:code-slots-per-simple-fun fun-index))))
-          (aver (functionp fun)) ; in case %CODE-ENTRY-POINT returns NIL
-          (setf (code-header-ref code-obj (+ w sb-vm:simple-fun-name-slot))
-                (entry-info-name entry-info)
-                (code-header-ref code-obj (+ w sb-vm:simple-fun-arglist-slot))
-                (entry-info-arguments entry-info)
-                (code-header-ref code-obj (+ w sb-vm:simple-fun-source-slot))
-                (entry-info-form/doc entry-info)
-                (code-header-ref code-obj (+ w sb-vm:simple-fun-info-slot))
-                (entry-info-type/xref entry-info))
-          (note-fun entry-info fun object))))
-
     (push debug-info (core-object-debug-info object))
 
-    (do ((index (+ sb-vm:code-constants-offset
-                   (* (length (ir2-component-entries 2comp))
-                      sb-vm:code-slots-per-simple-fun))
-                (1+ index)))
-        ((>= index (length constants)))
-      (let* ((const (aref constants index))
-             (kind (if (listp const) (car const) const)))
-        (case kind
-          (:entry
-           (reference-core-fun code-obj index (cadr const) object))
-          ((nil))
-          (t
-           (let ((referent
-                  (etypecase kind
-                   ((member :named-call :fdefinition) (cadr const))
-                   ((eql :known-fun)
-                    (%coerce-name-to-fun (cadr const)))
-                   (constant
-                    (constant-value const)))))
-             (if (eq kind :named-call)
-                 (set-code-fdefn code-obj index referent)
-                 (setf (code-header-ref code-obj index) referent)))))))
+    ;; Don't need code pinned now
+    ;; (It will implicitly be pinned on the conservatively scavenged backends)
+    (macrolet ((set-boxed-word (i val &optional fdefnp)
+                 (declare (ignorable fdefnp))
+                 ;; Thankfully we don't mix untagged fdefns with darwin-jit
+                 #+darwin-jit
+                 `(setf (svref boxed-data (- ,i ,sb-vm:code-constants-offset)) ,val)
+                 #-darwin-jit
+                 (if (not fdefnp) ; statically not an fdefn
+                     `(setf (code-header-ref code-obj ,i) ,val)
+                     `(if ,fdefnp ; else choose which setter to use
+                          (set-code-fdefn code-obj ,i ,val)
+                          (setf (code-header-ref code-obj ,i) ,val)))))
+
+      (let* ((entries (ir2-component-entries 2comp))
+             (fun-index (length entries)))
+        (dolist (entry-info entries)
+          (let ((fun (%code-entry-point code-obj (decf fun-index)))
+                (w (+ sb-vm:code-constants-offset
+                      (* sb-vm:code-slots-per-simple-fun fun-index))))
+            (aver (functionp fun)) ; in case %CODE-ENTRY-POINT returns NIL
+            (set-boxed-word (+ w sb-vm:simple-fun-name-slot) (entry-info-name entry-info))
+            (set-boxed-word (+ w sb-vm:simple-fun-arglist-slot) (entry-info-arguments entry-info))
+            (set-boxed-word (+ w sb-vm:simple-fun-source-slot) (entry-info-form/doc entry-info))
+            (set-boxed-word (+ w sb-vm:simple-fun-info-slot) (entry-info-type/xref entry-info))
+            (setf (gethash entry-info (core-object-entry-table object)) fun))))
+
+      (do ((index const-patch-start-index (1+ index)))
+          ((>= index n-boxed-words))
+        (let* ((const (aref constants index))
+               (kind (if (listp const) (car const) const)))
+          (set-boxed-word
+           index
+           (etypecase kind
+             ((eql :entry)
+              (the function (gethash (leaf-info (cadr const))
+                                     (core-object-entry-table object))))
+             ((member :named-call :fdefinition) (cadr const))
+             ((eql :known-fun) (%coerce-name-to-fun (cadr const)))
+             (constant (constant-value const)))
+           (eq kind :named-call))))
+
+      #+darwin-jit (assign-code-constants code-obj boxed-data))
+
     (when named-call-fixups
       (sb-vm::statically-link-code-obj code-obj named-call-fixups))
     (when sb-fasl::*show-new-code*
@@ -399,7 +409,6 @@
 (defun fix-core-source-info (info object &optional function)
   (declare (type core-object object))
   (declare (type (or null function) function))
-  (aver (zerop (hash-table-count (core-object-patch-table object))))
   (let ((source (debug-source-for-info info :function function)))
     (dolist (info (core-object-debug-info object))
       (setf (debug-info-source info) source)))
