@@ -324,25 +324,27 @@ Evaluate the FORMS in the specified SITUATIONS (any of :COMPILE-TOPLEVEL,
           (unless (listp arglist)
             (fail "The local macro argument list ~S is not a list."
                   arglist))
-          `(,name macro .
-                  ;; I guess the reason we want to compile here rather than possibly
-                  ;; using an interpreted lambda is that we generate the usual gamut
-                  ;; of style-warnings and such. One might wonder if this could somehow
-                  ;; go through the front-most part of the front-end, to deal with
-                  ;; semantics, but then generate an interpreted function or something
-                  ;; more quick to emit than machine code.
-                  ,(compile-in-lexenv
-                    (let (#-sb-xc-host
-                          (*macro-policy*
-                            ;; Make sure to save the sources if an
-                            ;; inlined functions closes over this
-                            ;; macro.
-                            (process-optimize-decl
-                             '(optimize (store-source-form 3))
-                             *macro-policy*)))
-                      (make-macro-lambda nil arglist body 'macrolet name))
-                    lexenv
-                    nil nil nil t nil))))))) ; name source-info tlf ephemeral errorp
+          (let (source-info
+                tlf)
+            #-sb-xc-host
+            (when (boundp '*source-paths*)
+              (push `(declare (source-form ,definition)) body)
+              (setf source-info *source-info*
+                    tlf (let ((path (or (get-source-path definition)
+                                        (and (boundp '*current-path*)
+                                             *current-path*))))
+                          (source-path-tlf-number path))))
+            `(,name macro .
+                    ;; I guess the reason we want to compile here rather than possibly
+                    ;; using an interpreted lambda is that we generate the usual gamut
+                    ;; of style-warnings and such. One might wonder if this could somehow
+                    ;; go through the front-most part of the front-end, to deal with
+                    ;; semantics, but then generate an interpreted function or something
+                    ;; more quick to emit than machine code.
+                    ,(compile-in-lexenv
+                      (make-macro-lambda nil arglist body 'macrolet name)
+                      lexenv
+                      nil source-info tlf t nil))))))))
 
 (defun funcall-in-macrolet-lexenv (definitions fun context)
   (%funcall-in-foomacrolet-lexenv
@@ -466,6 +468,35 @@ body, references to a NAME will effectively be replaced with the EXPANSION."
                                   (subseq args required min))
                                ,@(subseq args 0 required)
                                ,@(subseq args min)))))
+
+(defmacro inline-%primitive (template &rest args)
+  (let* ((required (length (template-arg-types template)))
+         (info (template-info-arg-count template))
+         (min (+ required info))
+         (nargs (length args)))
+    (if (template-more-args-type template)
+        (when (< nargs min)
+          (bug "Primitive was called with ~R argument~:P, ~
+                but wants at least ~R."
+               nargs
+               min))
+        (unless (= nargs min)
+          (bug "Primitive was called with ~R argument~:P, ~
+                but wants exactly ~R."
+               nargs
+               min)))
+
+    (when (template-conditional-p template)
+      (bug "%PRIMITIVE was used with a conditional template."))
+
+    (when (template-more-results-type template)
+      (bug "%PRIMITIVE was used with an unknown values template."))
+
+    `(%%primitive ',template
+                  ',(eval-info-args
+                     (subseq args required min))
+                  ,@(subseq args 0 required)
+                  ,@(subseq args min))))
 
 ;;;; QUOTE
 
@@ -513,6 +544,7 @@ Return VALUE without evaluating it."
 ;;; un-merged pathnames. I'm not daring enough to change it for everyone.
 ;;; It defaults to what it should, and is changed before saving the image.
 ;;;
+;;; FIXME: can't we just get rid of this and _never_ use TRUENAME?
 (declaim (type (member pathname truename) *name-context-file-path-selector*))
 (defglobal *name-context-file-path-selector* 'pathname)
 
@@ -540,6 +572,29 @@ Return VALUE without evaluating it."
       (list :in context))))
 
 ;;;; FUNCTION and NAMED-LAMBDA
+
+;;; `(NAMED-LAMBDA ,NAME ,@REST) is like `(FUNCTION (LAMBDA ,@REST)),
+;;; except that the value of NAME is passed to the compiler for use in
+;;; creation of debug information for the resulting function.
+;;;
+;;; NAME can be a legal function name or some arbitrary other thing.
+;;;
+;;; If NAME is a legal function name, then the caller should be
+;;; planning to set (FDEFINITION NAME) to the created function.
+;;; (Otherwise the debug names will be inconsistent and thus
+;;; unnecessarily confusing.)
+;;;
+;;; Arbitrary other things are appropriate for naming things which are
+;;; not the FDEFINITION of NAME. E.g.
+;;;   NAME = (:FLET FOO BAR)
+;;; for the FLET function in
+;;;   (DEFUN BAR (X)
+;;;     (FLET ((FOO (Y) (+ X Y)))
+;;;       FOO))
+;;; or
+;;;   NAME = (:METHOD PRINT-OBJECT :AROUND (STARSHIP T))
+;;; for the function used to implement
+;;;   (DEFMETHOD PRINT-OBJECT :AROUND ((SS STARSHIP) STREAM) ...).
 (defun name-lambdalike (thing)
   (case (car thing)
     ((named-lambda)
@@ -1177,7 +1232,7 @@ care."
              (compiler-error "~S is a constant and thus can't be set." name))
            (when (lambda-var-p leaf)
              (let ((home-lambda (ctran-home-lambda-or-null start)))
-               (when home-lambda
+               (when (and home-lambda (neq (lambda-var-home leaf) home-lambda))
                  (sset-adjoin leaf (lambda-calls-or-closes home-lambda))))
              (when (lambda-var-ignorep leaf)
                ;; ANSI's definition of "Declaration IGNORE, IGNORABLE"
