@@ -15,8 +15,6 @@
 
 ;;;; LAMBDA hackery
 
-;;;; FIXME: where is that file?
-;;;; Answer: As usual, with CMUCL.
 ;;;; Note: Take a look at the compiler-overview.tex section on "Hairy
 ;;;; function representation" before you seriously mess with this
 ;;;; stuff.
@@ -311,7 +309,7 @@
                 (dolist (default defaults)
                   (if (constantp default)
                       (default-vals default)
-                      (let ((var (sb-xc:gensym)))
+                      (let ((var (gensym)))
                         (default-bindings `(,var ,default))
                         (default-vals var))))
                 (let ((bindings (default-bindings))
@@ -451,10 +449,10 @@
                                  :type (leaf-type var)
                                  :where-from (leaf-where-from var))))
 
-    (let* ((n-context (sb-xc:gensym "N-CONTEXT-"))
+    (let* ((n-context (gensym "N-CONTEXT-"))
            (context-temp (make-lambda-var :%source-name n-context
                                           :arg-info (make-arg-info :kind :more-context)))
-           (n-count (sb-xc:gensym "N-COUNT-"))
+           (n-count (gensym "N-COUNT-"))
            (count-temp (make-lambda-var :%source-name n-count
                                         :type (specifier-type 'index)
                                         :arg-info (make-arg-info :kind :more-count))))
@@ -474,11 +472,11 @@
       ;; and take advantage of the base+index+displacement addressing
       ;; mode on x86oids.)
       (when (optional-dispatch-keyp res)
-        (let ((n-index (sb-xc:gensym "N-INDEX-"))
-              (n-key (sb-xc:gensym "N-KEY-"))
-              (n-value-temp (sb-xc:gensym "N-VALUE-TEMP-"))
-              (n-allowp (sb-xc:gensym "N-ALLOWP-"))
-              (n-lose (sb-xc:gensym "N-LOSE-"))
+        (let ((n-index (gensym "N-INDEX-"))
+              (n-key (gensym "N-KEY-"))
+              (n-value-temp (gensym "N-VALUE-TEMP-"))
+              (n-allowp (gensym "N-ALLOWP-"))
+              (n-lose (gensym "N-LOSE-"))
               (allowp (or (optional-dispatch-allowp res)
                           (policy *lexenv* (zerop safety))))
               (found-allow-p nil))
@@ -502,9 +500,9 @@
                      (keyword (arg-info-key info))
                      (supplied-p (arg-info-supplied-p info))
                      (supplied-used-p (arg-info-supplied-used-p info))
-                     (n-value (sb-xc:gensym "N-VALUE-"))
+                     (n-value (gensym "N-VALUE-"))
                      (clause (cond (supplied-p
-                                    (let ((n-supplied (sb-xc:gensym "N-SUPPLIED-")))
+                                    (let ((n-supplied (gensym "N-SUPPLIED-")))
                                       (temps (list n-supplied
                                                    (if supplied-used-p
                                                        nil
@@ -631,10 +629,10 @@
         ;; Make up two extra variables, and squirrel them away in
         ;; ARG-INFO-DEFAULT for transforming (VALUES-LIST REST) into
         ;; (%MORE-ARG-VALUES CONTEXT 0 COUNT) when possible.
-        (let* ((context-name (sb-xc:gensym "REST-CONTEXT-"))
+        (let* ((context-name (gensym "REST-CONTEXT-"))
                (context (make-lambda-var :%source-name context-name
                                          :arg-info (make-arg-info :kind :more-context)))
-               (count-name (sb-xc:gensym "REST-COUNT-"))
+               (count-name (gensym "REST-COUNT-"))
                (count (make-lambda-var :%source-name count-name
                                        :arg-info (make-arg-info :kind :more-count)
                                        :type (specifier-type 'index))))
@@ -662,7 +660,7 @@
         (main-vars val-temp)
         (bind-vars key)
         (cond ((or hairy-default supplied-p)
-               (let* ((n-supplied (sb-xc:gensym "N-SUPPLIED-"))
+               (let* ((n-supplied (gensym "N-SUPPLIED-"))
                       (supplied-temp (make-lambda-var
                                       :%source-name n-supplied)))
                  (unless supplied-p
@@ -873,6 +871,8 @@
 
     res))
 
+(defvar *lambda-conversions*)
+
 ;;; Convert a LAMBDA form into a LAMBDA leaf or an OPTIONAL-DISPATCH leaf.
 (defun ir1-convert-lambda (form &key (source-name '.anonymous.)
                            debug-name maybe-add-debug-catch
@@ -894,58 +894,61 @@
     (bug "Both SYSTEM-LAMBDA and MAYBE-ADD-DEBUG-CATCH specified"))
   (unless (or debug-name (neq '.anonymous. source-name))
     (setf debug-name (name-lambdalike form)))
-  (multiple-value-bind (forms decls doc) (parse-body (cddr form) t)
-    (let ((*lexenv* (process-muffle-decls decls *lexenv*)))
-      (multiple-value-bind (vars keyp allow-other-keys aux-vars aux-vals)
-          (make-lambda-vars (cadr form))
-        (binding* (((*lexenv* result-type post-binding-lexenv
-                              lambda-list explicit-check source-form
-                              local-policy)
-                    (process-decls decls (append aux-vars vars) nil
-                                   :binding-form-p t :allow-lambda-list t))
-                   (debug-catch-p (and maybe-add-debug-catch
-                                       *allow-instrumenting*
-                                       (policy *lexenv*
-                                           (>= insert-debug-catch 2))))
-                   (forms (if debug-catch-p
-                              (wrap-forms-in-debug-catch forms)
-                              forms))
-                   (forms (if (eq result-type *wild-type*)
-                              forms
-                              `((the ,(type-specifier result-type) (progn ,@forms)))))
-                   (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
-                   (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
-                               (ir1-convert-hairy-lambda forms vars keyp
-                                                         allow-other-keys
-                                                         aux-vars aux-vals
-                                                         :post-binding-lexenv post-binding-lexenv
-                                                         :source-name source-name
-                                                         :debug-name debug-name
-                                                         :system-lambda system-lambda))
-                              (t
-                               (ir1-convert-lambda-body forms vars
-                                                        :aux-vars aux-vars
-                                                        :aux-vals aux-vals
-                                                        :post-binding-lexenv post-binding-lexenv
-                                                        :source-name source-name
-                                                        :debug-name debug-name
-                                                        :system-lambda system-lambda
-                                                        :local-policy local-policy)))))
-          (when explicit-check
-            (setf (getf (functional-plist res) 'explicit-check) explicit-check))
-          (setf (functional-inline-expansion res) (or source-form form))
-          (setf (functional-arg-documentation res)
-                (if (eq lambda-list :unspecified)
-                    (strip-lambda-list (cadr form) :arglist)
-                    lambda-list))
-          (setf (functional-documentation res) doc)
-          (when (boundp '*lambda-conversions*)
-            ;; KLUDGE: Not counting TL-XEPs is a lie, of course, but
-            ;; keeps things less confusing to users of TIME, where this
-            ;; count gets used.
-            (unless (and (consp debug-name) (eq 'tl-xep (car debug-name)))
-              (incf *lambda-conversions*)))
-          res)))))
+  (binding* (((forms decls doc) (parse-body (cddr form) t))
+             ((*lexenv* source-form) (process-muffle-decls decls *lexenv*))
+             (*current-path* (or (and source-form
+                                      (get-source-path source-form))
+                                 *current-path*))
+             ((vars keyp allow-other-keys aux-vars aux-vals)
+              (make-lambda-vars (cadr form)))
+             ((*lexenv* result-type post-binding-lexenv
+                        lambda-list explicit-check source-form
+                        local-policy)
+              (process-decls decls (append aux-vars vars) nil
+                             :binding-form-p t :allow-lambda-list t))
+             (debug-catch-p (and maybe-add-debug-catch
+                                 *allow-instrumenting*
+                                 (policy *lexenv*
+                                     (>= insert-debug-catch 2))))
+             (forms (if debug-catch-p
+                        (wrap-forms-in-debug-catch forms)
+                        forms))
+             (forms (if (eq result-type *wild-type*)
+                        forms
+                        `((the ,(type-specifier result-type) (progn ,@forms)))))
+             (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
+             (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
+                         (ir1-convert-hairy-lambda forms vars keyp
+                                                   allow-other-keys
+                                                   aux-vars aux-vals
+                                                   :post-binding-lexenv post-binding-lexenv
+                                                   :source-name source-name
+                                                   :debug-name debug-name
+                                                   :system-lambda system-lambda))
+                        (t
+                         (ir1-convert-lambda-body forms vars
+                                                  :aux-vars aux-vars
+                                                  :aux-vals aux-vals
+                                                  :post-binding-lexenv post-binding-lexenv
+                                                  :source-name source-name
+                                                  :debug-name debug-name
+                                                  :system-lambda system-lambda
+                                                  :local-policy local-policy)))))
+    (when explicit-check
+      (setf (getf (functional-plist res) 'explicit-check) explicit-check))
+    (setf (functional-inline-expansion res) (or source-form form))
+    (setf (functional-arg-documentation res)
+          (if (eq lambda-list :unspecified)
+              (strip-lambda-list (cadr form) :arglist)
+              lambda-list))
+    (setf (functional-documentation res) doc)
+    (when (boundp '*lambda-conversions*)
+      ;; KLUDGE: Not counting TL-XEPs is a lie, of course, but
+      ;; keeps things less confusing to users of TIME, where this
+      ;; count gets used.
+      (unless (and (consp debug-name) (eq 'tl-xep (car debug-name)))
+        (incf *lambda-conversions*)))
+    res))
 
 (defun wrap-forms-in-debug-catch (forms)
   #+unwind-to-frame-and-call-vop
@@ -978,14 +981,6 @@
              (progn
                ,@forms))))))))
 
-;; FIXME: really should be an aspect of the lexical environment,
-;; but LEXENVs don't know whether they are toplevel or not.
-(defun has-toplevelness-decl (lambda-expr)
-  (dolist (expr (cddr lambda-expr)) ; Skip over (LAMBDA (ARGS))
-    (cond ((equal expr '(declare (top-level-form))) (return t))
-          ((typep expr '(or (cons (eql declare)) string))) ; DECL | DOCSTRING
-          (t (return nil)))))
-
 ;;; helper for LAMBDA-like things, to massage them into a form
 ;;; suitable for IR1-CONVERT-LAMBDA.
 (defun ir1-convert-lambdalike (thing
@@ -1003,7 +998,16 @@
     ((named-lambda)
      (let* ((name (cadr thing))
             (lambda-expression `(lambda ,@(cddr thing)))
-            (*inline-expansions* (list name 1 *inline-expansions*)))
+            (*inline-expansions* (list name 1 *inline-expansions*))
+            (simple-lexenv-p (simple-lexical-environment-p *lexenv*)))
+       ;; Discard any forward references to this function unless we
+       ;; are block compiling and the lexical environment doesn't
+       ;; contain any hair. If the lexical environment is too hairy, then we only
+       ;; install the definition during the processing of this NAMED-LAMBDA,
+       ;; ensuring that the function cannot be called outside of the correct
+       ;; environment.
+       (unless (and (block-compile *compilation*) simple-lexenv-p)
+         (remhash name (free-funs *ir1-namespace*)))
        (if (and name (legal-fun-name-p name))
            (let ((defined-fun-res (get-defined-fun name (second lambda-expression)))
                  (res (ir1-convert-lambda lambda-expression
@@ -1011,13 +1015,15 @@
                                           :source-name name))
                  (info (info :function :info name)))
              (setf (functional-inlinep res) (info :function :inlinep name)
-                   (defined-fun-named-lambda-p defined-fun-res) t)
-             (when (has-toplevelness-decl lambda-expression)
-               (setf (functional-top-level-defun-p res) t))
+                   (defined-fun-same-block-p defined-fun-res) t)
              ;; FIXME: Should non-entry block compiled defuns have
              ;; this propagate?
              (assert-global-function-definition-type name res)
-             (push res (defined-fun-functionals defined-fun-res))
+             ;; If in a simple environment, then we can allow
+             ;; backward references to this function from following
+             ;; top-level forms.
+             (when simple-lexenv-p
+               (push res (defined-fun-functionals defined-fun-res)))
              (unless (or
                       (eq (defined-fun-inlinep defined-fun-res) 'notinline)
                       ;; Don't treat recursive stubs like CAR as self-calls
@@ -1049,8 +1055,8 @@
 
 (declaim (end-block))
 
-
 ;;;; defining global functions
+
 ;;; Given a lambda-list, return a FUN-TYPE object representing the signature:
 ;;; return type is *, and each individual arguments type is T -- but we get
 ;;; the argument counts and keywords.
@@ -1072,111 +1078,59 @@
             (allow (when (ll-kwds-allowp llks) '(&allow-other-keys))))
         (careful-specifier-type `(function (,@reqs ,@opts ,@rest ,@keys ,@allow) *))))))
 
-(declaim (start-block maybe-inline-syntactic-closure))
-
-;;; Take the lexenv surrounding an inlined function and extract things
-;;; needed for the inline expansion suitable for dumping into fasls.
-;;; Right now it's MACROLET, SYMBOL-MACROLET, SPECIAL and
-;;; INLINE/NOTINLINE declarations. Upon encountering something else return NIL.
-;;; This is later used by PROCESS-INLINE-LEXENV to reproduce the lexenv.
-;;;
-;;; Previously it just used the functions and vars of the innermost
-;;; lexenv, but the body of macrolet can refer to other macrolets
-;;; defined earlier, so it needs to process all the parent lexenvs to
-;;; recover the proper order.
-(defun reconstruct-lexenv (lexenv)
-  (let (shadowed-funs
-        shadowed-vars
-        result)
-    (loop for env = lexenv then parent
-          for parent = (lexenv-parent env)
-          for vars = (lexenv-vars env)
-          for funs = (lexenv-funs env)
-          for declarations = nil
-          for symbol-macros = nil
-          for macros = nil
-          do
-          (loop for binding in vars
-                for (name . what) = binding
-                unless (and parent
-                            (find binding (lexenv-vars parent)))
-                do (typecase what
-                     (cons
-                      (aver (eq (car what) 'macro))
-                      (push name shadowed-vars)
-                      (push (list name (cdr what)) symbol-macros))
-                     (global-var
-                      (aver (eq (global-var-kind what) :special))
-                      (push `(special ,name) declarations))
-                     (t
-                      (unless (memq name shadowed-vars)
-                        (return-from reconstruct-lexenv)))))
-          (loop for binding in funs
-                for (name . what) = binding
-                unless (and parent
-                            (find binding (lexenv-funs parent)))
-                do
-                (typecase what
-                  (cons
-                   (push name shadowed-funs)
-                   (let ((expression (function-lambda-expression (cdr what))))
-                     (aver expression)
-                     (push (cons name expression) macros)))
-                  ;; FIXME: Is there a good reason for this not to be
-                  ;; DEFINED-FUN (which :INCLUDEs GLOBAL-VAR, in case
-                  ;; you're wondering how this ever worked :-)? Maybe
-                  ;; in conjunction with an AVERrance that it's not an
-                  ;; (AND GLOBAL-VAR (NOT GLOBAL-FUN))? -- CSR,
-                  ;; 2002-07-08
-                  (global-var
-                   (unless (defined-fun-p what)
-                     (return-from reconstruct-lexenv))
-                   (push `(,(car (defined-fun-inlinep what))
-                           ,name)
-                         declarations))
-                  (t
-                   (unless (memq name shadowed-funs)
-                     (return-from reconstruct-lexenv)))))
-          (when declarations
-            (setf result (list* :declare declarations (and result (list result)))))
-          (when symbol-macros
-            (setf result (list* :symbol-macro symbol-macros (and result (list result)))))
-          (when macros
-            (setf result (list* :macro macros (and result (list result)))))
-          while (and parent
-                     (not (null-lexenv-p parent))))
-    result))
-
-;;; Return a sexpr for LAMBDA in LEXENV such that loading it from fasl
-;;; preserves the original lexical environment for inlining.
-;;; Return NIL if the lexical environment is too complicated.
-(defun maybe-inline-syntactic-closure (lambda lexenv)
+;;; Return a lambda form that has been "closed" with respect to
+;;; LEXENV, returning a LAMBDA-WITH-LEXENV if there are interesting
+;;; declarations. To handle local macros, rather than closing over
+;;; definitions in the environment, expand all macros in the body of
+;;; LAMBDA, so that nothing in the syntactic environment is needed in
+;;; the expansion. If there is something too complex in the lexical
+;;; environment (like a lexical variable), then we return NIL.
+(defun inline-syntactic-closure-lambda (lambda lexenv)
   (declare (type list lambda) (type lexenv-designator lexenv))
   (aver (eql (first lambda) 'lambda))
-  ;; We used to have a trivial implementation, verifying that lexenv
-  ;; was effectively null. However, this fails to take account of the
-  ;; idiom
-  ;;
-  ;; (declaim (inline foo))
-  ;; (macrolet ((def (x) `(defun ,x () ...)))
-  ;;   (def foo))
-  ;;
-  ;; which, while too complicated for the cross-compiler to handle in
-  ;; unfriendly foreign lisp environments, would be good to support in
-  ;; the target compiler. -- CSR, 2002-05-13 and 2002-11-02
   (typecase lexenv
    (lexenv
     (let ((vars (lexenv-vars lexenv))
-          (funs (lexenv-funs lexenv)))
-      (acond ((or (lexenv-blocks lexenv) (lexenv-tags lexenv)) nil)
-             ((and (null vars) (null funs)) lambda)
-             ;; If the lexenv is too hairy for cross-compilation,
-             ;; you'll find out later, when trying to perform inlining.
-             ;; This is fine, because if the inline expansion is only
-             ;; for the target, it's totally OK to cross-compile this
-             ;; defining form. The syntactic env is correctly captured.
-             ((reconstruct-lexenv lexenv)
-              `(lambda-with-lexenv ,it ,@(cdr lambda))))))
+          (funs (lexenv-funs lexenv))
+          (decls ()))
+      (cond ((or (lexenv-blocks lexenv) (lexenv-tags lexenv)) nil)
+            ((and (null vars) (null funs)) lambda)
+            ((dolist (x vars nil)
+               (let ((name (car x))
+                     (what (cdr x)))
+                 (when (eq x (assoc name vars :test #'eq))
+                   (typecase what
+                     (cons
+                      (aver (eq (car what) 'macro)))
+                     (global-var
+                      (aver (eq (global-var-kind what) :special))
+                      (push `(special ,name) decls))
+                     (t (return t))))))
+             nil)
+            ((dolist (x funs nil)
+               (let ((name (car x))
+                     (what (cdr x)))
+                 (when (eq x (assoc name funs :test #'equal))
+                   (typecase what
+                     (cons)
+                     ;; FIXME: Is there a good reason for this not to be
+                     ;; DEFINED-FUN (which :INCLUDEs GLOBAL-VAR, in case
+                     ;; you're wondering how this ever worked :-)? Maybe
+                     ;; in conjunction with an AVERrance that it's not an
+                     ;; (AND GLOBAL-VAR (NOT GLOBAL-FUN))? -- CSR,
+                     ;; 2002-07-08
+                     (global-var
+                      (when (defined-fun-p what)
+                        (push `(,(car (defined-fun-inlinep what))
+                                ,name)
+                              decls)))
+                     (t (return t))))))
+             nil)
+            (t
+             (let ((expansion (sb-walker:macroexpand-all lambda lexenv)))
+               (if decls
+                   `(lambda-with-lexenv (:declare ,decls) ,@(cdr expansion))
+                   expansion))))))
    #+(and sb-fasteval (not sb-xc-host))
    (sb-interpreter:basic-env
     (awhen (sb-interpreter::reconstruct-syntactic-closure-env lexenv)
@@ -1184,8 +1138,6 @@
    #+sb-fasteval
    (null lambda))) ; trivial case. Never occurs in the compiler.
 
-(declaim (end-block))
-
 (declaim (start-block ir1-convert-inline-lambda))
 
 ;;; Convert the forms produced by RECONSTRUCT-LEXENV to LEXENV
@@ -1335,6 +1287,18 @@
                 "previous declaration"
                 "previous definition"))))
 
+;;; The lexical environment is hairy if it has stuff like lexical
+;;; variables, blocks, or tags that are not load-time constants. Local
+;;; functions and macros are OK, since we are worried about outside
+;;; calls. Macros get expanded, and local functions are load-time
+;;; constants.
+(defun simple-lexical-environment-p (lexenv)
+  (and (null (lexenv-blocks lexenv))
+       (null (lexenv-tags lexenv))
+       (every (lambda (entry)
+                (consp (cdr entry)))
+              (lexenv-vars lexenv))))
+
 ;;; Used for global inline expansion. Earlier something like this was
 ;;; used by %DEFUN too. FIXME: And now it's probably worth rethinking
 ;;; whether this function is a good idea at all.
@@ -1374,16 +1338,17 @@
         (note-inlining (optional-dispatch-more-entry fun))
         (mapc #'note-inlining (optional-dispatch-entry-points fun))))
     ;; substitute for any old references
-    (unless (or (eq (defined-fun-inlinep var) :notinline)
+    (unless (or (eq (defined-fun-inlinep var) 'notinline)
                 (not (block-compile *compilation*))
                 (and info
                      (or (fun-info-transforms info)
                          (fun-info-templates info)
                          (fun-info-ir2-convert info))))
       (substitute-leaf fun var)
-      ;; If in a simple environment, then we can allow backward references
-      ;; to this function from following top-level forms.
-      (when expansion
+      ;; If in a simple environment, then we can allow backward
+      ;; references to this function from following top-level
+      ;; forms.
+      (when (simple-lexical-environment-p *lexenv*)
         (push fun (defined-fun-functionals var))))
     fun))
 
