@@ -50,7 +50,9 @@
                         (funcall 'sb-pcl::fun-doc (transform-function transform)))
       "optimize"))
 
-(defprinter (transform) type note important)
+(defmethod print-object ((x transform) stream)
+  (print-unreadable-object (x stream :type t :identity t)
+    (princ (type-specifier (transform-type x)) stream)))
 
 ;;; Grab the FUN-INFO and enter the function, replacing any old
 ;;; one with the same type and note.
@@ -87,9 +89,10 @@
                             (type-specifier ctype)
                             ctype)))
     (dolist (name names)
-      (block ignore
-        (unless overwrite-fndb-silently
-          (let ((old-fun-info (info :function :info name)))
+      (let ((old-fun-info (info :function :info name))
+            inherit)
+        (block ignore
+          (unless overwrite-fndb-silently
             (when old-fun-info
               ;; This is handled as an error because it's generally a bad
               ;; thing to blow away all the old optimization stuff. It's
@@ -104,22 +107,35 @@
                   (cerror "Go ahead, overwrite it."
                           "~@<overwriting old FUN-INFO ~2I~_~S ~I~_for ~S~:>"
                           old-fun-info name)
+                (continue ()
+                  :report "Inherit templates and optimizers"
+                  (setf inherit t))
                 (ignore ()
-                  (return-from ignore))))))
-        (setf (info :function :type name) type-to-store)
-        (setf (info :function :where-from name) :declared)
-        (setf (info :function :kind name) :function)
-        (setf (info :function :info name)
-              (make-fun-info :attributes attributes
-                             :derive-type derive-type
-                             :optimizer optimizer
-                             :result-arg result-arg
-                             :call-type-deriver call-type-deriver
-                             :annotation annotation))
-        (if location
-            (setf (getf (info :source-location :declaration name) 'defknown)
-                  location)
-            (remf (info :source-location :declaration name) 'defknown)))))
+                  (return-from ignore)))))
+          (setf (info :function :type name) type-to-store)
+          (setf (info :function :where-from name) :declared)
+          (setf (info :function :kind name) :function)
+          (cond (inherit
+                 (when optimizer
+                   (setf (fun-info-optimizer old-fun-info) optimizer))
+                 (when derive-type
+                   (setf (fun-info-derive-type old-fun-info) derive-type))
+                 (setf (fun-info-attributes old-fun-info) attributes
+                       (fun-info-result-arg old-fun-info) result-arg
+                       (fun-info-annotation old-fun-info) annotation
+                       (fun-info-call-type-deriver old-fun-info) call-type-deriver))
+                (t
+                 (setf (info :function :info name)
+                       (make-fun-info :attributes attributes
+                                      :derive-type derive-type
+                                      :optimizer optimizer
+                                      :result-arg result-arg
+                                      :call-type-deriver call-type-deriver
+                                      :annotation annotation))))
+          (if location
+              (setf (getf (info :source-location :declaration name) 'defknown)
+                    location)
+              (remf (info :source-location :declaration name) 'defknown))))))
   names)
 
 
@@ -155,6 +171,8 @@
     (setq attributes (union '(unwind) attributes)))
   (when (member 'flushable attributes)
     (pushnew 'unsafely-flushable attributes))
+  #-(or x86-64 arm64) ;; Needs to be supported by the call VOPs
+  (setf attributes (remove 'no-verify-arg-count attributes))
   (multiple-value-bind (type annotation)
       (split-type-info arg-types result-type)
     `(%defknown ',(if (and (consp name)
@@ -256,10 +274,10 @@
 
 ;;;; generic type inference methods
 
-(defun symeval-derive-type (node &aux (args (basic-combination-args node))
+(defun symbol-value-derive-type (node &aux (args (basic-combination-args node))
                                       (lvar (pop args)))
   (unless (and lvar (endp args))
-    (return-from symeval-derive-type))
+    (return-from symbol-value-derive-type))
   (if (constant-lvar-p lvar)
       (let* ((sym (lvar-value lvar))
              (var (maybe-find-free-var sym))
@@ -287,15 +305,6 @@
     (let ((lvar (nth n (combination-args call))))
       (when lvar (lvar-type lvar)))))
 
-;;; Derive the result type according to the float contagion rules, but
-;;; always return a float. This is used for irrational functions that
-;;; preserve realness of their arguments.
-(defun result-type-float-contagion (call)
-  (declare (type combination call))
-  (reduce #'numeric-contagion (combination-args call)
-          :key #'lvar-type
-          :initial-value (specifier-type 'single-float)))
-
 (defun simplify-list-type (type &key preserve-dimensions)
   ;; Preserve all the list types without dragging
   ;; (cons (eql 10)) stuff in.
@@ -318,7 +327,7 @@
                                        preserve-vector-type)
   (lambda (call)
     (declare (type combination call))
-    (let ((lvar (nth (1- n) (combination-args call))))
+    (let ((lvar (nth n (combination-args call))))
       (when lvar
         (let ((type (lvar-type lvar)))
           (cond ((simplify-list-type type
@@ -340,7 +349,7 @@
 (defun result-type-specifier-nth-arg (n)
   (lambda (call)
     (declare (type combination call))
-    (let ((lvar (nth (1- n) (combination-args call))))
+    (let ((lvar (nth n (combination-args call))))
       (when (and lvar (constant-lvar-p lvar))
         (careful-specifier-type (lvar-value lvar))))))
 
@@ -358,7 +367,7 @@
 (defun creation-result-type-specifier-nth-arg (n)
   (lambda (call)
     (declare (type combination call))
-    (let ((lvar (nth (1- n) (combination-args call))))
+    (let ((lvar (nth n (combination-args call))))
       (when (and lvar (constant-lvar-p lvar))
         (let* ((specifier (lvar-value lvar))
                (lspecifier (if (atom specifier) (list specifier) specifier)))
@@ -450,16 +459,6 @@
             (max-dim (lvar-type lvar))
             (values max min))))))
 
-(defun position-derive-type (call)
-  (let ((dim (sequence-lvar-dimensions (second (combination-args call)))))
-    (when (integerp dim)
-      (specifier-type `(or (integer 0 (,dim)) null)))))
-
-(defun count-derive-type (call)
-  (let ((dim (sequence-lvar-dimensions (second (combination-args call)))))
-    (when (integerp dim)
-      (specifier-type `(integer 0 ,dim)))))
-
 ;;; This used to be done in DEFOPTIMIZER DERIVE-TYPE, but
 ;;; ASSERT-CALL-TYPE already asserts the ARRAY type, so it gets an extra
 ;;; assertion that may not get eliminated and requires extra work.
