@@ -67,12 +67,13 @@
   ;; The block may end up being deleted due to cast optimization
   ;; caused by USE-GOOD-FOR-DX-P
   (unless (node-to-be-deleted-p call)
-    (let ((dx-lvars
-            (loop for arg in (basic-combination-args call)
-                  for var in (lambda-vars fun)
-                  for dx = (leaf-dynamic-extent var)
-                  when (and dx arg (not (lvar-dynamic-extent arg)))
-                    append (handle-nested-dynamic-extent-lvars dx arg))))
+    (let* ((*dx-lexenv* (node-lexenv call))
+           (dx-lvars
+             (loop for arg in (basic-combination-args call)
+                   for var in (lambda-vars fun)
+                   for dx = (leaf-dynamic-extent var)
+                   when (and dx arg (not (lvar-dynamic-extent arg)))
+                   append (handle-nested-dynamic-extent-lvars dx arg))))
       (when dx-lvars
         (let* ((entry (with-ir1-environment-from-node call
                         (make-entry)))
@@ -175,7 +176,7 @@
   (declare (type functional fun))
   (etypecase fun
     (clambda
-     (let* ((n-supplied (sb-xc:gensym))
+     (let* ((n-supplied (gensym))
             (nargs (length (lambda-vars fun)))
             (temps (make-gensym-list nargs)))
        `(lambda (,n-supplied ,@temps)
@@ -188,7 +189,7 @@
      (let* ((min (optional-dispatch-min-args fun))
             (max (optional-dispatch-max-args fun))
             (more (optional-dispatch-more-entry fun))
-            (n-supplied (sb-xc:gensym))
+            (n-supplied (gensym))
             (temps (make-gensym-list max))
             (main (optional-dispatch-main-entry fun))
             (optional-vars (nthcdr min (lambda-vars main)))
@@ -1047,10 +1048,13 @@
                     (return-result (lambda-return (node-home-lambda call)))
                     (node-lvar call)))
           (call-type (node-derived-type call)))
-      (unless (eq call-type *wild-type*)
-        ;; FIXME: Replace the call with unsafe CAST. -- APD, 2003-01-26
-        (do-uses (use result)
-          (derive-node-type use call-type)))
+      ;; FIXME: Replace the call with unsafe CAST. -- APD, 2003-01-26
+      (do-uses (use result)
+        ;; CRETURN is an unknown value destination, now the
+        ;; destination might be consuming just one value.
+        ;; Reoptimize to help the VALUES transform, for example.
+        (reoptimize-node use)
+        (derive-node-type use call-type))
       (substitute-lvar-uses lvar result
                             (and lvar (eq (lvar-uses lvar) call)))))
 
@@ -1196,11 +1200,10 @@
              (not (functional-inline-expanded clambda)))
     ;; ANSI requires that explicit NOTINLINE be respected.
     (or (eq (lambda-inlinep clambda) 'notinline)
-        ;; If (= LET-CONVERSION 0) we can guess that inlining
-        ;; generally won't be appreciated, but if the user
-        ;; specifically requests inlining, that takes precedence over
-        ;; our general guess.
-        (and (policy clambda (= let-conversion 0))
+        ;; If (= LET-CONVERSION 0) or (= DEBUG 3) we can guess that inlining
+        ;; generally won't be appreciated, but if the user specifically requests
+        ;; inlining, that takes precedence over our general guess.
+        (and (or (policy clambda (or (= let-conversion 0) (= debug 3))))
              (not (eq (lambda-inlinep clambda) 'inline))))))
 
 ;;; We also don't convert calls to named functions which appear in the
