@@ -140,8 +140,7 @@
          (error "Class not yet defined: ~S" name))))
 
 (defoptimizer (%typep-wrapper constraint-propagate-if)
-    ((test-value variable type) node gen)
-  (declare (ignore test-value gen))
+    ((test-value variable type) node)
   (aver (constant-lvar-p type))
   (let ((type (lvar-value type)))
     (values variable (if (ctype-p type)
@@ -174,7 +173,6 @@
   nil)
 
 (defoptimizer (%type-constraint constraint-propagate) ((x type) node gen)
-  (declare (ignore node))
   (let ((var (ok-lvar-lambda-var x gen)))
     (when var
       (let ((type (lvar-value type)))
@@ -563,52 +561,60 @@
                   (intersection-type-types type))))
 
 ;;; If necessary recurse to check the cons type.
-(defun source-transform-cons-typep (object type)
-  (let* ((car-type (cons-type-car-type type))
-         (cdr-type (cons-type-cdr-type type))
-         (car-test-p (not (type= car-type *universal-type*)))
-         (cdr-test-p (not (type= cdr-type *universal-type*))))
-    (if (and (not car-test-p) (not cdr-test-p))
-        `(consp ,object)
-        ;; CONSP can be safely weakened to LISTP if either of the CAR
-        ;; or CDR test (or both) can distinguish LIST from CONS
-        ;; by never returning T when given an input of NIL.
-        (labels ((safely-weakened (ctype)
-                   (typecase ctype
-                     (member-type
-                      (not (member nil (member-type-members ctype))))
-                     (classoid
-                      ;; can't weaken if the specifier is (CONS SYMBOL)
-                      (not (ctypep nil ctype)))
-                     ;; these are disjoint from NIL
-                     ((or cons-type numeric-type array-type character-set-type)
-                      t)
-                     (intersection-type
-                      ;; at least one of them must not spuriously return T
-                      (some #'safely-weakened (compound-type-types ctype)))
-                     (union-type
-                      ;; require that none spuriously return T
-                      (every #'safely-weakened (compound-type-types ctype)))
-                     (hairy-type
-                      ;; hack - (CONS KEYWORD) is weakenable
-                      ;; because NIL is not a keyword.
-                      (equal (hairy-type-specifier ctype)
-                             '(satisfies keywordp))))))
-          (let ((car-test
-                  (and car-test-p
-                       `((typep (car ,object) ',(type-specifier car-type)))))
-                (cdr-test
-                  (and cdr-test-p
-                       `((typep (cdr ,object) ',(type-specifier cdr-type))))))
-            ;; Being paranoid, perform the safely weakenable test first
-            ;; so that the other part doesn't execute on an object that
-            ;; it would not have gotten, were the CONSP test not weakened.
-            (cond ((and car-test-p (safely-weakened car-type))
-                   `(and (listp ,object) ,@car-test ,@cdr-test))
-                  ((and cdr-test-p (safely-weakened cdr-type))
-                   `(and (listp ,object) ,@cdr-test ,@car-test))
-                  (t
-                   `(and (consp ,object) ,@car-test ,@cdr-test))))))))
+(defun source-transform-cons-typep
+    (object type &aux (car-type (cons-type-car-type type))
+                      (cdr-type (cons-type-cdr-type type))
+                      (car-test-p (not (type= car-type *universal-type*)))
+                      (cdr-test-p (not (type= cdr-type *universal-type*))))
+   ;; CONSP can be safely weakened to LISTP if either of the CAR
+   ;; or CDR test (or both) can distinguish LIST from CONS
+   ;; by never returning T when given an input of NIL.
+   (labels ((safely-weakened (ctype)
+              (typecase ctype
+                (member-type
+                 (not (member nil (member-type-members ctype))))
+                (classoid
+                 ;; can't weaken if the specifier is (CONS SYMBOL)
+                 (not (ctypep nil ctype)))
+                ;; these are disjoint from NIL
+                ((or cons-type numeric-type array-type character-set-type)
+                 t)
+                (intersection-type
+                 ;; at least one of them must not spuriously return T
+                 (some #'safely-weakened (compound-type-types ctype)))
+                (union-type
+                 ;; require that none spuriously return T
+                 (every #'safely-weakened (compound-type-types ctype)))
+                (hairy-type
+                 ;; hack - (CONS KEYWORD) is weakenable
+                 ;; because NIL is not a keyword.
+                 (equal (hairy-type-specifier ctype)
+                        '(satisfies keywordp))))))
+     (cond
+       ((and (not car-test-p) (not cdr-test-p))
+        `(consp ,object))
+       ((and (not cdr-test-p)
+             (member-type-p car-type)
+             (vop-existsp :translate car-eq-if-listp)
+             (type-singleton-p car-type)
+             (typep (first (member-type-members car-type)) '(and symbol (not null))))
+        `(car-eq-if-listp ,object ',(first (member-type-members car-type))))
+       (t
+        (let ((car-test
+               (and car-test-p
+                    `((typep (car ,object) ',(type-specifier car-type)))))
+              (cdr-test
+               (and cdr-test-p
+                    `((typep (cdr ,object) ',(type-specifier cdr-type))))))
+          ;; Being paranoid, perform the safely weakenable test first
+          ;; so that the other part doesn't execute on an object that
+          ;; it would not have gotten, were the CONSP test not weakened.
+          (cond ((and car-test-p (safely-weakened car-type))
+                 `(and (listp ,object) ,@car-test ,@cdr-test))
+                ((and cdr-test-p (safely-weakened cdr-type))
+                 `(and (listp ,object) ,@cdr-test ,@car-test))
+                (t
+                 `(and (consp ,object) ,@car-test ,@cdr-test))))))))
 
 (defun source-transform-character-set-typep (object type)
   (let ((pairs (character-set-type-pairs type)))
@@ -620,7 +626,7 @@
                 `(base-char-p ,object))
                ((= (cdar pairs) (1- char-code-limit))
                 `(characterp ,object))))
-        (let ((n-code (sb-xc:gensym "CODE")))
+        (let ((n-code (gensym "CODE")))
           `(and (characterp ,object)
                 (let ((,n-code (char-code ,object)))
                   (or
@@ -632,7 +638,7 @@
 (defun source-transform-simd-pack-typep (object type)
   (if (type= type (specifier-type 'simd-pack))
       `(simd-pack-p ,object)
-      (let ((n-tag (sb-xc:gensym "TAG")))
+      (let ((n-tag (gensym "TAG")))
         `(and
           (simd-pack-p ,object)
           (let ((,n-tag (%simd-pack-tag ,object)))
@@ -646,7 +652,7 @@
 (defun source-transform-simd-pack-256-typep (object type)
   (if (type= type (specifier-type 'simd-pack-256))
       `(simd-pack-256-p ,object)
-      (let ((n-tag (sb-xc:gensym "TAG")))
+      (let ((n-tag (gensym "TAG")))
         `(and
           (simd-pack-256-p ,object)
           (let ((,n-tag (%simd-pack-256-tag ,object)))
@@ -854,6 +860,19 @@
 ;;; object's layout can ever be EQ to that of the ancestor.
 ;;; e.g. a fixnum as representative of class REAL.
 ;;; So in actual practice, you can't make something that is a pure STREAM, etc.
+
+;;; TODOs:
+;;; 1. There is an additional tweak that can potentially return false in one fewer
+;;;    conditional branch if the layout being tested has depthoid 8 (or 9 if #+64-bit).
+;;;    In that scenario, if the ID word of the candidate structure's layout does not
+;;;    exist, then it's the 0th bitmap word and safe to read always. Therefore
+;;;    STRUCTURE-IS-A and depthoid can be tested in that order. If there is no ID match,
+;;;    there's no depthoid test. If there is an ID match, it's the same as before.
+;;; 2. Since all backends implement STRUCTURE-IS-A, is there any reason that the
+;;;    depthoid test is in the transform's expansion and not baked into that vop?
+;;;    Putting it in the vop could be better for some backends,
+;;;    and would eliminate the ad-hoc LAYOUT-DEPTHOID-GE vop.
+
 #-(or x86 x86-64) ; vop-translated for these 2
 (defmacro layout-depthoid-ge (layout depthoid)
   `(>= (wrapper-depthoid ,layout) ,depthoid))
@@ -1039,7 +1058,7 @@
 (defun source-transform-typep (object type)
   (when (typep type 'type-specifier)
     (check-deprecated-type type))
-  (let ((name (sb-xc:gensym "OBJECT")))
+  (let ((name (gensym "OBJECT")))
     (multiple-value-bind (transform error)
         (%source-transform-typep name type)
       (if error
@@ -1378,7 +1397,7 @@
         tval)))))
 
 (deftransform #+64-bit unsigned-byte-64-p #-64-bit unsigned-byte-32-p
-  ((value) (fixnum) * :important nil)
+  ((value) (sb-vm:signed-word) * :important nil)
   `(>= value 0))
 
 (deftransform %other-pointer-p ((object))
