@@ -40,13 +40,12 @@
 #include "validate.h"
 #include "gc-internal.h"
 #include "gc-private.h"
-#include "getallocptr.h"
 #include "code.h"
 
 #include <errno.h>
 
 #ifdef LISP_FEATURE_SB_CORE_COMPRESSION
-# include <zlib.h>
+# include <zstd.h>
 #endif
 
 /* build_id must match between the C code and .core file because a core
@@ -170,74 +169,53 @@ lose:
 
 #ifndef LISP_FEATURE_SB_CORE_COMPRESSION
 # define inflate_core_bytes(fd,offset,addr,len) \
-    lose("This runtime was not built with zlib-compressed core support... aborting")
+    lose("This runtime was not built with zstd-compressed core support... aborting")
 #else
-# define ZLIB_BUFFER_SIZE (1u<<16)
 static void inflate_core_bytes(int fd, os_vm_offset_t offset,
                                os_vm_address_t addr, int len)
 {
-    z_stream stream;
-    unsigned char* buf = successful_malloc(ZLIB_BUFFER_SIZE);
-    int ret;
-
 # ifdef LISP_FEATURE_WIN32
-    /* Ensure the memory is committed so zlib doesn't segfault trying to
-       inflate. */
+    /* Ensure the memory is committed so zstd doesn't segfault trying
+       to inflate. */
     os_commit_memory(addr, len);
 # endif
-
-    if (-1 == lseek(fd, offset, SEEK_SET)) {
+    if (-1 == lseek(fd, offset, SEEK_SET))
         lose("Unable to lseek() on corefile");
-    }
 
-    stream.zalloc = NULL;
-    stream.zfree = NULL;
-    stream.opaque = NULL;
-    stream.avail_in = 0;
-    stream.next_in = buf;
+    int ret;
+    size_t buf_size = ZSTD_DStreamInSize();
+    unsigned char* buf = successful_malloc(buf_size);
+    ZSTD_inBuffer input;
+    input.src = buf;
+    input.pos = 0;
 
-    ret = inflateInit(&stream);
-    if (ret != Z_OK)
-        lose("zlib error %i", ret);
+    ZSTD_outBuffer output;
+    output.dst = (void*)addr;
+    output.size = len;
+    output.pos = 0;
 
-    stream.next_out  = (void*)addr;
-    stream.avail_out = len;
+    ZSTD_DStream *stream = ZSTD_createDStream();
+    if (stream == NULL)
+        lose("unable to create zstd decompression context");
+    ret = ZSTD_initDStream(stream);
+    if (ZSTD_isError(ret))
+        lose("ZSTD_initDStream failed with error: %s", ZSTD_getErrorName(ret));
+
+    /* Read in exactly one frame. */
     do {
-        ssize_t count = read(fd, buf, ZLIB_BUFFER_SIZE);
+        ssize_t count = read(fd, buf, buf_size);
         if (count < 0)
             lose("unable to read core file (errno = %i)", errno);
-        stream.next_in = buf;
-        stream.avail_in = count;
-        if (count == 0) break;
-        ret = inflate(&stream, Z_NO_FLUSH);
-        switch (ret) {
-        case Z_STREAM_END:
-            break;
-        case Z_OK:
-            if (stream.avail_out == 0)
-                lose("Runaway gzipped core directory... aborting");
-            if (stream.avail_in > 0)
-                lose("zlib inflate returned without fully"
-                     "using up input buffer... aborting");
-            break;
-        default:
-            lose("zlib inflate error: %i", ret);
-            break;
-        }
-    } while (ret != Z_STREAM_END);
+        input.size = count;
+        input.pos = 0;
+        ret = ZSTD_decompressStream(stream, &output, &input);
+        if (ZSTD_isError(ret))
+            lose("ZSTD_decompressStream failed with error: %s",
+                 ZSTD_getErrorName(ret));
+    } while (ret != 0);
 
-    if (stream.avail_out > 0) {
-        if (stream.avail_out >= os_vm_page_size)
-            fprintf(stderr, "Warning: gzipped core directory significantly"
-                    "shorter than expected (%lu bytes)", (unsigned long)stream.avail_out);
-        /* Is this needed? */
-        memset(stream.next_out, 0, stream.avail_out);
-    }
-
-    inflateEnd(&stream);
-    free(buf);
+    ZSTD_freeDStream(stream);
 }
-# undef ZLIB_BUFFER_SIZE
 #endif
 
 #define DYNAMIC_SPACE_ADJ_INDEX 0
@@ -429,6 +407,20 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
             for (i=0; i<(nwords-1); ++i)
                 if (bitmap_logbitp(i, bitmap)) adjust_pointers(slots+i, 1, adj);
             continue;
+#ifdef LISP_FEATURE_COMPACT_SYMBOL
+          case SYMBOL_WIDETAG:
+            { // Copied from scav_symbol() in gc-common
+            struct symbol* s = (void*)where;
+            adjust_pointers(&s->value, 2, adj);
+            lispobj name = decode_symbol_name(s->name);
+            lispobj adjusted_name = adjust_word(adj, name);
+            // writeback the name if it changed
+            if (adjusted_name != name) set_symbol_name(s, adjusted_name);
+            int indicated_nwords = (*where>>N_WIDETAG_BITS) & 0xFF;
+            adjust_pointers(&s->fdefn, indicated_nwords - 4, adj);
+            }
+            continue;
+#endif
         case FDEFN_WIDETAG:
             adjust_pointers(where+1, 2, adj);
             // For most architectures, 'raw_addr' doesn't satisfy is_lisp_pointer()
@@ -438,12 +430,21 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
             continue;
         case CODE_HEADER_WIDETAG:
             if (filler_obj_p(where)) {
+                // OMGWTF! Why does a filler code object merit adjustment?
                 if (where[2]) adjust_word_at(where+2, adj);
                 continue;
             }
             // Fixup the constant pool. The word at where+1 is a fixnum.
             code = (struct code*)where;
             adjust_pointers(where+2, code_header_words(code)-2, adj);
+#ifdef LISP_FEATURE_UNTAGGED_FDEFNS
+            // Process each untagged fdefn pointer.
+            lispobj* fdefns_start = code->constants + code_n_funs(code)
+              * CODE_SLOTS_PER_SIMPLE_FUN;
+            int i;
+            for (i=code_n_named_calls(code)-1; i>=0; --i)
+                adjust_word_at(fdefns_start+i, adj);
+#endif
 #if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64 || \
     defined LISP_FEATURE_PPC || defined LISP_FEATURE_PPC64
             // Fixup absolute jump table
@@ -473,8 +474,8 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
             continue;
         case CLOSURE_WIDETAG:
             fix_fun_header_layout(where, adj);
-#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
-            // For x86[-64], the closure fun appears to be a fixnum,
+#if FUN_SELF_FIXNUM_TAGGED
+            // For x86[-64], arm64, the closure fun appears to be a fixnum,
             // and might need adjustment unless pointing to immobile code.
             // Then fall into the general case; where[1] won't get re-adjusted
             // because it doesn't satisfy is_lisp_pointer().
@@ -522,7 +523,9 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         case COMPLEX_VECTOR_WIDETAG:
         case COMPLEX_ARRAY_WIDETAG:
         // And the rest of the purely descriptor objects.
+#ifndef LISP_FEATURE_COMPACT_SYMBOL
         case SYMBOL_WIDETAG:
+#endif
         case VALUE_CELL_WIDETAG:
         case WEAK_POINTER_WIDETAG:
         case RATIO_WIDETAG:
@@ -551,6 +554,7 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
 #ifdef SIMD_PACK_256_WIDETAG
         case SIMD_PACK_256_WIDETAG:
 #endif
+        case FILLER_WIDETAG: // non-card-spanning object pages container fillers
             continue;
         default:
           if (other_immediate_lowtag_p(widetag)
@@ -562,8 +566,8 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         adjust_pointers(where+1, nwords-1, adj);
     }
 #if SHOW_SPACE_RELOCATION
-    fprintf(stderr, "space @ %p: fixed %d absolute + %d relative pointers\n",
-            (lispobj*)start, adj->n_relocs_abs, adj->n_relocs_rel);
+    fprintf(stderr, "space @ %p..%p: fixed %d absolute + %d relative pointers\n",
+            (lispobj*)start, end, adj->n_relocs_abs, adj->n_relocs_rel);
 #endif
 }
 
@@ -579,6 +583,7 @@ static void relocate_heap(struct heap_adjust* adj)
                         (char*)adj->range[i].start + adj->range[i].delta,
                         (char*)adj->range[i].end + adj->range[i].delta);
     }
+    relocate_space(NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, adj);
     relocate_space(STATIC_SPACE_OBJECTS_START, static_space_free_pointer, adj);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     relocate_space(FIXEDOBJ_SPACE_START, fixedobj_free_pointer, adj);
@@ -586,7 +591,8 @@ static void relocate_heap(struct heap_adjust* adj)
 #ifdef LISP_FEATURE_CHENEYGC
     relocate_space(DYNAMIC_0_SPACE_START, (lispobj*)get_alloc_pointer(), adj);
 #else
-    relocate_space(DYNAMIC_SPACE_START, (lispobj*)get_alloc_pointer(), adj);
+    relocate_space(DYNAMIC_SPACE_START, (lispobj*)dynamic_space_highwatermark(),
+                   adj);
 #endif
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     // Pointers within varyobj space to varyobj space do not need adjustment
@@ -635,7 +641,7 @@ void calc_asm_routine_bounds()
         asm_routines_start = READ_ONLY_SPACE_START;
     } else {
         lispobj *where = (lispobj*)STATIC_SPACE_OBJECTS_START;
-        for (; where < static_space_free_pointer; where += OBJECT_SIZE(*where, where))
+        for (; where < static_space_free_pointer; where += object_size(where))
             if (widetag_of((lispobj*)where) == CODE_HEADER_WIDETAG) {
                 asm_routines_start = (uword_t)where;
                 break;
@@ -676,6 +682,21 @@ void calc_immobile_space_bounds()
 }
 #endif
 
+__attribute__((unused)) static void check_dynamic_space_addr_ok(uword_t start, uword_t size)
+{
+#ifdef LISP_FEATURE_64_BIT // don't want a -Woverflow warning on 32-bit
+    uword_t end_word_addr = start + size - N_WORD_BYTES;
+    // Word-aligned pointers can't address more than 48 significant bits for now.
+    // If you want to lift that restriction, look at how SYMBOL-PACKAGE and
+    // SYMBOL-NAME are combined into one lispword.
+    uword_t unaddressable_bits = 0xFFFF000000000000;
+    if ((start & unaddressable_bits) || (end_word_addr & unaddressable_bits))
+        lose("Panic! This version of SBCL can not address memory\n"
+             "in the range %p:%p given by the OS.\nPlease report this as a bug.",
+             (void*)start, (void*)(start + size));
+#endif
+}
+
 /* TODO: If static + readonly were mapped as desired without disabling ASLR
  * but one of the large spaces couldn't be mapped as desired, start over from
  * the top, disabling ASLR. This should help to avoid relocating the heap
@@ -880,12 +901,13 @@ process_directory(int count, struct ndir_entry *entry,
                 }
 #else /* gencgc */
                 {
-                uword_t aligned_start = ALIGN_UP(addr, GENCGC_CARD_BYTES);
+                uword_t aligned_start = ALIGN_UP(addr, GENCGC_PAGE_BYTES);
                 /* Misalignment can happen only if card size exceeds OS page.
                  * Drop one card to avoid overrunning the allocated space */
                 if (aligned_start > addr) // not card-aligned
-                    dynamic_space_size -= GENCGC_CARD_BYTES;
+                    dynamic_space_size -= GENCGC_PAGE_BYTES;
                 DYNAMIC_SPACE_START = addr = aligned_start;
+                check_dynamic_space_addr_ok(addr, dynamic_space_size);
                 }
 #endif
                 break;
@@ -937,9 +959,10 @@ process_directory(int count, struct ndir_entry *entry,
              * at the time the core was saved.
              * For gencgc this is #defined as DYNAMIC_SPACE_START */
             current_dynamic_space = (lispobj *)addr;
+#else
+            next_free_page = ALIGN_UP(entry->nwords<<WORD_SHIFT, GENCGC_PAGE_BYTES)
+              / GENCGC_PAGE_BYTES;
 #endif
-            set_alloc_pointer((lispobj)free_pointer);
-
             anon_dynamic_space_start = (os_vm_address_t)(addr + len);
         }
     }
@@ -1103,7 +1126,7 @@ char* get_asm_routine_by_name(const char* name, int *index)
         for (i=2 ; i < vector_len(table) ; i += 2)
             if (lowtag_of(sym = table->data[i]) == OTHER_POINTER_LOWTAG
                 && widetag_of(&SYMBOL(sym)->header) == SYMBOL_WIDETAG
-                && !strcmp(name, (char*)(VECTOR(SYMBOL(sym)->name)->data))) {
+                && !strcmp(name, (char*)(symbol_name(SYMBOL(sym))->data))) {
                 lispobj value = table->data[i+1];
                 // value = (start-address . (end-address . index))
                 if (index)
@@ -1126,11 +1149,7 @@ void asm_routine_poke(const char* routine, int offset, char byte)
 }
 
 // Caution: use at your own risk
-#undef DEBUG_CORE_LOADING
-#ifdef DEBUG_CORE_LOADING
-#ifdef LISP_FEATURE_CHENEYGC
-#  error "Can't define DEBUG_CORE_LOADING for cheneygc"
-#endif
+#if defined DEBUG_CORE_LOADING && DEBUG_CORE_LOADING
 #include "hopscotch.h"
 #include "genesis/cons.h"
 #include "genesis/layout.h"
@@ -1148,6 +1167,7 @@ struct visitor {
     struct hopscotch_table *reached;
 };
 
+static void trace_sym(lispobj, struct symbol*, struct hopscotch_table*);
 #define RECURSE(x) if(is_lisp_pointer(x))graph_visit(ptr,x,seen)
 static void graph_visit(lispobj __attribute__((unused)) referer,
                         lispobj ptr,
@@ -1176,7 +1196,7 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
         case FUNCALLABLE_INSTANCE_WIDETAG:
             layout = layout_of(obj);
             graph_visit(ptr, layout, seen);
-            nwords = sizetab[widetag_of(obj)](obj);
+            nwords = headerobj_size(obj);
             struct bitmap bitmap = get_layout_bitmap(LAYOUT(layout));
             for (i=0; i<(nwords-1); ++i)
                 if (bitmap_logbitp(i, bitmap)) RECURSE(obj[1+i]);
@@ -1200,6 +1220,8 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
             for(i=2; i<=nwords; ++i) RECURSE(obj[i]);
             break;
         case SYMBOL_WIDETAG:
+            trace_sym(ptr, SYMBOL(ptr), seen);
+            break;
         case WEAK_POINTER_WIDETAG:
             nwords = TINY_BOXED_NWORDS(*obj);
             for(i=1; i<=nwords; ++i) RECURSE(obj[i]);
@@ -1211,11 +1233,21 @@ static void graph_visit(lispobj __attribute__((unused)) referer,
             break;
         default:
             if (!leaf_obj_widetag_p(widetag_of(obj))) {
-                int size = sizetab[widetag_of(obj)](obj);
+                int size = headerobj_size(obj);
                 for(i=1; i<size; ++i) RECURSE(obj[i]);
             }
       }
 }
+static void trace_sym(lispobj ptr, struct symbol* sym, struct hopscotch_table* seen)
+{
+    RECURSE(decode_symbol_name(sym->name));
+    RECURSE(sym->value);
+    RECURSE(sym->info);
+    RECURSE(sym->fdefn);
+    int indicated_nwords = HeaderValue(sym->header) & 0xFF;
+    if (indicated_nwords + 1 > SYMBOL_SIZE) // has one more slot with no name
+        RECURSE(1[&sym->fdefn]); // ASSUMPTION: slot order
+}
 
 static void tally(lispobj ptr, struct visitor* v)
 {
@@ -1224,9 +1256,9 @@ static void tally(lispobj ptr, struct visitor* v)
     else {
         lispobj* obj = native_pointer(ptr);
         lispobj header = *obj;
+        int words = object_size2(obj, header);
         int widetag = header_widetag(header);
         int header_index = widetag>>2;
-        int words = OBJECT_SIZE(header, obj);
         ++v->headers[header_index].count;
         v->headers[header_index].words += words;
         if (widetag == SIMPLE_VECTOR_WIDETAG) {
@@ -1241,24 +1273,39 @@ static void tally(lispobj ptr, struct visitor* v)
     }
 }
 
+/* This printing in here is useful, but it's too much to output in make-target-2,
+ * because genesis dumps a ton of unreachable objects.
+ * The reason is this: cold-load has no way of knowing if a literal loaded
+ * from fasl and written to core is really supposed to be consumed by target.
+ * e.g. if it's a string naming a foreign fixup, who reads that string?
+ * Answer: The host, but it appears in the cold core as if the target will need it.
+ * The only way to rectify this defect is just not worth doing - each literal
+ * would have to be kept only as a host proxy until such time as we actually refer
+ * to it from another object that is definitely in the cold core, via FOP-LOAD-CODE
+ * and who know what else. Thus it remains a graph tracing problem in nature,
+ * which is best left to GC */
 static uword_t visit(lispobj* where, lispobj* limit, uword_t arg)
 {
     struct visitor* v = (struct visitor*)arg;
     lispobj* obj = where;
     while (obj < limit) {
+        if (widetag_of(obj) == FILLER_WIDETAG) {
+            obj += object_size(obj);
+            continue;
+        }
         lispobj ptr = compute_lispobj(obj);
         tally(ptr, v);
-        // Weird stuff happens in genesis where this logic thinks that the
-        // cold core holds a ton of unreachable objects. I have no idea why.
-        if (!hopscotch_get(v->reached, ptr, 0))
-            printf("object not reached: %p\n", (void*)ptr);
-        obj += OBJECT_SIZE(*obj, obj);
+        if (!hopscotch_get(v->reached, ptr, 0)) printf("unreachable: %p\n", (void*)ptr);
+        obj += object_size(obj);
     }
     return 0;
 }
 
 #ifdef LISP_FEATURE_GENCGC
-#define dynamic_space_pointer_p(ptr) (find_page_index((void*)ptr) >= 0)
+#define count_this_pointer_p(ptr) (find_page_index((void*)ptr) >= 0)
+#endif
+#ifdef LISP_FEATURE_CHENEYGC
+#define count_this_pointer_p(ptr) (1)
 #endif
 
 static void sanity_check_loaded_core(lispobj initial_function)
@@ -1272,13 +1319,12 @@ static void sanity_check_loaded_core(lispobj initial_function)
                      1<<18, /* initial size */
                      0);
     {
+      trace_sym(NIL, SYMBOL(NIL), &reached);
       lispobj* where = (lispobj*)STATIC_SPACE_OBJECTS_START;
       lispobj* end = static_space_free_pointer;
       while (where<end) {
-        // This falsely treats NIL as 4 conses but it doesn't really matter.
-        // The garbage collectors do too.
         graph_visit(0, compute_lispobj(where), &reached);
-        where += OBJECT_SIZE(*where, where);
+        where += object_size(where);
       }
     }
     graph_visit(0, initial_function, &reached); // not otherwise reachable
@@ -1286,11 +1332,17 @@ static void sanity_check_loaded_core(lispobj initial_function)
     int key_index;
     lispobj ptr;
     for_each_hopscotch_key(key_index, ptr, reached)
-        if (dynamic_space_pointer_p(ptr))
-            tally(ptr, &v[0]);
+        if (count_this_pointer_p(ptr)) tally(ptr, &v[0]);
     // Pass 2: Count all heap objects
     v[1].reached = &reached;
+#ifdef LISP_FEATURE_GENCGC
     walk_generation(visit, -1, (uword_t)&v[1]);
+#endif
+#ifdef LISP_FEATURE_CHENEYGC
+    visit((lispobj*)READ_ONLY_SPACE_START, read_only_space_free_pointer, (uword_t)&v[1]);
+    visit((lispobj*)STATIC_SPACE_START, static_space_free_pointer, (uword_t)&v[1]);
+#endif
+
     // Pass 3: Compare
     // Start with the conses
     v[0].headers[0].words = v[0].headers[0].count * 2;
@@ -1306,7 +1358,6 @@ static void sanity_check_loaded_core(lispobj initial_function)
             ((strncmp(widetag_names[i], "unk", 3)
               && (i != CHARACTER_WIDETAG>>2)
               && (i != SIMPLE_FUN_WIDETAG>>2)
-              && (i != NO_TLS_VALUE_MARKER_WIDETAG>>2)
               && (i != UNBOUND_MARKER_WIDETAG>>2)))) {
             int mismatch = v[0].headers[i].count != v[1].headers[i].count;
             printf("%8d %11d  | %8d %11d   | %s%s\n",
