@@ -52,6 +52,17 @@
 #define BIGNUM_MARK_BIT MARK_BIT
 #endif
 
+extern lispobj lisp_init_function, gc_object_watcher;
+
+static inline uword_t boxedobj_widetag_to_markbit(int widetag) {
+    return (widetag == FDEFN_WIDETAG || widetag == WEAK_POINTER_WIDETAG)
+         ? FDEFN_MARK_BIT : MARK_BIT;
+}
+static inline uword_t general_widetag_to_markbit(int widetag) {
+    return (widetag == FDEFN_WIDETAG || widetag == WEAK_POINTER_WIDETAG)
+        ? FDEFN_MARK_BIT : ((widetag == BIGNUM_WIDETAG) ? BIGNUM_MARK_BIT : MARK_BIT);
+}
+
 #define interesting_pointer_p(x) \
   (find_page_index((void*)x) >= 0 || immobile_space_p(x))
 
@@ -77,13 +88,16 @@ static page_index_t free_page;
  * If it collides with 'next_free_page', then you lose.
  * TODO: It would be reasonably simple to have this request more memory from
  * the OS instead of failing on overflow */
-static void* get_free_page() {
+static void* get_free_page(boolean prezero) {
     --free_page;
     if (free_page < next_free_page)
         lose("Needed more space to GC");
-    page_table[free_page].type = UNBOXED_PAGE_FLAG;
+    page_table[free_page].type = PAGE_TYPE_UNBOXED;
     char* mem = page_address(free_page);
-    zero_dirty_pages(free_page, free_page, 0);
+#ifdef LISP_FEATURE_DARWIN_JIT
+    prezero = 1; // Might need to alter MMU-based protection
+#endif
+    if (prezero) zeroize_pages_if_needed(free_page, free_page, 0);
     return mem;
 }
 
@@ -92,18 +106,23 @@ static void* get_free_page() {
  * then one GC page can hold 2K cons cells.
  * One byte marks 8 conses (1 bit per cons), 256 bytes mark 2048 conses.
  * 128 blocks of 256 bytes fit on a 32K GC page. */
-char *suballocator_free_ptr, *suballocator_end_ptr;
+static char *suballocator_free_ptr, *suballocator_end_ptr;
 
+#ifndef LISP_FEATURE_USE_CONS_REGION
+static inline uword_t compute_page_key(lispobj cons) {
+    return ALIGN_DOWN(cons, GENCGC_PAGE_BYTES);
+}
 static void* allocate_cons_mark_bits() {
-    int nbytes = GENCGC_CARD_BYTES / (2 * N_WORD_BYTES) / 8;
+    int nbytes = GENCGC_PAGE_BYTES / (2 * N_WORD_BYTES) / 8;
     if (suballocator_free_ptr + nbytes > suballocator_end_ptr) {
-        suballocator_free_ptr = get_free_page();
-        suballocator_end_ptr = suballocator_free_ptr + GENCGC_CARD_BYTES;
+        suballocator_free_ptr = get_free_page(1);
+        suballocator_end_ptr = suballocator_free_ptr + GENCGC_PAGE_BYTES;
     }
-    void* mem = suballocator_free_ptr;
+    char* mem = suballocator_free_ptr;
     suballocator_free_ptr += nbytes;
     return mem;
 }
+#endif
 
 static void gc_enqueue(lispobj object)
 {
@@ -114,13 +133,14 @@ static void gc_enqueue(lispobj object)
         next = scav_queue.recycler;
         if (next) {
             scav_queue.recycler = next->next;
-            next->next = 0;
             dprintf(("Popped recycle list\n"));
         } else {
-            next = (struct Qblock*)get_free_page();
+            next = (struct Qblock*)get_free_page(0);
             dprintf(("Alloc'd new block\n"));
         }
         block = block->next = next;
+        block->next = 0;
+        block->tail = block->count = 0;
         scav_queue.tail_block = block;
     }
     block->elements[block->tail] = object;
@@ -152,18 +172,20 @@ static lispobj gc_dequeue()
  * for headerless objects (conses) */
 struct hopscotch_table mark_bits;
 
-static inline uword_t compute_page_key(lispobj cons) {
-    return ALIGN_DOWN(cons, GENCGC_CARD_BYTES);
-}
 static inline int compute_dword_number(lispobj cons) {
-    return (cons & (GENCGC_CARD_BYTES - 1)) >> (1+WORD_SHIFT);
+    return (cons & (GENCGC_PAGE_BYTES - 1)) >> (1+WORD_SHIFT);
 }
 
 static inline int cons_markedp(lispobj pointer) {
+    unsigned int index = compute_dword_number(pointer);
+#ifdef LISP_FEATURE_USE_CONS_REGION /* all conses must be on cons pages */
+    unsigned char* bits = (unsigned char*)ALIGN_DOWN(pointer, GENCGC_PAGE_BYTES)
+                          + CONS_PAGE_USABLE_BYTES;
+#else
     unsigned char* bits = (unsigned char*)
         hopscotch_get(&mark_bits, compute_page_key(pointer), 0);
     if (!bits) return 0;
-    int index = compute_dword_number(pointer);
+#endif
     return (bits[index / 8] >> (index % 8)) & 1;
 }
 
@@ -178,13 +200,36 @@ static inline int pointer_survived_gc_yet(lispobj pointer)
     int widetag = header_widetag(header);
     switch (widetag) {
     case BIGNUM_WIDETAG: return (header & BIGNUM_MARK_BIT) != 0;
-    case FDEFN_WIDETAG : return (header & FDEFN_MARK_BIT) != 0;
+    case WEAK_POINTER_WIDETAG:
+    case FDEFN_WIDETAG: return (header & FDEFN_MARK_BIT) != 0;
     }
     if (embedded_obj_p(widetag))
         header = *fun_code_header(native_pointer(pointer));
     return (header & MARK_BIT) != 0;
 }
 
+extern page_index_t contiguous_block_final_page(page_index_t);
+void dump_marked_objects() {
+    fprintf(stderr, "Marked objects:\n");
+    page_index_t first = 0;
+    int n = 0;
+    while (first < next_free_page) {
+        page_index_t last = contiguous_block_final_page(first);
+        lispobj* where = (lispobj*)page_address(first);
+        lispobj* limit = (lispobj*)page_address(last) + page_words_used(last);
+        while (where < limit) {
+            lispobj obj = compute_lispobj(where);
+            if (pointer_survived_gc_yet(obj)) {
+                ++n;
+                fprintf(stderr, " %"OBJ_FMTX"\n", obj);
+            }
+            where += object_size(where);
+        }
+        first = 1 + last;
+    }
+    fprintf(stderr, "Total: %d\n", n);
+}
+
 void __mark_obj(lispobj pointer)
 {
     gc_dcheck(is_lisp_pointer(pointer));
@@ -203,7 +248,7 @@ void __mark_obj(lispobj pointer)
                 pointer = make_lispobj(base, OTHER_POINTER_LOWTAG);
                 header = *base;
             }
-            uword_t markbit = (widetag == FDEFN_WIDETAG) ? FDEFN_MARK_BIT : MARK_BIT;
+            uword_t markbit = boxedobj_widetag_to_markbit(widetag);
             if (header & markbit) return; // already marked
             *base |= markbit;
         }
@@ -219,15 +264,23 @@ void __mark_obj(lispobj pointer)
 #endif
         if (leaf_obj_widetag_p(widetag)) return;
     } else {
+        unsigned int index = compute_dword_number(pointer);
+#ifdef LISP_FEATURE_USE_CONS_REGION /* all conses must be on cons pages */
+        unsigned char* bits = (unsigned char*)ALIGN_DOWN(pointer, GENCGC_PAGE_BYTES)
+                              + CONS_PAGE_USABLE_BYTES;
+#else
+        /* Only _some_ conses are on cons pages depending if they got moved there.
+         * We could use a hybrid of storing mark bits on the page if possible,
+         * or the hash-table if not. However, I'd rather not complicate things.
+         * All architectures should be made to allocate directly on cons pages! */
         uword_t key = compute_page_key(pointer);
-        int index = compute_dword_number(pointer);
         unsigned char* bits = (unsigned char*)hopscotch_get(&mark_bits, key, 0);
         if (!bits) {
             bits = allocate_cons_mark_bits();
             hopscotch_insert(&mark_bits, key, (sword_t)bits);
-        } else if (bits[index / 8] & (1 << (index % 8))) {
-            return;
         }
+#endif
+        if (bits[index / 8] & (1 << (index % 8))) return;
         // Mark the cons
         bits[index / 8] |= 1 << (index % 8);
     }
@@ -283,14 +336,9 @@ static void trace_object(lispobj* where)
     lispobj header = *where;
     int widetag = header_widetag(header);
 
-    switch (widetag) {
-    case INSTANCE_WIDETAG:
-        return trace_using_layout(instance_layout(where),
-                                  where, instance_length(header));
-    case FUNCALLABLE_INSTANCE_WIDETAG:
-        return trace_using_layout(funinstance_layout(where),
-                                  where, HeaderValue(header) & SHORT_HEADER_MAX_WORDS);
-    }
+    if (instanceoid_widetag_p(widetag))
+        return trace_using_layout(layout_of(where), where,
+                                  instanceoid_length(header));
     sword_t scan_from = 1;
     sword_t scan_to = sizetab[widetag](where);
     sword_t i;
@@ -332,7 +380,7 @@ static void trace_object(lispobj* where)
             return;
         }
         break;
-#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
+#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64) || defined (LISP_FEATURE_ARM64)
     /* on x86[-64], closure->fun is a fixnum-qua-pointer. Convert it to a lisp
      * pointer to mark it, but not on platforms where it's already a descriptor */
     case CLOSURE_WIDETAG:
@@ -358,6 +406,19 @@ static void trace_object(lispobj* where)
         return;
 #endif
         break;
+#ifdef LISP_FEATURE_COMPACT_SYMBOL
+    case SYMBOL_WIDETAG:
+        {
+        struct symbol* s = (void*)where;
+        gc_mark_obj(decode_symbol_name(s->name));
+        gc_mark_obj(s->value);
+        gc_mark_obj(s->info);
+        gc_mark_obj(s->fdefn);
+        // process the unnamed slot of augmented symbols
+        if ((s->header & 0xFF00) == (SYMBOL_SIZE<<8)) gc_mark_obj(*(1+&s->fdefn));
+        }
+        return;
+#endif
     case FDEFN_WIDETAG:
         gc_mark_obj(fdefn_callee_lispobj((struct fdefn*)where));
         scan_to = 3;
@@ -376,24 +437,22 @@ static void trace_object(lispobj* where)
 
 void prepare_for_full_mark_phase()
 {
-    // FIXME: Estimate how large to create mark_bits based on dynamic space size.
-    // Guess 8 words per object, and X% of the objects are conses.
-    // The problem is guessing how localized the conses are: guess that N conses
-    // will reside on fraction*N different pages, which guides us as to how many
-    // hash table entries are needed.
+#ifndef LISP_FEATURE_USE_CONS_REGION
     hopscotch_create(&mark_bits, HOPSCOTCH_HASH_FUN_DEFAULT,
                      N_WORD_BYTES, /* table values are machine words */
                      65536, /* initial size */
                      0);
+#endif
 
     free_page = page_table_pages;
     suballocator_free_ptr = suballocator_end_ptr = 0;
-    struct Qblock* block = (struct Qblock*)get_free_page();
+    struct Qblock* block = (struct Qblock*)get_free_page(0);
     dprintf(("Queue block holds %d objects\n", (int)QBLOCK_CAPACITY));
     scav_queue.head_block = block;
     scav_queue.tail_block = block;
     scav_queue.recycler   = 0;
-    gc_assert(!scav_queue.head_block->count);
+    block->next = 0;
+    block->tail = block->count = 0;
 }
 
 void execute_full_mark_phase()
@@ -402,12 +461,13 @@ void execute_full_mark_phase()
     struct rusage before, after;
     getrusage(RUSAGE_SELF, &before);
 #endif
+    trace_object((lispobj*)NIL_SYMBOL_SLOTS_START);
     lispobj* where = (lispobj*)STATIC_SPACE_OBJECTS_START;
     lispobj* end = static_space_free_pointer;
     while (where < end) {
         lispobj obj = compute_lispobj(where);
         gc_enqueue(obj);
-        where += listp(obj) ? 2 : sizetab[widetag_of(where)](where);
+        where += listp(obj) ? 2 : headerobj_size(where);
     }
 #ifdef LISP_FEATURE_METASPACE
     where = (lispobj*)METASPACE_START;
@@ -415,9 +475,13 @@ void execute_full_mark_phase()
     while (where < end) {
         lispobj obj = compute_lispobj(where);
         gc_enqueue(obj);
-        where += listp(obj) ? 2 : sizetab[widetag_of(where)](where);
+        where += listp(obj) ? 2 : headerobj_size(where);
     }
 #endif
+    gc_mark_obj(lisp_package_vector);
+    gc_mark_obj(lisp_init_function);
+    gc_mark_obj(gc_object_watcher);
+    gc_mark_obj(alloc_profile_data);
     do {
         lispobj ptr = gc_dequeue();
         gc_dcheck(ptr != 0);
@@ -447,8 +511,8 @@ static void local_smash_weak_pointers()
     struct weak_pointer *wp, *next_wp;
     for (wp = weak_pointer_chain; wp != WEAK_POINTER_CHAIN_END; wp = next_wp) {
         gc_assert(widetag_of(&wp->header) == WEAK_POINTER_WIDETAG);
-        next_wp = wp->next;
-        wp->next = NULL;
+        next_wp = get_weak_pointer_next(wp);
+        reset_weak_pointer_next(wp);
         lispobj pointee = wp->value;
         gc_assert(is_lisp_pointer(pointee));
         if (!pointer_survived_gc_yet(pointee))
@@ -513,7 +577,7 @@ static void sweep_fixedobj_pages(long *zeroed)
         lispobj *limit = (lispobj*)((char*)obj + IMMOBILE_CARD_BYTES - obj_spacing);
         for ( ; obj <= limit ; obj = (lispobj*)((char*)obj + obj_spacing) ) {
             lispobj header = *obj;
-            uword_t markbit = (header_widetag(header) == FDEFN_WIDETAG) ? FDEFN_MARK_BIT : MARK_BIT;
+            uword_t markbit = boxedobj_widetag_to_markbit(header_widetag(header));
             if (fixnump(header)) { // is a hole
             } else if (header & markbit) { // live object
                 *obj = header ^ markbit;
@@ -528,60 +592,59 @@ static void sweep_fixedobj_pages(long *zeroed)
     lispobj* limit = fixedobj_page_address(FIXEDOBJ_RESERVED_PAGES);
     while (obj < limit) {
         lispobj header = *obj;
-        uword_t markbit = (header_widetag(header) == FDEFN_WIDETAG) ? FDEFN_MARK_BIT : MARK_BIT;
+        uword_t markbit = boxedobj_widetag_to_markbit(header_widetag(header));
         if (header & markbit) *obj = header ^ markbit;
-        obj += sizetab[widetag_of(obj)](obj);
+        obj += headerobj_size(obj);
     }
 }
 #endif
 
-static uword_t sweep(lispobj* where, lispobj* end, uword_t arg)
+/* Overwrite exactly 1 object wit non-pointer words of some sort.
+ * This eliminates tenured garbage in pseudo-static-generation,
+ * and does NOT strive to to write as few words as possible,
+ * unlike deposit_filler() which tries to be efficient */
+static void clobber_headered_object(lispobj* addr, sword_t nwords)
+{
+    // FIXME: clobbering an object on single-object pages should free entire pages
+    page_index_t page = find_page_index(addr);
+    if (page < 0) { // code space
+        struct code* code  = (struct code*)addr;
+        if (!filler_obj_p((lispobj*)code)) {
+            code->boxed_size = 0;
+            code->header = (nwords << CODE_HEADER_SIZE_SHIFT)
+                           | CODE_HEADER_WIDETAG;
+            memset(addr+2, 0, (nwords - 2) * N_WORD_BYTES);
+        }
+    } else if ((SINGLE_OBJECT_FLAG|page_table[page].type) == (SINGLE_OBJECT_FLAG|PAGE_TYPE_CODE)) {
+        // Code pages don't want (0 . 0) fillers, otherwise heap checking
+        // gets an error: "object @ 0x..... is non-code on code page"
+        addr[0] = make_filler_header(nwords);
+        addr[1] = 0;
+    } else {
+        memset(addr, 0, nwords * N_WORD_BYTES);
+    }
+}
+
+static uword_t sweep(lispobj* where, lispobj* end,
+                     __attribute__((unused)) uword_t arg)
 {
-    long *zeroed = (long*)arg; // one count per generation
     sword_t nwords;
 
-    // TODO: consecutive dead objects on same page should be merged.
     for ( ; where < end ; where += nwords ) {
         lispobj word = *where;
+        int livep = 0;
         if (is_header(word)) {
-            nwords = sizetab[header_widetag(word)](where);
-            lispobj markbit = MARK_BIT;
-            switch (header_widetag(word)) {
-            case BIGNUM_WIDETAG: markbit = BIGNUM_MARK_BIT; break;
-            case FDEFN_WIDETAG : markbit = FDEFN_MARK_BIT; break;
-            }
-            if (word & markbit)
-                *where = word ^ markbit;
-            else {
-                // Turn the object into either a (0 . 0) cons
-                // or an unboxed filler depending on size.
-                if (nwords <= 2) // could be SAP, SIMPLE-ARRAY-NIL, 1-word bignum, etc
-                    goto cons;
-                struct code* code  = (struct code*)where;
-                // Keep in sync with the definition of filler_obj_p()
-                if (!filler_obj_p((lispobj*)code)) {
-                    page_index_t page = find_page_index(where);
-                    int gen = page >= 0 ? page_table[page].gen
-                      : immobile_obj_gen_bits(where);
-                    NOTE_GARBAGE(gen, where, nwords, zeroed, {
-                        code->boxed_size = 0;
-                        code->header = (nwords << CODE_HEADER_SIZE_SHIFT)
-                                     | CODE_HEADER_WIDETAG;
-                        memset(where+2, 0, (nwords - 2) * N_WORD_BYTES);
-                    })
-                }
-            }
+            nwords = headerobj_size2(where, word);
+            lispobj markbit = general_widetag_to_markbit(header_widetag(word));
+            if (word & markbit) livep = 1, *where = word ^ markbit;
+            if (!livep) clobber_headered_object(where, nwords);
         } else {
             nwords = 2;
-            if (!cons_markedp((lispobj)where)) {
-                if (where[0] | where[1]) {
-               cons:
-                    gc_dcheck(!immobile_space_p((lispobj)where));
-                    NOTE_GARBAGE(page_table[find_page_index(where)].gen,
-                                 where, 2, zeroed,
-                                 where[0] = where[1] = 0);
-                }
-            }
+            livep = cons_markedp((lispobj)where);
+            /* This bit pattern indicates unambiguously that a cons cell
+             * has gotten clobbered. Using (0 . 0) leaves open to question
+             * whether the cons is actually live */
+            if (!livep) where[0] = where[1] = (uword_t)-1;
         }
     }
     return 0;
@@ -630,28 +693,28 @@ void execute_full_sweep_phase()
             fprintf(stderr, "%ld%s", words_zeroed[i], i?"+":"");
         fprintf(stderr, " words zeroed]\n");
     }
+    page_index_t page;
+#ifdef LISP_FEATURE_USE_CONS_REGION
+    // Clear all cons mark bits
+    for (page = 0; page < next_free_page; ++page)
+        if (page_table[page].type == PAGE_TYPE_CONS) {
+            char* base = page_address(page);
+            char* bits = base + CONS_PAGE_USABLE_BYTES;
+            char* end  = base + GENCGC_PAGE_BYTES;
+            memset(bits, 0, end-bits);
+        }
+#else
     hopscotch_destroy(&mark_bits);
+#endif
+
     if (sweeplog)
         fflush(sweeplog);
 
-#ifdef LISP_FEATURE_SOFT_CARD_MARKS
-    free_page = next_free_page;
-#else
-    page_index_t first_page, last_page;
-    for (first_page = 0; first_page < next_free_page; ++first_page)
-        if (PAGE_WRITEPROTECTED_P(first_page)
-            && protection_mode(first_page) == PHYSICAL) {
-            last_page = first_page;
-            while (PAGE_WRITEPROTECTED_P(last_page+1)
-                   && protection_mode(last_page+1) == PHYSICAL)
-                ++last_page;
-            os_protect(page_address(first_page),
-                       (last_page - first_page + 1) * GENCGC_CARD_BYTES,
-                       OS_VM_PROT_READ | OS_VM_PROT_EXECUTE);
-            first_page = last_page;
-        }
-#endif
-    while (free_page < page_table_pages) {
-        page_table[free_page++].type = FREE_PAGE_FLAG;
-    }
+    // Give back all private-use pages and indicate need-to-zero
+    for (page = free_page; page < page_table_pages; ++page) {
+        gc_assert((page_table[page].type & PAGE_TYPE_MASK) == PAGE_TYPE_UNBOXED);
+        gc_assert(!page_bytes_used(page));
+        set_page_needs_zerofill(page);
+        page_table[page].type = FREE_PAGE_FLAG;
+      }
 }
