@@ -34,11 +34,11 @@ gc_general_alloc(struct alloc_region* region, sword_t nbytes, int page_type)
     }
     return collector_alloc_fallback(region, nbytes, page_type);
 }
-lispobj copy_possibly_large_object(lispobj object, sword_t nwords,
+lispobj copy_potential_large_object(lispobj object, sword_t nwords,
                                    struct alloc_region*, int page_type);
 #else
 void *gc_general_alloc(void*,sword_t,int);
-lispobj copy_possibly_large_object(lispobj object, sword_t nwords,
+lispobj copy_potential_large_object(lispobj object, sword_t nwords,
                                    void*, int page_type);
 #endif
 
@@ -270,15 +270,14 @@ static inline struct bitmap get_layout_bitmap(struct layout* layout)
     struct bitmap bitmap;
     const int layout_id_vector_fixed_capacity = 7;
 #ifdef LISP_FEATURE_64_BIT
-    sword_t depthoid = layout->flags;
+    sword_t depthoid = layout->sw_flags;
     // Depthoid is stored in the upper 4 bytes of 'flags', as a fixnum.
     depthoid >>= (32 + N_FIXNUM_TAG_BITS);
     int extra_id_words =
       (depthoid > layout_id_vector_fixed_capacity) ?
       ALIGN_UP(depthoid - layout_id_vector_fixed_capacity, 2) / 2 : 0;
 #else
-    sword_t depthoid = layout->depthoid;
-    depthoid >>= N_FIXNUM_TAG_BITS;
+    sword_t depthoid = fixnum_value(layout->depthoid);
     int extra_id_words = (depthoid > layout_id_vector_fixed_capacity) ?
       depthoid - layout_id_vector_fixed_capacity : 0;
 #endif
@@ -445,13 +444,9 @@ static inline int instance_length(lispobj header)
     return (((unsigned int)header >> INSTANCE_LENGTH_SHIFT) & 0x3FFF) + extra;
 }
 
-// One bit differentiates FUNCALLABLE_INSTANCE_WIDETAG from INSTANCE_WIDETAG.
-// This is index of that bit.
-#ifdef LISP_FEATURE_64_BIT
-#  define FUNINSTANCE_SELECTOR_BIT_NUMBER 3
-#else
-#  define FUNINSTANCE_SELECTOR_BIT_NUMBER 2
-#endif
+// This is index of the bit that differentiates FUNCALLABLE_INSTANCE_WIDETAG
+// from INSTANCE_WIDETAG.
+#define FUNINSTANCE_SELECTOR_BIT_NUMBER 2
 static inline boolean instanceoid_widetag_p(unsigned char widetag) {
     return (widetag | (1<<FUNINSTANCE_SELECTOR_BIT_NUMBER)) == FUNCALLABLE_INSTANCE_WIDETAG;
 }
@@ -484,13 +479,15 @@ static inline int instanceoid_length(lispobj header) {
 #endif
 
 static inline int layout_depth2_id(struct layout* layout) {
-    int32_t* vector = (int32_t*)&layout->id_word0;
+    int32_t* vector = (int32_t*)&layout->uw_id_word0;
     return vector[0];
 }
-// Keep in sync with hardwired IDs in src/compiler/generic/genesis.lisp
+// Keep in sync with hardwired IDs in CHOOSE-LAYOUT-ID
+// in src/compiler/generic/layout-ids.lisp
 #define WRAPPER_LAYOUT_ID 2
 #define LAYOUT_LAYOUT_ID 3
 #define LFLIST_NODE_LAYOUT_ID 4
+#define BROTHERTREE_UNARY_NODE_LAYOUT_ID 5
 
 /// Return true if 'thing' is a layout.
 /// This predicate is careful, as is it used to verify heap invariants.
@@ -550,4 +547,30 @@ struct slab_header {
 static const int CONS_PAGE_USABLE_BYTES = MAX_CONSES_PER_PAGE*CONS_SIZE*N_WORD_BYTES;
 #endif
 
+#include "genesis/cons.h"
+/* Return true if 'addr' has a lowtag and widetag that correspond,
+ * given that the words at 'addr' are within range for an allocated page.
+ * 'addr' could be a pointer to random data, and this check is merely
+ * a heuristic. False positives are possible. */
+static inline boolean plausible_tag_p(lispobj addr)
+{
+    if (listp(addr))
+        return is_cons_half(CONS(addr)->car)
+            && is_cons_half(CONS(addr)->cdr)
+            // -1 can be left by the */signed=>integer vop
+            // and is also useful as filler on cons pages.
+            && CONS(addr)->car != (uword_t)-1;
+    unsigned char widetag = widetag_of(native_pointer(addr));
+    return other_immediate_lowtag_p(widetag)
+        && lowtag_of(addr) == LOWTAG_FOR_WIDETAG(widetag);
+}
+
+#ifdef LISP_FEATURE_64_BIT
+# define make_filler_header(n) (((uword_t)(n)<<32)|FILLER_WIDETAG)
+# define filler_total_nwords(header) ((header)>>32)
+#else
+# define make_filler_header(n) (((n)<<N_WIDETAG_BITS)|FILLER_WIDETAG)
+# define filler_total_nwords(header) ((header)>>N_WIDETAG_BITS)
+#endif
+
 #endif /* _GC_PRIVATE_H_ */
