@@ -30,8 +30,13 @@
 #endif
 
 #define GENCGC_PAGE_WORDS (GENCGC_PAGE_BYTES/N_WORD_BYTES)
-extern char *page_address(page_index_t);
-int gencgc_handle_wp_violation(void *);
+
+/* Calculate the start address for the given page number. */
+static inline char *page_address(page_index_t page_num) {
+    return (void*)(DYNAMIC_SPACE_START + (page_num * GENCGC_PAGE_BYTES));
+}
+
+int gencgc_handle_wp_violation(void*, void*);
 
 
 #if N_WORD_BITS == 64
@@ -80,6 +85,7 @@ int gencgc_handle_wp_violation(void *);
 #define PAGE_TYPE_SMALL_MIXED  4 // #b100
 #define PAGE_TYPE_CONS         5 // #b101
 #define PAGE_TYPE_CODE         7 // #b111
+#define THREAD_PAGE_FLAG       8
 #define SINGLE_OBJECT_FLAG    16
 #define OPEN_REGION_PAGE_FLAG 32
 #define FREE_PAGE_FLAG        0
@@ -120,7 +126,6 @@ struct page {
     // !!! If bit positions are changed, be sure to reflect the changes into
     // page_extensible_p() as well as ALLOCATION-INFORMATION in sb-introspect
     // and WALK-DYNAMIC-SPACE.
-    unsigned char
         /*
          * The 4 low bits of 'type' are defined by PAGE_TYPE_x constants.
          *  0000 free
@@ -129,15 +134,7 @@ struct page {
          *  ?011 mixed boxed/unboxed non-code objects
          *  ?111 code
          * The next two bits are SINGLE_OBJECT and OPEN_REGION */
-        type :6,
-        /* Whether the page was used at all. This is the only bit that can
-         * be 1 on a free page */
-        need_zerofill :1,
-        /* If this page should not be moved during a GC then this flag
-         * is set. It's only valid during a GC for allocated pages,
-         * and only meaningful for pages in the condemned set.
-         * (It can be spuriously 1 on a page in any generation) */
-        pinned :1;
+    unsigned char type;
 
     /* the generation that this page belongs to. This should be valid
      * for all pages that may have objects allocated, even current
@@ -257,7 +254,7 @@ static inline void gc_set_region_empty(struct alloc_region *region)
 
 static inline void gc_init_region(struct alloc_region *region)
 {
-    region->last_page = 0; // must always be a valid page index
+    // A distinction without a difference (it used to do one more assignment)
     gc_set_region_empty(region);
 }
 
@@ -269,8 +266,7 @@ static inline void gc_init_region(struct alloc_region *region)
 
 /* Find the page index within the page_table for the given
  * address. Return -1 on failure. */
-static inline page_index_t
-find_page_index(void *addr)
+static inline page_index_t find_page_index(void *addr)
 {
     if (addr >= (void*)DYNAMIC_SPACE_START) {
         page_index_t index = ((uintptr_t)addr -
@@ -283,22 +279,34 @@ find_page_index(void *addr)
 
 #define page_single_obj_p(page) ((page_table[page].type & SINGLE_OBJECT_FLAG)!=0)
 
+extern unsigned char* gc_page_pins;
 static inline boolean pinned_p(lispobj obj, page_index_t page)
 {
     extern struct hopscotch_table pinned_objects;
     // Single-object pages can be pinned, but the object doesn't go
-    // in the hashtable. I'm a little surprised that the return value
-    // should be 0 in such case, but I think this never gets called
-    // on large objects because they've all been "moved" to newspace
-    // by adjusting the page table. Perhaps this should do:
-    //   gc_assert(!page_single_obj_p(page))
-    if (!page_table[page].pinned || page_single_obj_p(page)) return 0;
+    // in the hashtable. pinned_p can be queried on those pages,
+    // but the answer is always 'No', because if pinned, the page would
+    // already have had its generation changed to newspace.
+    if (page_single_obj_p(page)) return 0;
+
 #ifdef RETURN_PC_WIDETAG
-    if (widetag_of(native_pointer(obj)) == RETURN_PC_WIDETAG)
+    // Yet another complication from the despised LRA objects- with the
+    // refinement of 8 pin bits per page, we either must set all possible bits
+    // for a simple-fun, or map LRAs to the code base address.
+    if (widetag_of(native_pointer(obj)) == RETURN_PC_WIDETAG) {
+        // The hash-table stores tagged pointers.
         obj = make_lispobj(fun_code_header(native_pointer(obj)),
                            OTHER_POINTER_LOWTAG);
+        page = find_page_index((void*)obj);
+    }
 #endif
-    return hopscotch_containsp(&pinned_objects, obj);
+
+    unsigned char pins = gc_page_pins[page];
+    if (!pins) return 0;
+    unsigned addr_lowpart = obj & (GENCGC_PAGE_BYTES-1);
+    // Divide the page into 8 parts, see whether that part is pinned.
+    unsigned subpage = addr_lowpart / (GENCGC_PAGE_BYTES/8);
+    return (pins & (1<<subpage)) && hopscotch_containsp(&pinned_objects, obj);
 }
 
 // Return true only if 'obj' must be *physically* transported to survive gc.
@@ -309,16 +317,18 @@ from_space_p(lispobj obj)
 {
     gc_dcheck(compacting_p());
     page_index_t page_index = find_page_index((void*)obj);
-    return page_index >= 0
-        && page_table[page_index].gen == from_space
-        && !pinned_p(obj, page_index);
+    // NOTE: It is legal to access page_table at index -1,
+    // and the 'gen' of page -1 is an otherwise unused value.
+    return page_table[page_index].gen == from_space && !pinned_p(obj, page_index);
 }
 
 static boolean __attribute__((unused)) new_space_p(lispobj obj)
 {
     gc_dcheck(compacting_p());
     page_index_t page_index = find_page_index((void*)obj);
-    return page_index >= 0 && page_table[page_index].gen == new_space;
+    // NOTE: It is legal to access page_table at index -1,
+    // and the 'gen' of page -1 is an otherwise unused value.
+    return page_table[page_index].gen == new_space;
 }
 
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
