@@ -49,6 +49,7 @@
 #include "thread.h"
 #include "gc-internal.h"
 #include <fcntl.h>
+#include <sys/prctl.h>
 
 #ifdef LISP_FEATURE_X86
 /* Prototype for personality(2). Done inline here since the header file
@@ -264,6 +265,9 @@ void os_init()
 #ifdef LISP_FEATURE_SB_FUTEX
     futex_init();
 #endif
+#ifdef LISP_FEATURE_SB_DEVEL
+    prctl(PR_SET_PTRACER, PR_SET_PTRACER_ANY);
+#endif
 }
 
 #if (defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)) \
@@ -356,27 +360,6 @@ int os_preinit(char *argv[], char *envp[])
 #endif
     return 0;
 }
-
-void
-os_protect(os_vm_address_t address, os_vm_size_t length, os_vm_prot_t prot)
-{
-#ifdef LISP_FEATURE_SOFT_CARD_MARKS
-    // dynamic space should not have protections manipulated
-    if (find_page_index(address) >= 0)
-        lose("unexpected call to os_protect with software card marks");
-#endif
-    if (mprotect(address, length, prot)) {
-        if (errno == ENOMEM) {
-            lose("An mprotect call failed with ENOMEM. This probably means that the maximum amount\n"
-                 "of separate memory mappings was exceeded. To fix the problem, either increase\n"
-                 "the maximum with e.g. 'echo 262144 > /proc/sys/vm/max_map_count' or recompile\n"
-                 "SBCL with a larger value for GENCGC-PAGE-BYTES in\n"
-                 "'src/compiler/"SBCL_TARGET_ARCHITECTURE_STRING"/parms.lisp'.");
-        } else {
-            perror("mprotect");
-        }
-    }
-}
 
 /*
  * any OS-dependent special low-level handling for signals
@@ -406,7 +389,7 @@ sigsegv_handler(int signal, siginfo_t *info, os_context_t *context)
 #endif
 
 #ifdef LISP_FEATURE_GENCGC
-    if (gencgc_handle_wp_violation(addr)) return;
+    if (gencgc_handle_wp_violation(context, addr)) return;
 #else
     if (cheneygc_handle_wp_violation(context, addr)) return;
 #endif
