@@ -30,7 +30,6 @@
 #include "arch.h"
 #include "interr.h"
 #include "search.h"
-#include "purify.h"
 #include "globals.h"
 #include "lispregs.h"
 #include "interrupt.h"
@@ -86,7 +85,7 @@ struct crash_thread_preamble {
 const int CRASH_PREAMBLE_SIGNATURE =
     (sizeof (struct crash_preamble) << 16) | sizeof (struct crash_thread_preamble);
 
-#if defined LISP_FEATURE_X86_64 // un-tested elsewhere
+#if defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64 // un-tested elsewhere
 #include <errno.h>
 static void checked_write(int fd, void* buf, long nbytes)
 {
@@ -96,7 +95,7 @@ static void checked_write(int fd, void* buf, long nbytes)
 
 #include "immobile-space.h"
 void save_gc_crashdump(char *pathname,
-                       uword_t approx_stackptr_at_gc_start)
+                       lispobj* cur_thread_approx_stackptr)
 {
     extern int pin_all_dynamic_space_code;
     int fd = open(pathname, O_WRONLY|O_CREAT, 0666);
@@ -158,20 +157,27 @@ void save_gc_crashdump(char *pathname,
         int ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,th));
         os_context_t* threadcontext = nth_interrupt_context(0, th);
         uword_t sp;
-        if (ici)
+        if (ici) {
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
             sp = *os_context_register_addr(threadcontext, reg_SP);
 #else
             sp = *os_context_register_addr(threadcontext, reg_CSP);
 #endif
-        else {
+        } else if (th->state_word.state == STATE_DEAD) {
+            gc_assert(th->binding_stack_pointer == th->binding_stack_start);
+            // FIXME: assumes stack grows down
+            sp = (uword_t)th->control_stack_end;
+        } else {
             if (th != get_sb_vm_thread()) {
-              char msg[] = "No stackptr for crash dump\n";
-              write(2, msg, sizeof msg-1);
+              char msg[80];
+              int n = snprintf(msg, sizeof msg,
+                               "thread %p state %d - No stackptr for crash dump\n",
+                               th, th->state_word.state);
+              write(2, msg, n);
               _exit(1);
             }
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
-            sp = approx_stackptr_at_gc_start;
+            sp = (uword_t)cur_thread_approx_stackptr;
 #else
             sp = access_control_stack_pointer(th);
 #endif
@@ -210,9 +216,7 @@ void save_gc_crashdump(char *pathname,
 static cmd call_cmd, dump_cmd, print_cmd, quit_cmd, help_cmd;
 static cmd flush_cmd, regs_cmd, exit_cmd;
 static cmd print_context_cmd, pte_cmd, search_cmd;
-static cmd backtrace_cmd, purify_cmd, catchers_cmd;
-static cmd grab_sigs_cmd;
-static cmd kill_cmd;
+static cmd backtrace_cmd, catchers_cmd;
 static cmd threads_cmd;
 
 extern void gc_stop_the_world(), gc_start_the_world();
@@ -224,7 +228,7 @@ static void suspend_other_threads() {
     // versus doing this loop
     struct thread *th;
     for_each_thread(th) { gc_close_thread_regions(th); }
-    gc_close_collector_regions();
+    gc_close_collector_regions(0);
 }
 static void unsuspend_other_threads() {
 #ifdef LISP_FEATURE_SB_THREAD
@@ -233,14 +237,14 @@ static void unsuspend_other_threads() {
 }
 
 static int save_cmd(char **ptr) {
-#if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_SB_THREAD
     char *name  = parse_token(ptr);
     if (!name) {
         fprintf(stderr, "Need filename\n");
         return 1;
     }
+#if (defined LISP_FEATURE_X86 || defined LISP_FEATURE_X86_64) && defined LISP_FEATURE_SB_THREAD
     suspend_other_threads();
-    save_gc_crashdump(name, (uword_t)__builtin_frame_address(0));
+    save_gc_crashdump(name, (lispobj*)__builtin_frame_address(0));
     unsuspend_other_threads();
 #else
     fprintf(stderr, "Unimplemented\n");
@@ -261,7 +265,7 @@ static int threads_cmd(char __attribute__((unused)) **ptr) {
 static int verify_cmd(char __attribute__((unused)) **ptr) {
     gencgc_verbose = 1;
     suspend_other_threads();
-    verify_heap(0);
+    verify_heap(0, 0);
     unsuspend_other_threads();
     return 0;
 }
@@ -293,13 +297,6 @@ static struct cmd {
     {"d", "(an alias for dump)", dump_cmd},
     {"exit", "Exit this instance of the monitor.", exit_cmd},
     {"flush", "Flush all temp variables.", flush_cmd},
-    /* (Classic CMU CL had a "gc" command here, which seems like a
-     * reasonable idea, but the code was stale (incompatible with
-     * gencgc) so I just flushed it. -- WHN 20000814 */
-    {"grab-signals", "Set the signal handlers to call LDB.", grab_sigs_cmd},
-    {"kill", "Kill ourself with signal number N (useful if running under gdb)",
-     kill_cmd},
-    {"purify", "Purify. (Caveat purifier!)", purify_cmd},
     {"print", "Print object at ADDRESS.", print_cmd},
     {"p", "(an alias for print)", print_cmd},
     {"pte", "Page table entry for address", pte_cmd},
@@ -413,7 +410,7 @@ dump_cmd(char **ptr)
                 if (word != 0 && !is_lisp_pointer(word)
                     && valid_widetag_p(header_widetag(word))) {
                     printf(" %s", widetag_names[header_widetag(word)>>2]);
-                    next_object += sizetab[header_widetag(word)](next_object);
+                    next_object += headerobj_size2(next_object, word);
                 } else if (!is_header(word)) {
                     next_object += CONS_SIZE;
                 } else { // disable decoder if weirdness observed
@@ -449,16 +446,6 @@ pte_cmd(char **ptr)
     return 0;
 }
 
-static int
-kill_cmd(char **ptr)
-{
-#ifndef LISP_FEATURE_WIN32
-    int sig;
-    if (parse_number(ptr, &sig)) kill(getpid(), sig);
-#endif
-    return 0;
-}
-
 static int
 regs_cmd(char __attribute__((unused)) **ptr)
 {
@@ -603,13 +590,6 @@ exit_cmd(char __attribute__((unused)) **ptr)
     return 1; // 'done' flag
 }
 
-static int
-purify_cmd(char __attribute__((unused)) **ptr)
-{
-    purify(NIL, NIL);
-    return 0;
-}
-
 static void
 print_context(os_context_t *context)
 {
@@ -643,12 +623,11 @@ print_context_cmd(char **ptr)
             printf("printing context %d\n", index);
             print_context(nth_interrupt_context(index, thread));
         } else {
-            printf("There aren't that many/few contexts.\n");
             printf("There are %d interrupt contexts.\n", free_ici);
         }
     } else {
         if (free_ici == 0)
-            printf("There are no interrupt contexts!\n");
+            printf("There are no interrupt contexts.\n");
         else {
             printf("There are %d interrupt contexts.\n", free_ici);
             printf("printing context %d\n", free_ici - 1);
@@ -714,34 +693,6 @@ catchers_cmd(char __attribute__((unused)) **ptr)
     return 0;
 }
 
-/* SIGINT handler that invokes the monitor (for when Lisp isn't up to it) */
-static void
-sigint_handler(int __attribute__((unused)) signal,
-               siginfo_t __attribute__((unused)) *info,
-               void *context)
-{
-    extern void ldb_monitor();
-    fprintf(stderr, "\nSIGINT hit at %p\n", (void*)os_context_pc(context));
-    ldb_monitor();
-    fprintf(stderr, "Returning to lisp (if you're lucky).\n");
-}
-
-static int
-grab_sigs_cmd(char __attribute__((unused)) **ptr)
-{
-#ifdef LISP_FEATURE_WIN32
-    fprintf(stderr, "sorry no can do\n"); fflush(stderr);
-#else
-    printf("Grabbing SIGINT.\n");
-    struct sigaction sa;
-    sigemptyset(&sa.sa_mask);
-    sa.sa_flags = SA_SIGINFO;
-    sa.sa_sigaction = sigint_handler;
-    sigaction(SIGINT, &sa, 0);
-#endif
-    return 0;
-}
-
 extern boolean gc_active_p;
 extern FILE *gc_activitylog_file;
 void
@@ -753,7 +704,8 @@ ldb_monitor(void)
     int ambig;
 
     printf("Welcome to LDB, a low-level debugger for the Lisp runtime environment.\n");
-    if (gc_active_p) printf("(GC in progress)\n");
+    if (gc_active_p) printf("(GC in progress, oldspace=%d, newspace=%d)\n",
+                            from_space, new_space);
     if (gc_activitylog_file) fflush(gc_activitylog_file);
     if (!ldb_in) {
 #ifndef LISP_FEATURE_WIN32
@@ -812,7 +764,9 @@ monitor_or_something()
     ldb_monitor();
 }
 
-#ifdef STANDALONE
+#ifdef STANDALONE_LDB
+void gc_stop_the_world() { } // do nothing
+void gc_start_the_world() { } // do nothing
 #include <errno.h>
 #include "core.h"
 #include "gencgc-private.h"
@@ -872,17 +826,18 @@ int load_gc_crashdump(char* pathname)
     struct crash_preamble preamble;
     struct crash_thread_preamble thread_preamble;
     checked_read(fd, &preamble, sizeof preamble);
-    printf("static=%lx nbytes=%x\n", preamble.static_start, (int)preamble.static_nbytes);
-    printf("heap_start=%lx npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
+    printf("static=%"OBJ_FMTX" nbytes=%x\n", preamble.static_start, (int)preamble.static_nbytes);
+    printf("heap_start=%"OBJ_FMTX" npages=%d\n", preamble.dynspace_start, (int)preamble.dynspace_npages);
     // pin_dynspace_code is for display only. It gets recomputed as the
     // logical OR of all threads' values of *GC-PIN-CODE-PAGES*.
     printf("sprof_enabled=%d pin_dynspace_code=%d packages=%p\n",
            preamble.sprof_enabled, preamble.pin_dynspace_code,
            (void*)preamble.lisp_package_vector);
+    lisp_package_vector = preamble.lisp_package_vector;
     sb_sprof_enabled = preamble.sprof_enabled;
     if (preamble.signature != CRASH_PREAMBLE_SIGNATURE)
-        lose("Can't load crashdump: bad header (have %lx, expect %lx)",
-             preamble.signature, (long)CRASH_PREAMBLE_SIGNATURE);
+        lose("Can't load crashdump: bad header (have %"OBJ_FMTX", expect %"OBJ_FMTX")",
+             preamble.signature, (uword_t)CRASH_PREAMBLE_SIGNATURE);
     if (preamble.card_size != GENCGC_CARD_BYTES)
         lose("Can't load crashdump: memory parameters differ");
     gc_card_table_nbits = preamble.card_table_nbits;
@@ -898,7 +853,7 @@ int load_gc_crashdump(char* pathname)
         lose("Didn't map dynamic space where expected: %p vs %p",
              dynspace, (char*)preamble.dynspace_start);
     checked_read(fd, (char*)DYNAMIC_SPACE_START, dynspace_nbytes);
-    fprintf(stderr, "snapshot: %ld pages in use (%ld bytes)\n",
+    fprintf(stderr, "snapshot: %"PRIdPTR" pages in use (%ld bytes)\n",
             next_free_page, dynspace_nbytes);
     checked_read(fd, page_table, sizeof (struct page) * next_free_page);
     recompute_gen_bytes_allocated();
@@ -951,6 +906,10 @@ int load_gc_crashdump(char* pathname)
         }
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
         *os_context_sp_addr(context) = (uword_t)stackptr;
+#ifdef LISP_FEATURE_X86
+        *os_context_register_addr(context,reg_SP) = (os_context_register_t)stackptr;
+#endif
+        gc_assert(*os_context_register_addr(context,reg_SP) == (os_context_register_t)stackptr);
 #else
         *os_context_register_addr(context, reg_CSP) = (uword_t)stackptr;
 #endif
