@@ -14,11 +14,11 @@
 #include "genesis/layout.h"
 #include "genesis/package.h"
 #include "genesis/vector.h"
-#include "getallocptr.h" // for get_alloc_pointer()
 #include "search.h"
 #include "genesis/avlnode.h"
 #include "genesis/sap.h"
 #include "genesis/thread-instance.h"
+#include "print.h"
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -159,8 +159,8 @@ static int find_ref(lispobj* source, lispobj target)
         check_ptr(1, source[1]);
         return -1;
     }
+    scan_limit = headerobj_size(source);
     int widetag = widetag_of(source);
-    scan_limit = sizetab[widetag](source);
     switch (widetag) {
     case INSTANCE_WIDETAG:
     case FUNCALLABLE_INSTANCE_WIDETAG:
@@ -430,20 +430,20 @@ static inline lispobj decode_pointer(uint32_t encoding)
 
 static void maybe_show_object_name(lispobj obj, FILE* stream)
 {
-    extern void safely_show_lstring(lispobj* string, int quotes, FILE *s);
     lispobj package, package_name;
     if (lowtag_of(obj)==OTHER_POINTER_LOWTAG)
         switch(widetag_of(native_pointer(obj))) {
         case SYMBOL_WIDETAG:
             putc(',', stream);
-            if ((package = SYMBOL(obj)->package) == NIL) {
+            package = symbol_package(SYMBOL(obj));
+            if (package == NIL) {
                 fprintf(stream, "#:");
             } else {
                 package_name = ((struct package*)native_pointer(package))->_name;
-                safely_show_lstring(native_pointer(package_name), 0, stream);
+                safely_show_lstring(VECTOR(package_name), 0, stream);
                 fputs("::", stream);
             }
-            safely_show_lstring(native_pointer(SYMBOL(obj)->name), 0, stream);
+            safely_show_lstring(symbol_name(SYMBOL(obj)), 0, stream);
             break;
         }
 }
@@ -463,7 +463,7 @@ static boolean root_p(lispobj ptr, int criterion)
 static lispobj mkcons(lispobj car, lispobj cdr)
 {
     struct cons *cons = (struct cons*)
-        gc_general_alloc(sizeof(struct cons), BOXED_PAGE_FLAG);
+        gc_general_alloc(cons_region, sizeof(struct cons), PAGE_TYPE_CONS);
     cons->car = car;
     cons->cdr = cdr;
     return make_lispobj(cons, LIST_POINTER_LOWTAG);
@@ -474,7 +474,7 @@ static lispobj liststar3(lispobj x, lispobj y, lispobj z) {
 static lispobj make_sap(char* value)
 {
     struct sap *sap = (struct sap*)
-        gc_general_alloc(sizeof(struct sap), BOXED_PAGE_FLAG);
+        gc_general_alloc(unboxed_region, sizeof(struct sap), PAGE_TYPE_UNBOXED);
     sap->header = (1<<N_WIDETAG_BITS) | SAP_WIDETAG;
     sap->pointer = value;
     return make_lispobj(sap, OTHER_POINTER_LOWTAG);
@@ -526,8 +526,7 @@ static lispobj trace1(lispobj object,
                     if (hopscotch_containsp(visited, (lispobj)ptr))
                         fprintf(stderr, "%p ", ptr);
                     else {
-                        lispobj word = *ptr;
-                        int nwords = OBJECT_SIZE(word, ptr);
+                        int nwords = object_size(ptr);
                         fprintf(stderr, "%p+%d ", ptr, nwords);
                     }
                     list1 = cell[1];
@@ -720,7 +719,7 @@ static uword_t build_refs(lispobj* where, lispobj* end,
     boolean count_only = !ss->record_ptrs;
     for ( ; where < end ; where += nwords ) {
         if (ss->ignored_objects && ignorep(where, ss->ignored_objects)) {
-            nwords = OBJECT_SIZE(*where, where);
+            nwords = object_size(where);
             continue;
         }
         ++n_objects;
@@ -731,8 +730,8 @@ static uword_t build_refs(lispobj* where, lispobj* end,
             check_ptr(where[1]);
             continue;
         }
+        nwords = scan_limit = headerobj_size2(where, word);
         int widetag = header_widetag(word);
-        nwords = scan_limit = sizetab[widetag](where);
         switch (widetag) {
         case INSTANCE_WIDETAG:
         case FUNCALLABLE_INSTANCE_WIDETAG:
@@ -785,6 +784,7 @@ static uword_t build_refs(lispobj* where, lispobj* end,
                 }
             }
             break;
+        case FILLER_WIDETAG: continue;
         default:
             if (!(other_immediate_lowtag_p(widetag) && LOWTAG_FOR_WIDETAG(widetag)))
               lose("Unknown widetag %x", widetag);
@@ -834,6 +834,7 @@ static uword_t build_refs(lispobj* where, lispobj* end,
 static void scan_spaces(struct scan_state* ss)
 {
     struct scan_state old = *ss;
+    build_refs((lispobj*)NIL_SYMBOL_SLOTS_START, (lispobj*)NIL_SYMBOL_SLOTS_END, ss);
     build_refs((lispobj*)STATIC_SPACE_OBJECTS_START, static_space_free_pointer, ss);
     show_tally(old, ss);
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
@@ -981,12 +982,18 @@ static int trace_paths(void (*context_scanner)(),
                                   inverted_heap, &scratchpad,
                                   n_pins, pins, context_scanner, criterion);
             lispobj* elt = VECTOR(paths)->data + i;
-            ensure_ptr_word_writable(elt);
+            notice_pointer_store(elt);
             if ((*elt = path) != 0) ++n_found;
         }
         ++i;
     } while (weak_pointers != NIL);
-    ensure_region_closed(&boxed_region, BOXED_PAGE_FLAG);
+    // gc_close_collector_regions() won't suffice here.
+    // Depending on whether this is called from prove_liveness() or
+    // gc_prove_liveness() via collect_garbage(), there may be an open
+    // region that has the THREAD_PAGE_FLAG in its PTE. But that's not
+    // correct for the allocations made for traceroot's answers.
+    ensure_region_closed(unboxed_region, PAGE_TYPE_UNBOXED);
+    ensure_region_closed(cons_region, PAGE_TYPE_CONS);
     os_invalidate(scratchpad.base, scratchpad.end-scratchpad.base);
 #if TRACEROOT_USE_ABSL_HASHMAP
     absl_hashmap_destroy(inverted_heap);
@@ -1046,15 +1053,6 @@ int gc_prove_liveness(void(*context_scanner)(),
         CONS(output)->car = make_fixnum(0);
         return 0;
     }
-    // Put back lowtags on pinned objects, since wipe_nonpinned_words() removed
-    // them. But first test whether lowtags were already repaired
-    // in case prove_liveness() is called after gc_prove_liveness().
-    if (n_pins>0 && !is_lisp_pointer(pins[0])) {
-        int i;
-        for(i=n_pins-1; i>=0; --i) {
-            pins[i] = compute_lispobj((lispobj*)pins[i]);
-        }
-    }
     n_paths = trace_paths(context_scanner, input, paths, ignore,
                           n_pins, (lispobj*)pins, criterion);
     CONS(output)->car = make_fixnum(n_paths);
@@ -1067,6 +1065,7 @@ int gc_prove_liveness(void(*context_scanner)(),
 int prove_liveness(lispobj objects, int criterion)
 {
     extern struct hopscotch_table pinned_objects;
-    extern int gc_n_stack_pins;
-    return gc_prove_liveness(0, objects, gc_n_stack_pins, pinned_objects.keys, criterion);
+    extern int gc_pin_count;
+    extern lispobj* gc_filtered_pins;
+    return gc_prove_liveness(0, objects, gc_pin_count, gc_filtered_pins, criterion);
 }
