@@ -61,6 +61,7 @@
 #include "unaligned.h"
 
 #include "gc.h"
+#include "gc-private.h"
 #include "gencgc-internal.h"
 #include <wincrypt.h>
 #include <stdarg.h>
@@ -791,7 +792,8 @@ set_up_win64_seh_thunk(size_t page_size)
     struct win64_seh_data *seh_data = (void *) WIN64_SEH_DATA_ADDR;
     DWORD64 base = (DWORD64) seh_data;
 
-    uint8_t *dthunk = seh_data->direct_thunk;
+    // 'volatile' works around "warning: writing 1 byte into a region of size 0 [-Wstringop-overflow=]"
+    volatile uint8_t *dthunk = seh_data->direct_thunk;
     dthunk[0] = 0x41; // pop r15
     dthunk[1] = 0x5F;
     dthunk[2] = 0xFF; // call rbx
@@ -801,7 +803,7 @@ set_up_win64_seh_thunk(size_t page_size)
     dthunk[6] = 0xC3; // ret
     dthunk[7] = 0x90; // nop (padding)
 
-    uint8_t *ithunk = seh_data->indirect_thunk;
+    volatile uint8_t *ithunk = seh_data->indirect_thunk;
     ithunk[0] = 0x41; // pop r15
     ithunk[1] = 0x5F;
     ithunk[2] = 0xFF; // call qword ptr [rbx]
@@ -811,7 +813,7 @@ set_up_win64_seh_thunk(size_t page_size)
     ithunk[6] = 0xC3; // ret
     ithunk[7] = 0x90; // nop (padding)
 
-    uint8_t *tramp = seh_data->handler_trampoline;
+    volatile uint8_t *tramp = seh_data->handler_trampoline;
     tramp[0] = 0xFF; // jmp qword ptr [rip+2]
     tramp[1] = 0x25;
     UNALIGNED_STORE32((tramp+2), 2);
@@ -860,9 +862,6 @@ void os_init()
     GetSystemInfo(&system_info);
     os_vm_page_size = system_info.dwPageSize > BACKEND_PAGE_BYTES?
         system_info.dwPageSize : BACKEND_PAGE_BYTES;
-#if defined(LISP_FEATURE_X86)
-    fast_bzero_pointer = fast_bzero_detect;
-#endif
     os_number_of_processors = system_info.dwNumberOfProcessors;
 
 #ifdef LISP_FEATURE_X86_64
@@ -1040,17 +1039,17 @@ static int
 handle_breakpoint_trap(os_context_t *ctx, struct thread* self)
 {
 #ifdef LISP_FEATURE_UD2_BREAKPOINTS
-    if (((unsigned short *)*os_context_pc_addr(ctx))[0] != 0x0b0f)
+    if (((unsigned short *)OS_CONTEXT_PC(ctx))[0] != 0x0b0f)
         return -1;
 #endif
 
     /* Unlike some other operating systems, Win32 leaves EIP
      * pointing to the breakpoint instruction. */
-    (*os_context_pc_addr(ctx)) += TRAP_CODE_WIDTH;
+    OS_CONTEXT_PC(ctx) += TRAP_CODE_WIDTH;
 
     /* Now EIP points just after the INT3 byte and aims at the
      * 'kind' value (eg trap_Cerror). */
-    unsigned trap = *(unsigned char *)(*os_context_pc_addr(ctx));
+    unsigned trap = *(unsigned char *)OS_CONTEXT_PC(ctx);
 
 #ifdef LISP_FEATURE_SB_THREAD
     /* Before any other trap handler: gc_safepoint ensures that
@@ -1132,12 +1131,16 @@ handle_access_violation(os_context_t *ctx,
 
     /* dynamic space */
     page_index_t page = find_page_index(fault_address);
-    if (page != -1 && !PAGE_WRITEPROTECTED_P(page)) {
+    if (page != -1
+#ifndef LISP_FEATURE_SOFT_CARD_MARKS
+        && !PAGE_WRITEPROTECTED_P(page)
+#endif
+        ) {
         os_commit_memory(PTR_ALIGN_DOWN(fault_address, os_vm_page_size),
                          os_vm_page_size);
         return 0;
     }
-    if (gencgc_handle_wp_violation(fault_address)) {
+    if (gencgc_handle_wp_violation(ctx, fault_address)) {
         return 0;
     }
 
@@ -1541,10 +1544,10 @@ io_end_interruptible(HANDLE handle)
 {
     if (!ptr_CancelIoEx)
         return;
-    thread_mutex_lock(&interrupt_io_lock);
+    EnterCriticalSection(&interrupt_io_lock);
     __sync_bool_compare_and_swap(&thread_extra_data(get_sb_vm_thread())->synchronous_io_handle_and_flag,
                                  handle, 0);
-    thread_mutex_unlock(&interrupt_io_lock);
+    LeaveCriticalSection(&interrupt_io_lock);
 }
 #define WITH_INTERRUPTIBLE_IO(handle)      \
     if (!io_begin_interruptible(handle)) { \
@@ -1624,7 +1627,7 @@ int console_handle_p(HANDLE handle)
 
 static __stdcall unsigned int tty_read_line_server(LPVOID arg)
 {
-    thread_mutex_lock(&ttyinput.lock);
+    EnterCriticalSection(&ttyinput.lock);
     while (ttyinput.handle) {
         DWORD nchars;
         BOOL ok;
@@ -1632,7 +1635,7 @@ static __stdcall unsigned int tty_read_line_server(LPVOID arg)
         while (!ttyinput.in_progress)
           SleepConditionVariableCS(&ttyinput.cond_has_client,&ttyinput.lock,INFINITE);
 
-        thread_mutex_unlock(&ttyinput.lock);
+        LeaveCriticalSection(&ttyinput.lock);
 #ifdef LISP_FEATURE_SB_UNICODE
         ok = ReadConsoleW(ttyinput.handle,
                           &ttyinput.buffer[ttyinput.tail],
@@ -1645,7 +1648,7 @@ static __stdcall unsigned int tty_read_line_server(LPVOID arg)
                          &nchars,NULL);
 #endif
 
-        thread_mutex_lock(&ttyinput.lock);
+        EnterCriticalSection(&ttyinput.lock);
 
         if (ok) {
             ttyinput.tail += nchars;
@@ -1653,7 +1656,7 @@ static __stdcall unsigned int tty_read_line_server(LPVOID arg)
         }
         ttyinput.in_progress = 0;
     }
-    thread_mutex_unlock(&ttyinput.lock);
+    LeaveCriticalSection(&ttyinput.lock);
     return 0;
 }
 
@@ -1684,7 +1687,7 @@ win32_tty_listen(HANDLE handle)
     boolean result = 0;
     INPUT_RECORD ir;
     DWORD nevents;
-    thread_mutex_lock(&ttyinput.lock);
+    EnterCriticalSection(&ttyinput.lock);
     if (!tty_maybe_initialize_unlocked(handle))
         result = 0;
 
@@ -1700,7 +1703,7 @@ win32_tty_listen(HANDLE handle)
             }
         }
     }
-    thread_mutex_unlock(&ttyinput.lock);
+    LeaveCriticalSection(&ttyinput.lock);
     return result;
 }
 
@@ -1716,7 +1719,7 @@ static int win32_read_console(HANDLE handle, void* buf, int count)
 
     count = nchars*sizeof(console_char);
 
-    thread_mutex_lock(&ttyinput.lock);
+    EnterCriticalSection(&ttyinput.lock);
 
     if (!tty_maybe_initialize_unlocked(handle)) {
         result = -1;
@@ -1773,7 +1776,7 @@ static int win32_read_console(HANDLE handle, void* buf, int count)
         }
     }
 unlock:
-    thread_mutex_unlock(&ttyinput.lock);
+    LeaveCriticalSection(&ttyinput.lock);
     return result;
 }
 
@@ -1789,7 +1792,7 @@ win32_maybe_interrupt_io(void* thread)
 #endif
 
     if (ptr_CancelIoEx) {
-        thread_mutex_lock(&interrupt_io_lock);
+        EnterCriticalSection(&interrupt_io_lock);
         HANDLE h = (HANDLE)
             InterlockedExchangePointer((volatile LPVOID *)
                                        &thread_extra_data(th)->synchronous_io_handle_and_flag,
@@ -1797,9 +1800,9 @@ win32_maybe_interrupt_io(void* thread)
 
         if (h && (h!=INVALID_HANDLE_VALUE)) {
             if (console_handle_p(h)) {
-                thread_mutex_lock(&ttyinput.lock);
+                EnterCriticalSection(&ttyinput.lock);
                 WakeAllConditionVariable(&ttyinput.cond_has_data);
-                thread_mutex_unlock(&ttyinput.lock);
+                LeaveCriticalSection(&ttyinput.lock);
                 done = 1;
                 goto unlock;
             }
@@ -1809,7 +1812,7 @@ win32_maybe_interrupt_io(void* thread)
             done |= !!ptr_CancelIoEx(h,NULL);
         }
     unlock:
-        thread_mutex_unlock(&interrupt_io_lock);
+        LeaveCriticalSection(&interrupt_io_lock);
     }
     return done;
 }
