@@ -135,8 +135,11 @@
   (sb-sys:with-code-pages-pinned (:dynamic) (gc))
   ;; this should not fail to find FUN at its old address
   (let ((fun (sb-kernel:make-lisp-obj (first *some-object-handles*))))
-    ;; this should fail to find a string at its old address
-    (assert (not (nth-value 1 (sb-kernel:make-lisp-obj (second *some-object-handles*) nil))))
+    ;; To prove that _some_ things moved in memory,
+    ;; assert that we don't see the arbitrary string at its old address.
+    (multiple-value-bind (thing existsp)
+        (sb-kernel:make-lisp-obj (second *some-object-handles*) nil)
+      (assert (or (not existsp) (not (typep thing '(string 7))))))
     ;; this should similarly fail- STRING-TWO was transitively reachable but movable
     (multiple-value-bind (obj validp) (sb-kernel:make-lisp-obj (third *some-object-handles*) nil)
       (if validp
@@ -144,11 +147,12 @@
     ;; (assert (not (nth-value 1 (sb-kernel:make-lisp-obj (third *some-object-handles*) nil))))
     (assert (string= (sb-kernel:%simple-fun-name fun) "two potayto"))))
 
-#+immobile-space
 (with-test (:name :generation-of-fdefn)
-  ;; generation-of broke when fdefns stopped storing a generation in word 0
+  ;; GENERATION-OF broke when fdefns stopped storing a generation in word 0.
+  ;; Normally we expect to see SB-VM:+PSEUDO-STATIC-GENERATION+
+  ;; but allow for varied definition of CORE_PAGE_GENERATION.
   (assert (= (sb-kernel:generation-of (sb-int:find-fdefn 'car))
-             sb-vm:+pseudo-static-generation+)))
+             (sb-kernel:generation-of #'car))))
 
 (with-test (:name :static-fdefn-space)
   (sb-int:dovector (name sb-vm:+static-fdefns+)
@@ -453,7 +457,8 @@
                (/ sb-vm:large-object-size sb-vm:n-word-bytes)))
 (gc)
 #+gencgc
-(with-test (:name :page-protected-p :broken-on :x86)
+(with-test (:name :page-protected-p :broken-on :x86
+                  :fails-on (and :big-endian :ppc64))
   (if (= (sb-kernel:generation-of *vvv*) 0) (gc))
   (assert (= (sb-kernel:generation-of *vvv*) 1))
   (assert (sb-kernel:page-protected-p *vvv*))
@@ -477,3 +482,16 @@
   (gc)
   (let ((marks (sb-kernel:object-card-marks *vvv*)))
     (assert (not (find 1 marks)))))
+
+(with-test (:name :%shrink-vector)
+  (let ((v (make-array 25 :initial-element 'yikes)))
+    (sb-sys:with-pinned-objects (v)
+      ;; Can't call VECTOR-SAP on simple-vector.
+      ;; (Honestly I don't see what purpose that limitation serves)
+      (let ((sap (sb-sys:sap+ (sb-sys:int-sap (sb-kernel:get-lisp-obj-address v))
+                              (- (ash sb-vm:vector-data-offset sb-vm:word-shift)
+                                 sb-vm:other-pointer-lowtag))))
+        (sb-impl::%shrink-vector v 9)
+        (loop for i from 10 to 24
+              do
+           (assert (= (sb-sys:sap-ref-word sap (ash i sb-vm:word-shift)))))))))
